#include <algorithm> // for max
#include <string.h>

#include <iostream>

#include "tasks.h"
#include "task_io.h"

void Task::init(unsigned long wcet,
		unsigned long period,
		unsigned long deadline)
{
    this->wcet     = wcet;
    this->period   = period;
    this->deadline = deadline;
}

bool Task::has_implicit_deadline() const
{
    return deadline == 0 || deadline == period;
}

bool Task::has_constrained_deadline() const
{
    return deadline <= period;
}

bool Task::is_feasible() const
{
    return get_deadline() >= get_wcet()
        && get_period() >= get_wcet()
        && get_wcet() > 0;
}

void Task::get_utilization(mpq_class &util) const
{
    // assumes period != 0
    util  = get_wcet();
    util /= get_period();
}

void Task::get_density(mpq_class &density) const
{
    // assumes deadline != 0
    density  = get_wcet();
    density /= get_deadline();
}

std::ostream& operator<<(std::ostream &os, const Task &t)
{
    os << "Task(" << t.get_wcet() << ", " << t.get_period();
    if (!t.has_implicit_deadline())
        os << ", " << t.get_deadline();
    os << ")";
    return os;
}

TaskSet::TaskSet(unsigned int num_tasks)
{
    this->num_tasks = num_tasks;
    allocate();
}

TaskSet::TaskSet(const TaskSet &original)
{
    num_tasks = original.num_tasks;
    allocate();
    memcpy(tasks, original.tasks, sizeof(Task) * num_tasks);
}

void TaskSet::allocate()
{
    tasks = new Task[num_tasks];
}

TaskSet::~TaskSet()
{
    delete[] tasks;
}

#define FORALL(i, pred)                             \
    for (unsigned int i = 0; i < num_tasks; i++)    \
    {                                               \
        if (!pred)                                  \
            return false;                           \
    }                                               \
    return true;                                    \

bool TaskSet::has_only_implicit_deadlines() const
{
    FORALL(i, tasks[i].has_implicit_deadline());
}

bool TaskSet::has_only_constrained_deadlines() const
{
    FORALL(i, tasks[i].has_constrained_deadline());
}

bool TaskSet::has_only_feasible_tasks() const
{
    FORALL(i, tasks[i].is_feasible());
}

void TaskSet::get_utilization(mpq_class &util) const
{
    mpq_class tmp;
    util = 0;
    for (unsigned int i = 0; i < num_tasks; i++)
    {
        tasks[i].get_utilization(tmp);
        util += tmp;
    }
}

void TaskSet::get_density(mpq_class &density) const
{
    mpq_class tmp;
    density = 0;
    for (unsigned int i = 0; i < num_tasks; i++)
    {
        tasks[i].get_density(tmp);
        density += tmp;
    }
}

void TaskSet::get_max_density(mpq_class &max_density) const
{
    mpq_class tmp;
    max_density = 0;

    for (unsigned int i = 0; i < num_tasks; i++)
    {
        tasks[i].get_density(tmp);
        max_density = std::max(max_density, tmp);
    }
}

bool TaskSet::is_not_overutilized(unsigned int num_processors) const
{
    mpq_class util;
    get_utilization(util);
    return util <= num_processors;
}

