import optparse
import sys
import binpack

import gedf

from math import ceil

from model.tasks import Task, TaskSystem

from generator.tasksets import DIST

AGE_OF_THE_UNIVERSE = 13.75e9 # years

def comment(str, out=sys.stdout):
    out.write("# %s\n" % str)
    out.flush()

def write_bash_output(ts, want_part=False, out=sys.stdout):
    for t in ts:
        if want_part:
            out.write("task %d %d %d    # util:%.3f\n" %
                      (t.cost, t.period, t.partition, t.utilization()))
        else:
            out.write("task %d %d # util: %.3f\n" %
                      (t.cost, t.period, t.utilization()))

def part_and_split(ts, cpus):
    ok = False
    count = 0
    while not ok:
        try:
            parts = binpack.worst_fit_decreasing(ts, cpus,
                                                 weight=Task.utilization,
                                                 misfit=binpack.report_failure)
            parts = [TaskSystem(p) for p in parts]
            ok = True
        except binpack.DidNotFit, ex:
            count += 1
            if count > 100:
                print "Giving up."
                raise ex
#            print ex
            t = ex.item
#            print ts.utilization(), len(ts), t
            ts.remove(t)
            t1 = Task(t.cost, t.period * 2)
            t2 = Task(t.cost, t.period * 2)
            ts.append(t1)
            ts.append(t2)
 #           print ts.utilization(), len(ts), t1, t2
    for i, part in enumerate(parts):
        comment("partition %d: %d tasks, util=%.2f" %
                (i, len(part), part.utilization()))
        for t in part:
            t.partition = i

def quantize(ts):
    for t in ts:
        t.cost = ceil(t.cost)

def gen_task_set(cpus, gen, umax=None, tasks=None,
                 want_part=False):
    if umax is None:
        umax = cpus
    ts = gen(umax, tasks)
    quantize(ts)
    while ts.utilization() > umax:
        del ts[-1]
    if want_part:
        part_and_split(ts, cpus)
    return ts

def write_bash_task_set(ts, cpus, want_part, out, test_gedf):
    comment("%d tasks, total utilization: %.2f" % (len(ts), ts.utilization()),
            out)

    hyp    = ts.hyperperiod()
    hyp_yr = ts.hyperperiod() / 1000.0 / 3600.0 / 24.0 / 365.0

    fmt = lambda num: ("%.2e" if num > 1e9 else "%.2f") % num

    comment("Hyperperiod: %sms "
            "(%s years, %s times the age of the universe)" %
            (fmt(hyp), fmt(hyp_yr), fmt(hyp_yr / AGE_OF_THE_UNIVERSE)),
            out)

    write_bash_output(ts, want_part, out)

    ts.make_integral_us()

    for m in test_gedf:
        ok = gedf.is_schedulable(m, ts)
        comment("G-EDF schedulability test for %d CPUs: %s" %
                (m, "passed" if ok else "failed"),
                out)

def generate_bash_task_sets(cpus, gen, umax=None, tasks=None,
                            want_part=False, test_gedf=[],
                            out=sys.stdout):

    ts = gen_task_set(cpus, gen, umax, tasks, want_part)
    write_bash_task_set(ts, cpus, want_part, out, test_gedf)

o = optparse.make_option
opts = [
    o('-d', '--distribution', action='store', dest='dist', type='choice',
      choices=DIST.keys(),
      help='what distribution should be used for task set generation?'),

    o('-m', '--num-cpus', action='store', dest='cpus', type='int',
      help='The number of CPUs in the system.'),

    o('-u', '--max-util', action='store', dest='umax', type='float',
      help='Maximum task set utilization.'),

    o('-o', '--output', action='store', dest='file',
      help='where to store the result'),

    o(None, '--max-tasks', action='store', type='int', dest='tasks',
      help='maximum number of tasks'),

    o(None, '--part', action='store_true', dest='want_part',
      help='Pre partition task set (generator only).'),

    o(None, '--test-gedf', action='store_true', dest='test_gedf',
      help='Output if taskset is schedulable.')

    ]

defaults = {
    'dist'         : 'uni_mixed',
    'umax'         : None,
    'cpus'         : 32,
    'tasks'        : None,    # default: no limit
    'file'         : None,

    'want_part'    : False,
    'test_gedf'    : False,
    }


if __name__ == '__main__':
    parser = optparse.OptionParser(option_list=opts)
    parser.set_defaults(**defaults)
    (options, args) = parser.parse_args()

    f = None
    if options.file:
        try:
            f  = open(options.file, 'w')
            out = f
        except IOError, msg:
            print >> sys.stderr, msg

    generate_bash_task_sets(options.cpus, DIST[options.dist],
                            options.umax, options.tasks,
                            options.want_part,
                            [options.cpus] if options.test_gedf else [])

    if f:
        f.close()
