#!/usr/bin/env python

import sys
import os
import signal

from datetime import datetime
from math import ceil
from itertools import cycle
from functools import partial
from random import uniform

from util import Timer, Counter
from model import serialize
import generator.tasksets as gen
import cpptest
import gedf

MAX_CPUS=1000

def run_py_test(ts, is_schedulable, delim=None, max_cpus=None, start_cpu=None):
    if start_cpu is None:
        m = int(ceil(ts.utilization()))
    else:
        m = start_cpu
    while not is_schedulable(m, ts) and (max_cpus is None or m < max_cpus):
        m += 1
        if not delim is None:
            print delim
    return m

def run_cpp_test(ts, test_class, delim=None, max_cpus=None, start_cpu=None):
    if start_cpu is None:
        m = int(ceil(ts.utilization()))
    else:
        m = start_cpu
    fts = ts.as_cpp_taskset()
    while not test_class(m).is_schedulable(fts) \
            and (max_cpus is None or m < max_cpus):
        m += 1
        if not delim is None:
            print delim
    return m

def cpp_rta(m, min_delta=1000):
    return gedf.cpp.RTAGedf(m, min_delta)

tests = [
    ("GFB", gedf.gfb.is_schedulable, gedf.cpp.GFBGedf),
    ("BAK", gedf.bak.is_schedulable, gedf.cpp.BakerGedf),
    ("BCL", gedf.bcl.is_schedulable, gedf.cpp.BCLGedf),
    ("BCL-I", gedf.bcli.is_schedulable, gedf.cpp.BCLIterativeGedf),
    ("SKB", gedf.bar.is_schedulable, gedf.cpp.BaruahGedf),
    ("RTA", partial(gedf.rta.is_schedulable,  min_fixpoint_step=1000), cpp_rta),
    ("FF-DBF", gedf.ffdbf.is_schedulable, gedf.cpp.FFDBFGedf),
    ("ALL", partial(gedf.all.is_schedulable_py, baruah_threshold=None), gedf.cpp.GlobalEDF),
    ]

def run_test_loop(constrained=False, stop_after=None, umax=10.0,
                  want_us=True, test_names=None):
    ok  = 0
    bad = 0
    ts_count = 0
    if not test_names:
        # use all
        tests_to_use = tests
    else:
        tests_to_use = []
        for t in tests:
            if t[0] in test_names:
                tests_to_use.append(t)
    timers  = [(Timer(), Timer()) for _ in xrange(len(tests_to_use))]
    totals  = [Counter() for _ in xrange(len(tests_to_use))]
    skipped = [Counter() for _ in xrange(len(tests_to_use))]

    sim_timer   = Timer()
    sim_counter = Counter()

    try:
        for gn in cycle(gen.DIST.keys()):
            g = gen.DIST[gn]
            ucap = uniform(1.0, umax)
            ts = g(ucap, None)
            if constrained:
                for t in ts:
                    t.deadline = uniform(t.cost, t.period)
            ts.make_integral_us()
            if not want_us:
                ts.round_to_ms()
            print "%s: testing %.2f / %d task set / %.2f max density" \
                % (gn, ts.utilization(), len(ts), ts.max_density())
#            for t in ts:
#                print t, t.density()

            sim_timer.start()
            lower_bound = run_py_test(ts, gedf.sim.no_counter_example,
                                      max_cpus=MAX_CPUS)
            sim_timer.stop()
            sim_counter.inc(lower_bound)
            print ' lower:', lower_bound

            for ((label, py, cpp), (t1, t2), counter, skip) in \
                    zip(tests_to_use, timers, totals, skipped):
                print ' -- py', label, datetime.now()
                t1.start()
                py_res  = run_py_test(ts, py, max_cpus=MAX_CPUS)
                t1.stop()
                print ' -- cpp', label, datetime.now()
                t2.start()
                cpp_res = run_cpp_test(ts, cpp, max_cpus=MAX_CPUS)
                t2.stop()
                if py_res != cpp_res:
                    bad += 1
                    print "Task set %d failed %s comparison: py=%d cpp=%d" % \
                        (ok + bad, label, py_res, cpp_res)
                    fname = "/tmp/failed-task-set-%d-%d-%s.xml" % (bad, os.getpid(), label)
                    serialize.write_ts(ts, fname)
                    print "Dumped task set to %s." % fname
                elif py_res < lower_bound:
                    bad += 1
                    print "Task set %d failed %s bound: test=%d lower=%d" % \
                        (ok + bad, label, test, lower_bound)
                    fname = "/tmp/failed-task-set-%d-%d-%s.xml" % (bad, os.getpid(), label)
                    serialize.write_ts(ts, fname)
                    print "Dumped task set to %s." % fname
                else:
                    print ' =>', py_res
                    if py_res != MAX_CPUS:
                        counter.inc(py_res)
                    else:
                        skip.inc()
                    ok += 1
            ts_count += 1
            if stop_after != None and ts_count >= stop_after:
                break
            # make sure output is produced
            sys.stdout.flush()

    except KeyboardInterrupt:
        pass
    print "Test aborted; %d ok and %d bad, tested %d tasksets." \
        % (ok, bad, ts_count)
    print "%-7s%17s%17s%17s%17s%17s" % ("Test", "Python", "C++", "Ratio", "Total", "Skipped")
    for ((label, py, cpp), (t1, t2), ctr, skip) in \
            zip(tests_to_use, timers, totals, skipped):
        print "%-7s%17s%17s%17.2f%17d%17d" \
            % (label, t1.elapsed, t2.elapsed,
               t1.seconds() / t2.seconds() if t2.seconds() != 0 else 0,
               ctr.val, skip.val)
    print "Sim: ", sim_timer.elapsed, "Total:", sim_counter.val


def run_test_file(file, labels=[]):
    ts = serialize.load(file)
    ts.make_integral()
    print 'Taskset: '
    print ts
    lower_bound = run_py_test(ts, gedf.sim.no_counter_example,
                              max_cpus=MAX_CPUS)
    print 'Lower bound:', lower_bound
    print '==='
    for (label, py, cpp) in tests:
        if label in labels or not labels:
            print 'Py:'
            py_res  = run_py_test(ts, py, '---')
            print '===\nC++:'
            cpp_res = run_cpp_test(ts, cpp, '---')
            print "===\n%-5s: py=%d cpp=%d" % (label, py_res, cpp_res)


def sig_handler(signum, frame):
    print "Signaled! %d" % signum
    raise KeyboardInterrupt

def abort_on_sigusr1():
    signal.signal(signal.SIGUSR1, sig_handler)

if __name__ == '__main__':
    import sys
    args = sys.argv[1:]
    if len(args) >= 1:
        if args[0] == 'loop':
            if len(args) > 1:
                constrained = args[1] == 'constrained'
            else:
                constrained = False
            abort_on_sigusr1()
            run_test_loop(constrained=constrained, umax=6.0)
        else:
            run_test_file(args[0], args[1:])
    else:
        print "Usage: %s <taskset.xml> TEST1 TEST2..." % sys.argv[0]
