import random

from math import ceil, floor

from util import vector_xadd

def util_range(start, end, step):
    x = start
    while x <= end:
        yield x
        x += step

def gen_task_system(generator, umax, prng_state, tmax=None):
    # configure task set generation
    random.setstate(prng_state)
    # generate task set
    return generator(umax, tmax)

def make_bins(num_tests, umax, binsize):
    # initialize to zero, one extra column for total
    num_rows = int(ceil(umax / float(binsize)))
    return [[0] * (num_tests + 1) for _ in xrange(num_rows)]

def grow_and_test_seq(num_task_systems,
                      generator, cpus, infeasible,
                      tests, bins=None, binsize=0.1,
                      make_test_ts = lambda ts: ts):

    if not bins:
        bins = make_bins(len(tests), cpus, binsize)

    def get_bins_for_util(u):
        "lookup row in bins matrix"
        row = int(floor(u / binsize))
        row = min(row, len(bins) - 1)
        return bins[row]

    count = 0

    while count < num_task_systems:
        # 1) generate cpus + 1 tasks
        ts = generator(None, cpus + 1)
        ts.make_integral_us()
        testing = [True for _ in tests]

        # 2) test and grow
        while not infeasible(ts) and count < num_task_systems:
            b = get_bins_for_util(ts.utilization())
            count += 1
            # count total
            b[0] = b[0] + 1
            tts = make_test_ts(ts)
            for i,t in enumerate(tests):
                # see if this one was ok last time
                if testing[i]:
                    if t(tts):
                        # count test success
                        b[i + 1] = b[i + 1] + 1
                    else:
                        testing[i] = False
            # grow
            task = generator(None, 1)
            task.make_integral_us()
            ts += task

    return bins

def no_parallelism():
    global _grow_and_test
    _grow_and_test = grow_and_test_seq

try:
    # Python >= 2.6
    import multiprocessing as par

    from util import add_disable_parallelism_callback

    _generator = None
    _tests = None
    _make_test_ts = None
    _infeasible = None

    def _grow_and_test_helper(num_task_systems, cpus, binsize):
        return grow_and_test_seq(
            num_task_systems, _generator, cpus,
            _infeasible, _tests, None, binsize, _make_test_ts)

    # create and test task systems in parallel
    def grow_and_test_par(
        num_task_systems,
        generator, cpus, infeasible,
        tests, bins=None, binsize=0.1,
        make_test_ts = lambda ts: ts):

        global _generator
        global _tests
        global _make_test_ts
        global _infeasible

        _generator    = generator
        _tests        = tests
        _make_test_ts = make_test_ts
        _infeasible   = infeasible

        workers = par.Pool()
        cores = par.cpu_count()
        chunks = [num_task_systems / cores for _ in xrange(cores)]
        # last guy get's the rest
        chunks[-1] = num_task_systems - (cores - 1) * chunks[0]
        results = [workers.apply_async(_grow_and_test_helper,
                                       (chunk, cpus, binsize))
                   for chunk in chunks]
        bins = None
        for res in results:
            b = res.get()
            if not bins:
                bins = b
            else:
                # combine each row
                for r1, r2 in zip(bins, b):
                    vector_xadd(r1, r2)
        return bins

    _grow_and_test = grow_and_test_par

    add_disable_parallelism_callback(no_parallelism)

except ImportError:
    # must do it sequentially
    _grow_and_test = grow_and_test_seq


def grow_and_test(*args, **kargs):
    return _grow_and_test(*args, **kargs)

