import random

from itertools import imap

from overheads import SystemOverheads, Overheads, CPMD, const
import sample

from exp   import gen_task_system, util_range
from util  import us2ms, ms2us, sec2us, flatten, storage

from model.tasks import TaskSystem

import gedf
from semipart import edfwm, edffm, npsf, cedf, edfos, cfl

from util import ExecTimer, disable_parallelism

benchmark_data = []

def record_benchmark_data(tests):
    global benchmark_data
    benchmark_data = [
        ['TOTAL'] + [t.seconds() for t in tests],
        ['MAX']   + [t.seconds_max() for t in tests],
        ['AVG']   + [t.seconds_avg() for t in tests],
        ['COUNT'] + [t.count for t in tests],
        ]

def make_cedf_test(generator,
                   cluster_size, num_clusters,
                   sched_test,
                   cpmd_cost,
                   oheads):
    """
    generator --- task set generator
    """

    # (ucap, wss) -> schedulable?
    def cedf_test(ucap, wss, prng_state):
        # get migration overhead for working set size
        cache_cost = cpmd_cost(wss)
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # we have per-cluster overheads for C-EDF
        so = SystemOverheads(reference=oheads, n=len(ts))
        so.tick_period = ms2us(1)

        # convert to microsecond resolution
        ts.make_integral_us()

        bins = cedf.assign_ts_with_overheads(ts, num_clusters, cluster_size,
                                             so, cache_cost, want_bins=True)

        if bins:
            # test each cluster one at a time
            for b in bins:
                cluster = TaskSystem()
                # convert bin to task system
                for it in b.items:
                    # it is an inflated task wrapper
                    t = it.params.task
                    # update cost
                    t.cost = it.inflated_cost
                    cluster.append(t)
#                print 'calling %s(%d,...)' % (sched_test.__name__, cluster_size)
                if not sched_test(cluster_size, cluster):
                    return False
            return True
        else:
            # couldn't even bin-pack
            return False

    return cedf_test

def make_cfl_test(generator, cluster_size, num_clusters, sched_test,
                  cpmd_cost, oheads, bound_version):
    """
    generator --- task set generator
    """

    # (ucap, wss) -> schedulable?
    def cfl_test(ucap, wss, prng_state):
        # get migration overhead for working set size
        cache_cost = cpmd_cost(wss)
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # we have per-cluster overheads for C-EDF
        so = SystemOverheads(reference=oheads, n=len(ts))
        so.tick_period = ms2us(1)

        # convert to microsecond resolution
        ts.make_integral_us()

        result = cfl.assign_ts_with_overheads(ts, num_clusters, cluster_size,
                                            so, cache_cost)
        if (bound_version):
            if (result):
                return max([t.tardiness for t in ts])
            else:
                return 0
        else:
            return result

    return cfl_test

def make_edfwm_test(generator,
                    num_cpus,
                    oheads,
                    make_cost_function,
                    bound_version):
    # (ucap, wss) -> schedulable?
    def edfwm_test(ucap, wss, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # account for CPMD based on reference migration cost
        cf = make_cost_function(wss)

        so = SystemOverheads(reference=oheads, n=len(ts),
                             migration_cost=cf)
        so.tick_period = ms2us(1)

        # convert to microsecond resolution
        ts.make_integral_us()

        result = edfwm.assign_ts_with_overheads(ts, num_cpus, so)
        if (bound_version):
            if (result):
                return max([t.tardiness for t in ts])
            else:
                return 0
        else:
            return result

    return edfwm_test


def make_cnpsf_test(generator,
                    num_cpus,
                    cluster_size,
                    oheads,
                    make_cost_function,
                    max_cpmd,
                    delta):
    # (ucap, wss) -> schedulable?
    def cnpsf_test(ucap, wss, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # account for CPMD based on reference migration cost
        cf = make_cost_function(wss)

        so = SystemOverheads(reference=oheads, n=len(ts),
                             migration_cost=cf)
        so.tick_period = ms2us(1)

        # convert to microsecond resolution
        ts.make_integral_us()

        return npsf.clustered_assign_ts_with_overheads(ts, num_cpus, cluster_size,  delta, so,
                                             max_cpmd(wss))
    return cnpsf_test


def make_npsf_test(generator,
                   num_cpus,
                   oheads,
                   make_cost_function,
                   cpmd,
                   delta):
    # (ucap, wss) -> schedulable?
    def npsf_test(ucap, wss, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # account for CPMD based on reference migration cost
        cf = make_cost_function(wss)

        so = SystemOverheads(reference=oheads, n=len(ts),
                             migration_cost=cf)
        so.tick_period = ms2us(1)

        # convert to microsecond resolution
        ts.make_integral_us()

        return npsf.assign_ts_with_overheads(ts, num_cpus, delta, so,
                                             cpmd.max_cost(wss))

    return npsf_test


def make_edffm_test(generator,
                    num_cpus,
                    oheads,
                    make_cost_function,
                    bound_version):
    # (ucap, wss) -> schedulable?
    def edffm_test(ucap, wss, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # account for CPMD based on reference migration cost
        cf = make_cost_function(wss)

        # convert to microsecond resolution
        ts.make_integral_us()

        so = SystemOverheads(reference=oheads, n=len(ts),
                             migration_cost=cf)

        # try to assign tasks
        result = edffm.assign_ts_with_overheads(ts, num_cpus, so)
        if (bound_version):
            if (result):
                return max([t.tardiness for t in ts])
            else:
                return 0
        else:
            return result

    return edffm_test

def make_edfos_test(generator,
                    num_cpus,
                    oheads,
                    make_cost_function,
                    bound_version):
    # (ucap, wss) -> schedulable?
    def edfos_test(ucap, wss, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        # account for CPMD based on reference migration cost
        cf = make_cost_function(wss)

        # convert to microsecond resolution
        ts.make_integral_us()

        so = SystemOverheads(reference=oheads, n=len(ts),
                             migration_cost=cf)

        # try to assign tasks
        result = edfos.assign_ts_with_overheads(ts, num_cpus, so)
        if (bound_version):
            if (result):
                return max([t.tardiness for t in ts])
            else:
                return 0
        else:
            return result
        

    return edfos_test


# Hard-code machine topology of Ludwig.
#
# Socket0  Socket1  Socket2  Socket3
# ------   -------  -------  -------
# | 0, 4|  | 1, 5|  | 2, 6|  | 3, 7|
# | 8,12|  | 9,13|  |10,14|  |11,15|
# |16,20|  |17,21|  |18,22|  |19,23|
# -------  -------  -------  -------

def ludwig_l1(x, y):
    return x == y

def ludwig_l2(x, y):
    # x left column, y right column, or # y left column, x, right column
    return (x % 8 < 4 and x + 4 == y) or \
        (y % 8 < 4 and x - 4 == y)

def ludwig_l3(x, y):
    # same socket
    # not a a shared L2
    # not identical
    return (y % 4) == (x % 4) and \
        not ludwig_l2(x, y) and \
        x != y

def ludwig_mem(x, y):
    return abs(y - x) % 4 != 0

# ludwig_id = floor(cpuid / 6) + (cpuid % 6) * 4
def map_ludwig_cpuid(linear_cpuid):
    lcpuid = int(linear_cpuid)
    return (lcpuid / 6) + (lcpuid % 6) * 4

def ludwig_shared_cache(x, y):
    if ludwig_l1(x, y):
        return CPMD.L1
    elif ludwig_l2(x, y):
        return CPMD.L2
    elif ludwig_l3(x, y):
        return CPMD.L3
    else:
        return CPMD.MEM

def linear_shared_cache(x, y):
    """Ludwig-like core grouping with linear CPU ids"""
    if x == y:
        return CPMD.L1
    elif x / 2 == y / 2:
        return CPMD.L2
    elif x / 6 == y /  6:
        return CPMD.L3
    else:
        return CPMD.MEM

def compare_mappings():
    for x in range(24):
        for y in range(24):
            lx, ly = map_ludwig_cpuid(x), map_ludwig_cpuid(y)
            c = linear_shared_cache(x, y)
            lc = ludwig_shared_cache(lx, ly)
            print "%s: (%d,%d)->%d != (%d,%d)->%d" % ("OK" if c == lc else "ERROR",
                                                      x, y, c, lx, ly, lc)

def make_wss_cost_function(cpmd, cache_level):

    def make_cpmd(wss):
        cost_table = [cost(wss) for cost in cpmd.mem_hierarchy]
        # Sanitize cost table---preemptions that are more expensive than
        # migrations are due to measurement error and should be ignored.
        cost_table[CPMD.L2]   = max(cost_table[CPMD.L2],  cost_table[CPMD.L1])
        cost_table[CPMD.L3]   = max(cost_table[CPMD.L3],  cost_table[CPMD.L2])
        cost_table[CPMD.MEM]  = max(cost_table[CPMD.MEM], cost_table[CPMD.L3])
        def cpmd_cost(from_cpu, to_cpu):
            return cost_table[cache_level(from_cpu, to_cpu)]
        return cpmd_cost
    return make_cpmd

def make_sample_ucap_range(tests, samples, points, **kargs):
    def ucap_sched_exp(wss):
        # sample
        gen_setup_point = lambda u: lambda: [u, wss, random.getstate()]
        partial_result = sample.range(points,
                                      gen_setup_point,
                                      tests, samples=samples, **kargs)
        for row in partial_result:
            # add cache cost coordinate
            row.insert(0, wss)
        return partial_result

    return ucap_sched_exp

# Number of CPUs (24) hard-coded.
def run(generator, soft,
        samples,
        ucap_step, # steps at each point
        ws_sizes,       # working set sizes, the sampling points
        want_benchmark,
        **kargs):

    cpus = 24

    if soft:
        ############# SOFT ################
        ohead_cedf    = Overheads.from_file('ohead-data/rtas11_C-EDF_avg.csv')
        ohead_cfl     = Overheads.from_file('ohead-data/oh_host=ludwig_scheduler=C-FL-split-L3_stat=avg.csv')
        ohead_npsf    = Overheads.from_file('ohead-data/rtas11_NPS-F_avg.csv')
        ohead_cnpsf   = Overheads.from_file('ohead-data/rtas11_C-NPS-F_avg.csv')
        ohead_edfwm   = Overheads.from_file('ohead-data/rtas11_EDF-WM_avg.csv')
        ohead_edfwm_wc   = Overheads.from_file('ohead-data/rtas11_EDF-WM_wc.csv')
        ohead_edffm   = Overheads.from_file('ohead-data/rtas11_EDF-fm_avg.csv')
        ohead_edfos   = Overheads.from_file('ohead-data/sa_avg_ohead_dist=uni_light_algo=EDF-os.csv')

        idle_cpmd = CPMD.from_file('ohead-data/pmo-ospert10-idle-avg.csv')
        idle_cpmd_wc = CPMD.from_file('ohead-data/pmo-ospert10-idle-wc.csv')
        load_cpmd = CPMD.from_file('ohead-data/pmo-ospert10-load-avg.csv')
        load_cpmd_wc = CPMD.from_file('ohead-data/pmo-ospert10-load-wc.csv')

        cedf_schedulable = gedf.soft.has_bounded_tardiness
        cfl_schedulable = gedf.soft.has_bounded_tardiness
        cluster_size = 6
        cedf_migrations = CPMD.L3
        cfl_migrations = CPMD.L3


    else:
        ############## HARD ##############
        ohead_cedf    = Overheads.from_file('ohead-data/rtas11_P-EDF_wc.csv')
        ohead_cfl     = Overheads.from_file('ohead-data/oh_host=ludwig_scheduler=C-FL-split-L3_stat=wc.csv')
        ohead_npsf    = Overheads.from_file('ohead-data/rtas11_NPS-F_wc.csv')
        ohead_cnpsf   = Overheads.from_file('ohead-data/rtas11_C-NPS-F_wc.csv')
        ohead_edfwm   = Overheads.from_file('ohead-data/rtas11_EDF-WM_wc.csv')
        ohead_edffm   = Overheads.from_file('ohead-data/rtas11_EDF-fm_wc.csv')
        ohead_edfos   = Overheads.from_file('ohead-data/sa_wc_ohead_dist=uni_light_algo=EDF-os.csv')

        idle_cpmd = CPMD.from_file('ohead-data/pmo-ospert10-idle-wc.csv')
        load_cpmd = CPMD.from_file('ohead-data/pmo-ospert10-load-wc.csv')

        cedf_schedulable = gedf.is_schedulable
        cluster_size = 1
        cedf_migrations = CPMD.L1


    # EDF-WM data set does not have data for PULL_TIMER.
    # Andrea verified that PULL_TIMER is equivalent under EDF-fm.
    # So use EDF-fm data for EDF-WM PULL_TIMER for now.
    ohead_edfwm.pull_timer = ohead_edffm.pull_timer

    # combine migration costs with machine topology
    idle_cost = make_wss_cost_function(idle_cpmd, linear_shared_cache)
    idle_cost_wc = make_wss_cost_function(idle_cpmd_wc, linear_shared_cache)
    load_cost = make_wss_cost_function(load_cpmd, linear_shared_cache)
    load_cost_wc = make_wss_cost_function(load_cpmd_wc, linear_shared_cache)


#    tests = [
#        # C-EDF (load)
#        make_cedf_test(generator,  cluster_size, cpus / cluster_size,
#                       cedf_schedulable, load_cpmd.mem_hierarchy[cedf_migrations],
#                       ohead_cedf),
#        # C-EDF (idle)
#        make_cedf_test(generator,  cluster_size, cpus / cluster_size,
#                       cedf_schedulable, idle_cpmd.mem_hierarchy[cedf_migrations],
#                       ohead_cedf),
#
#        # EDF-WM (load)
#        make_edfwm_test(generator, cpus, ohead_edfwm, load_cost),
#        # EDF-WM (idle)
#        make_edfwm_test(generator, cpus, ohead_edfwm, idle_cost),
#
#        # NPS-F (load, delta=1)
#        make_npsf_test(generator, cpus, ohead_npsf, load_cost, load_cpmd, 1),
#        # NPS-F (load, delta=4)
#        make_npsf_test(generator, cpus, ohead_npsf, load_cost, load_cpmd, 4),
#
#        # NPS-F (idle, delta=1)
#        make_npsf_test(generator, cpus, ohead_npsf, idle_cost, idle_cpmd, 1),
#        # NPS-F (idle, delta=4)
#        make_npsf_test(generator, cpus, ohead_npsf, idle_cost, idle_cpmd, 4),
#
#        # C-NPS-F (load, delta=1) --- only L3-based clusters.
#        make_cnpsf_test(generator, cpus, 6, ohead_cnpsf, load_cost, load_cpmd.L3, 1),
#        # C-NPS-F (load, delta=4)
#        make_cnpsf_test(generator, cpus, 6, ohead_cnpsf, load_cost, load_cpmd.L3, 4),
#
#        # C-NPS-F (idle, delta=1)
#        make_cnpsf_test(generator, cpus, 6, ohead_cnpsf, idle_cost, idle_cpmd.L3, 1),
#        # C-NPS-F (idle, delta=4)
#        make_cnpsf_test(generator, cpus, 6, ohead_cnpsf, idle_cost, idle_cpmd.L3, 4)
#        ]

    cfl_tests = [
        make_cfl_test(generator,  cluster_size, cpus / cluster_size,
                       cedf_schedulable, load_cpmd.mem_hierarchy[cfl_migrations],
                       ohead_cfl, False),
        make_cfl_test(generator,  cluster_size, cpus / cluster_size,
                       cedf_schedulable, load_cpmd.mem_hierarchy[cfl_migrations],
                       ohead_cfl, True),
        # C-EDF (idle)
        make_cfl_test(generator,  cluster_size, cpus / cluster_size,
                       cedf_schedulable, idle_cpmd.mem_hierarchy[cfl_migrations],
                       ohead_cfl, False),
        make_cfl_test(generator,  cluster_size, cpus / cluster_size,
                       cedf_schedulable, idle_cpmd.mem_hierarchy[cfl_migrations],
                       ohead_cfl, True),
        ]

    edfwm_tests = [
        # EDF-WM (load)
        make_edfwm_test(generator, cpus, ohead_edfwm, load_cost, False),
        make_edfwm_test(generator, cpus, ohead_edfwm, load_cost, True),
        # EDF-WM (idle)
        make_edfwm_test(generator, cpus, ohead_edfwm, idle_cost, False),
        make_edfwm_test(generator, cpus, ohead_edfwm, idle_cost, True),
        # EDF-WM (load)
        make_edfwm_test(generator, cpus, ohead_edfwm_wc, load_cost_wc, False),
        make_edfwm_test(generator, cpus, ohead_edfwm_wc, load_cost_wc, True),
        # EDF-WM (idle)
        make_edfwm_test(generator, cpus, ohead_edfwm_wc, idle_cost_wc, False),
        make_edfwm_test(generator, cpus, ohead_edfwm_wc, idle_cost_wc, True),
        ]

    edffm_tests = [
        # EDF-fm (load)
        make_edffm_test(generator, cpus, ohead_edffm, load_cost, False),
        make_edffm_test(generator, cpus, ohead_edffm, load_cost, True),
        # EDF-fm (idle)
        make_edffm_test(generator, cpus, ohead_edffm, idle_cost, False),
        make_edffm_test(generator, cpus, ohead_edffm, idle_cost, True),
        ]

    edfos_tests = [
        # EDF-fm (load)
        make_edfos_test(generator, cpus, ohead_edfos, load_cost, False),
        make_edfos_test(generator, cpus, ohead_edfos, load_cost, True),
        # EDF-os (idle)
        make_edfos_test(generator, cpus, ohead_edfos, idle_cost, False),
        make_edfos_test(generator, cpus, ohead_edfos, idle_cost, True),
        ]

    #if soft:
    #    tests += edffm_tests
    tests = cfl_tests + edfwm_tests + edffm_tests + edfos_tests

    if want_benchmark:
        # for now just a process-local timer
        disable_parallelism()
        tests = [ExecTimer(f) for f in tests]


    ucap_points = list(util_range(1.0, 24.0, ucap_step))

    sample_ucap_range = make_sample_ucap_range(
        tests, samples, ucap_points, **kargs)
    result = []
    for wss in ws_sizes:
        rows = sample_ucap_range(wss)
        result += rows

    if want_benchmark:
        record_benchmark_data(tests)

    return [flatten(r) for r in result]

if __name__ == '__main__':
    import generator.tasksets
    import util
    gen = generator.tasksets.uni_medium
    soft = True

    util.disable_parallelism()

    ws_sizes = [0.0, 500.0, 1000.0, 1500.0, 10000.0]

    data = run(gen, soft, 30, 1, ws_sizes, False,
               feedback=sample.TraceProgress())

    util.print_rows(data, col_sep=',', prepend=' ')
