import random
import sample

from functools import partial

from util import listify, flatten, us2ms, ExecTimer, disable_parallelism

from exp import gen_task_system, grow_and_test

from util.stats import stats
import gedf
import binpack

baruah_test_skipped = 0
benchmark_data = []

def record_benchmark_data(tests):
    global benchmark_data
    benchmark_data = [
        ['TOTAL'] + [t.seconds() for t in tests],
        ['MAX']   + [t.seconds_max() for t in tests],
        ['AVG']   + [t.seconds_avg() for t in tests],
        ['COUNT'] + [t.count for t in tests],
        ]

def make_test(generator, cpus, is_schedulable, **extra_params):
    def generate_and_test(ucap, prng_state):
        ts = gen_task_system(generator, ucap, prng_state)
        ts.make_integral_us()
        return is_schedulable(cpus, ts, **extra_params)
    return generate_and_test

def make_cpp_test(generator, cpus, schedtest_ctor):
    def generate_and_test_cpp(ucap, prng_state):
        ts = gen_task_system(generator, ucap, prng_state)
        ts.make_integral_us()
        fts = ts.as_cpp_taskset()
        return schedtest_ctor(cpus).is_schedulable(fts)
    return generate_and_test_cpp

def cpp_rta(m, min_delta=1000):
    return gedf.cpp.RTAGedf(m, min_delta)

def py_bar(cpus, tasks, threshold=3000.0):
    global baruah_test_skipped
    if gedf.all.should_use_baruah_test(threshold, tasks, cpus):
        return gedf.bar.is_schedulable(cpus, tasks)
    else:
        baruah_test_skipped += 1
        return False

def cpp_bar(cpus, tasks, threshold=3000.0, already_cpp=False):
    global baruah_test_skipped
    fts = None
    if already_cpp:
        fts = tasks
        tasks = fts.original
    if gedf.all.should_use_baruah_test(threshold, tasks, cpus):
        if not fts:
            fts = tasks.as_cpp_taskset()
        return gedf.cpp.BaruahGedf(cpus).is_schedulable(fts)
    else:
        baruah_test_skipped += 1
        return False

def pedf_is_schedulable(cpus, tasks):
    try:
        parts = tasks.partition(cpus, 1, use_density=True)
    except binpack.DidNotFit:
        return False
    # if it fits (i.e, density less than one), then it is schedulale under P-EDF
    return True

def run(generator, cpus,
        umin, umax, epsilon,
        samples, min_delta,
        points=None,
        benchmark=False,
        **kargs):
    if gedf.using_cpp:
        tests = [
            make_test(generator, cpus, gedf.sim.no_counter_example),
            make_test(generator, cpus, pedf_is_schedulable),
            make_test(generator, cpus, gedf.all.is_schedulable_cpp),
            make_cpp_test(generator, cpus, gedf.cpp.GFBGedf),
            make_cpp_test(generator, cpus, gedf.cpp.BakerGedf),
            make_cpp_test(generator, cpus, gedf.cpp.BCLGedf),
            make_cpp_test(generator, cpus, gedf.cpp.BCLIterativeGedf),
            make_test(generator, cpus, cpp_bar),
            make_cpp_test(generator, cpus, cpp_rta),
            make_cpp_test(generator, cpus, gedf.cpp.FFDBFGedf),
            ]
    else:
        tests = [
            make_test(generator, cpus, gedf.sim.no_counter_example),
            make_test(generator, cpus, pedf_is_schedulable),
            make_test(generator, cpus, gedf.all.is_schedulable_py),
            make_test(generator, cpus, gedf.gfb.is_schedulable),
            make_test(generator, cpus, gedf.bak.is_schedulable),
            make_test(generator, cpus, gedf.bcl.is_schedulable),
            make_test(generator, cpus, gedf.bcli.is_schedulable),
            make_test(generator, cpus, py_bar),
            make_test(generator, cpus, gedf.rta.is_schedulable, min_fixpoint_step=1000),
            make_test(generator, cpus, gedf.ffdbf.is_schedulable),
            ]

    if benchmark:
        # for now just a process-local timer
        disable_parallelism()
        tests = [ExecTimer(f) for f in tests]

    if not points:
        ret = sample.range_adaptive([umin], [umax],
                                    sample.close_enough_epsilon(epsilon, min_delta=min_delta),
                                    sample.middle_point_scalar,
                                    lambda u: lambda: [u, random.getstate()],
                                    tests, samples=samples, **kargs)
    else:
        ret = sample.range(points,
                           lambda u: lambda: [u, random.getstate()],
                           tests, samples=samples, **kargs)
    if benchmark:
        record_benchmark_data(tests)
    ret.sort(key=lambda x: x[0])
    return ret

def run_sim_length(generator, cpus,
                   umin, umax, epsilon,
                   samples, min_delta,
                   points=None,
                   **kargs):
    if not gedf.using_cpp:
        print '# This is only implemented for C++!'
        return []

    def sim(secs):
        return partial(gedf.sim.no_counter_example, simulation_length=secs)

    heuristics = [
        make_test(generator, cpus, sim(1)),
        make_test(generator, cpus, sim(5)),
        make_test(generator, cpus, sim(10)),
        make_test(generator, cpus, sim(30)),
        make_test(generator, cpus, sim(60)),
        make_test(generator, cpus, sim(90)),
        make_test(generator, cpus, sim(120)),
        ]
    if not points:
        ret = sample.range_adaptive([umin], [umax],
                                    sample.close_enough_epsilon(epsilon, min_delta=min_delta),
                                    sample.middle_point_scalar,
                                    lambda u: lambda: [u, random.getstate()],
                                    heuristics, samples=samples, **kargs)
    else:
        ret = sample.range(points,
                           lambda u: lambda: [u, random.getstate()],
                           heuristics, samples=samples, **kargs)
    ret.sort(key=lambda x: x[0])
    return ret

def run_first_miss(generator, cpus,
                   umin, umax, epsilon,
                   samples, min_delta,
                   points=None,
                   **kargs):
    if not gedf.using_cpp:
        print '# This is only implemented for C++!'
        return []

    @listify
    def find_first_miss(ucap, prng_state):
        ts = gen_task_system(generator, ucap, prng_state)
        ts.make_integral_us()
        return gedf.sim.first_violation(cpus, ts, simulation_length=300)

    def us2ms_(val):
        if val is None:
            return 0.0
        else:
            return us2ms(val)

    def normalize(results, all, count):
        missed = [x for x in results if x > 0]
        max, avg, med, stdev = stats(missed, want_max=True)
        return [us2ms_(max), us2ms_(med), us2ms_(avg), us2ms_(stdev),
                len(missed) / float(len(results))]

    heuristics = [
        find_first_miss
        ]
    if not points:
        ret = sample.range_adaptive([umin], [umax],
                                    sample.close_enough_epsilon(epsilon, min_delta=min_delta),
                                    sample.middle_point_scalar,
                                    lambda u: lambda: [u, random.getstate()],
                                    heuristics, samples=samples,
                                    init=list,
                                    normalize=normalize, **kargs)
    else:
        ret = sample.range(points,
                           lambda u: lambda: [u, random.getstate()],
                           heuristics, samples=samples,
                           init=list,
                           normalize=normalize, **kargs)
    ret.sort(key=lambda x: x[0])
    ret = [flatten(r) for r in ret]
#    ret = [[(0 if x is None else x) for x in row] for row in ret]
    return ret

def run_grow_and_test(generator, cpus,
                      samples, epsilon,
                      benchmark=False):

    if gedf.using_cpp:
        convert_ts = lambda ts: ts.as_cpp_taskset()
        tests = [
            lambda fts: pedf_is_schedulable(cpus, fts.original),
            lambda fts: gedf.all.is_schedulable_cpp(cpus, fts.original),
            gedf.all.cpp.GFBGedf(cpus).is_schedulable,
            gedf.all.cpp.BakerGedf(cpus).is_schedulable,
            gedf.all.cpp.BCLGedf(cpus).is_schedulable,
            gedf.all.cpp.BCLIterativeGedf(cpus).is_schedulable,
            partial(cpp_bar, cpus, already_cpp=True),
            cpp_rta(cpus).is_schedulable,
            gedf.all.cpp.FFDBFGedf(cpus).is_schedulable,
            ]
    else:
        convert_ts = lambda ts: ts
        tests = [
            partial(pedf_is_schedulable, cpus),
            partial(gedf.all.is_schedulale_py, cpus),
            partial(gedf.gfb.is_schedulable, cpus),
            partial(gedf.bak.is_schedulable, cpus),
            partial(gedf.bcl.is_schedulable, cpus),
            partial(gedf.bcli.is_schedulable, cpus),
            partial(py_bar, cpus),
            partial(gedf.rta.is_schedulable, cpus, min_fixpoint_step=1000),
            partial(gedf.ffdbf.is_schedulable, cpus)
            ]

    def is_unschedulable(ts):
        return ts.utilization() > cpus or gedf.sim.misses_deadline(cpus, ts)

    if benchmark:
        # for now just a process-local timer
        disable_parallelism()
        tests = [ExecTimer(f) for f in tests]
        is_unschedulable = ExecTimer(is_unschedulable)

    bins = grow_and_test(
        samples,
        generator, cpus, is_unschedulable, tests,
        binsize=epsilon, # percentage of total utilization
        make_test_ts=convert_ts
        )

    # prepend utilization "center" of bins
    bins = [[epsilon / 2 + x * epsilon] + bins[x]
            for x in xrange(len(bins))]

    if benchmark:
        record_benchmark_data([is_unschedulable] + tests)

    return bins
