import random
import sample

from overheads import Overheads
from util      import us2ms
from exp       import gen_task_system, util_range

import quanta
import irq
import gedf
import gedf.quantum_based as q_gedf

import hier.generalized_tardiness as gt
import hier.virt_cpu as vt

def gedf_no_overheads_test(cpus, soft, preemptive=True, hard_test=gedf.is_schedulable):
    def pure_theory(generator, u, state):
        ts = gen_task_system(generator, u, state)
        ts.make_integral_us()
        return gedf.soft.is_schedulable(cpus, ts) if soft else hard_test(cpus, ts)
    return pure_theory

def gedf_quantum_centric_irq_test(cpus, soft, oheads=None, preemptive=True,
                                  stagger=False, release_master=False):
    if not oheads:
        oheads = Overheads() # default: zero overheads
    def gedf_qc_exp(generator, u, state):
        ts = gen_task_system(generator, u, state)
        pre  = ts.utilization()
        qlen = us2ms(oheads.quantum_size)
        if not soft:
            quanta.account_for_delayed_release(qlen + us2ms(oheads.release(ts)), ts)
            if stagger:
                quanta.account_for_staggering(qlen, cpus, ts)
        quanta.quantize_period(qlen, ts)
        q_gedf.account_for_overheads(oheads, ts, preemptive, release_master)
        ts.make_integral_us()
        ok = gedf.soft.is_schedulable(cpus, ts) if soft else gedf.is_schedulable(cpus, ts)
#        print 'QC', 'rm', release_master, 'pre', pre, 'm', cpus, 'u', ts.utilization(), ok
        return ok
    return gedf_qc_exp

def gedf_task_centric_irq_test(cpus, soft, oheads=None, preemptive=True,
                               release_master=False,
                               rm_inf_exec=True,
                               rm_multiplexed=True):
    if not oheads:
        oheads = Overheads() # default: zero overheads
    if soft:
        def gedf_soft_tc_exp(generator, u, state):
            tardiness = 0
#            print '>>'
            while True:
                ts = gen_task_system(generator, u, state)
                for t in ts:
                    t.completion = t.deadline + t.cost + tardiness
                gedf.account_for_scheduling_overheads(oheads, ts,
                                                      preemptive)
                irq.global_release_delay(oheads, ts, release_master, rm_inf_exec, rm_multiplexed)
                if not release_master:
                    irq.global_release_tick_task_centric(oheads, cpus, ts)
                else:
                    irq.global_tick_only_task_centric(oheads, cpus, ts)
                if ts.utilization() > cpus:
                    return False
                bound = gedf.soft.task_tardiness_bound(cpus, ts, preemptive)
#                print "testing tardiness=%s, bound=%s, usum=%s" % (tardiness, bound, ts.utilization())
                if bound is None:
                    return False
                if bound <= tardiness:
                    return True
                tardiness = bound
        return gedf_soft_tc_exp
    else:
        def gedf_hard_tc_exp(generator, u, state):
            ts = gen_task_system(generator, u, state)
            pre = ts.utilization()
            gedf.account_for_scheduling_overheads(oheads, ts, preemptive)
            irq.global_release_delay(oheads, ts, release_master, rm_inf_exec, rm_multiplexed)
            if not release_master:
                irq.global_release_tick_task_centric(oheads, cpus, ts)
            else:
                irq.global_tick_only_task_centric(oheads, cpus, ts)
            ts.make_integral_us()
            ok = gedf.is_schedulable(cpus, ts)
#            print 'TC', 'rm', release_master, 'pre', pre, 'm', cpus, 'u', ts.utilization(), ok
            return ok
        return gedf_hard_tc_exp

def gedf_processor_centric_irq_test(cpus, soft, oheads=None, preemptive=True,
                                    release_master=False,
                                    rm_inf_exec=True,
                                    rm_multiplexed=True):
    if not oheads:
        oheads = Overheads() # default: zero overheads
    if soft:
        def gedf_soft_pc_exp(generator, u, state):
            ts = gen_task_system(generator, u, state)
            gedf.account_for_scheduling_overheads(oheads, ts, preemptive)
            irq.global_release_delay(oheads, ts, release_master, rm_inf_exec, rm_multiplexed)
            sf = gt.ServiceFunction.from_overheads(oheads, ts, release_master)
            return gt.has_bounded_tardiness([sf] * cpus, ts)
        return gedf_soft_pc_exp
    else:
        def gedf_hard_pc_exp(generator, u, state):
            ts = gen_task_system(generator, u, state)
            gedf.account_for_scheduling_overheads(oheads, ts, preemptive)
            irq.global_release_delay(oheads, ts, release_master, rm_inf_exec, rm_multiplexed)
            ts.make_integral_us() # switch to us resolution;
            ir = vt.InterruptedResource.from_overheads(oheads, cpus, ts, release_master=release_master)
            ok = ir.is_schedulable(ts)
#            print 'PC', 'rm', release_master, 'ir', ir.utilization(), 'm', cpus, 'u', ts.utilization(), ok
            return ok
        return gedf_hard_pc_exp

def gedf_proc_or_task_centric_irq_test(cpus, soft, oheads=None, preemptive=True, release_master=False):
    task_centric = gedf_task_centric_irq_test(cpus, soft, oheads, preemptive, release_master)
    proc_centric = gedf_processor_centric_irq_test(cpus, soft, oheads, preemptive, release_master)
    def either_test(generator, u, state):
        return task_centric(generator, u, state) \
            or proc_centric(generator, u, state)
    return either_test

# used for JSA 2010 Paper, submitted in 2009 (Accounting for Interrupts...)
def run(generator, soft,
        umin, umax, epsilon,
        cpus, samples, **kargs):
    # use RTSS'09 overheads
    conf    = Overheads.jsa10_soft    if soft else Overheads.jsa10_hard
    conf_rm = Overheads.jsa10_rm_soft if soft else Overheads.jsa10_rm_hard

    ohead_100    = Overheads()
    ohead_20     = Overheads()
    ohead_100_rm = Overheads()
    ohead_20_rm  = Overheads()

    conf(ohead_100)
    conf(ohead_20)
    conf_rm(ohead_100_rm)
    conf_rm(ohead_20_rm)

    ohead_20.scale(0.2)
    ohead_20_rm.scale(0.2)

    tests = [
# ----------------------------------------------------------
        gedf_quantum_centric_irq_test(cpus, soft, ohead_100),
        gedf_task_centric_irq_test(cpus, soft, ohead_100),
        gedf_processor_centric_irq_test(cpus, soft, ohead_100),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_100_rm,
                                   release_master=True, rm_inf_exec=True, rm_multiplexed=False),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_100_rm,
                                   release_master=True, rm_inf_exec=False, rm_multiplexed=False),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_100_rm,
                                   release_master=True, rm_inf_exec=True, rm_multiplexed=True),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_100_rm,
                                   release_master=True, rm_inf_exec=False, rm_multiplexed=True),
# ----------------------------------------------------------
        gedf_quantum_centric_irq_test(cpus, soft, ohead_20),
        gedf_task_centric_irq_test(cpus, soft, ohead_20),
        gedf_processor_centric_irq_test(cpus, soft, ohead_20),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_20_rm,
                                   release_master=True, rm_inf_exec=True, rm_multiplexed=False),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_20_rm,
                                   release_master=True, rm_inf_exec=False, rm_multiplexed=False),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_20_rm,
                                   release_master=True, rm_inf_exec=True, rm_multiplexed=True),
        gedf_task_centric_irq_test(cpus - 1, soft, ohead_20_rm,
                                   release_master=True, rm_inf_exec=False, rm_multiplexed=True)
# ----------------------------------------------------------
        ]
    ret = sample.range_adaptive([umin], [umax],
                                sample.close_enough_epsilon(epsilon, min_delta=0.01),
                                sample.middle_point_scalar,
                                lambda u: lambda: [generator, u,
                                                   random.getstate()],
                                tests, samples=samples, **kargs)
    ret.sort(key=lambda x: x[0])
    return ret
