import random

from itertools import imap


import irq
import gedf
import pfair
import quanta
import binpack
import sample

from overheads import Overheads
from exp   import gen_task_system, util_range
from util      import ms2us, us2ms

def inflate_for_overheads(oheads, cluster_size, tasks, preemptive):
    # scheduling, context switch, cache affinity loss
    gedf.account_for_scheduling_overheads(oheads, tasks, preemptive)
    # IPI delay: the delay on release due to the latency of an IPI
    if cluster_size > 1:
        irq.global_release_delay(oheads, tasks)
    # account for intterrupts after accounting for other overheads
    # tick and release delay
    irq.global_release_tick_task_centric(oheads, cluster_size, tasks)


def make_cedf_test(generator,
                   cluster_size, num_clusters,
                   sched_test,
                   partition_failure,
                   preemptive,
                   oheads):
    """
    generator --- task set generator
    """

    # (ucap, cache_cost) -> schedulable?
    def cedf_test(ucap, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)
        # partition if necessary
        if num_clusters > 1:
            try:
                clusters = ts.partition(num_clusters, cluster_size)
            except binpack.DidNotFit:
                # If we can't partition, then this
                # task set is not schedulable.
                return partition_failure
        else:
            # fake single "cluster"
            clusters = [ts]

        for c in clusters:
            # inflate
            inflate_for_overheads(oheads, cluster_size, c, preemptive)
            # convert to microsecond resolution
            c.make_integral_us()

        # test, but test clusters lazily
        return sched_test(cluster_size, clusters)

    return cedf_test



def make_pfair_test(generator, soft, staggered, no_cpus, oheads):

    # (ucap, cache_cost) -> schedulable?
    def pfair_test(ucap, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)

        if not soft:
            # account for delay of release due to quantization
            quanta.account_for_delayed_release(1, ts)
        pfair.account_for_overheads(oheads, ts)
        if staggered and not soft:
            quanta.account_for_staggering(us2ms(oheads.quantum_size), no_cpus, ts)

        # make sure periods are quantum multiples
        quanta.quantize_period(us2ms(oheads.quantum_size), ts)

        return pfair.is_schedulable(no_cpus, ts)

    return pfair_test


def run(generator, soft,
        umin, umax, epsilon,
        samples, min_delta=0.01,
        points=None,
        **kargs):

    def mk_ohead(f):
        o = Overheads()
        f(o)
        return o

    if soft:
        ohead_pedf = mk_ohead(Overheads.pedf_soft)
        ohead_cedf = mk_ohead(Overheads.cedf_soft)
        ohead_gedf = mk_ohead(Overheads.gedf_soft)
        ohead_pfair = mk_ohead(Overheads.pfair_soft)
        ohead_spfair = mk_ohead(Overheads.spfair_soft)
        ohead_gnpedf = mk_ohead(Overheads.gnpedf_soft)

        def gedf_test(cluster_size, clusters):
            # test, but test clusters lazily
            return  all(imap(lambda c: gedf.soft.is_schedulable(cluster_size, c),
                             clusters))

    else:
        ohead_pedf = mk_ohead(Overheads.pedf_hard)
        ohead_cedf = mk_ohead(Overheads.cedf_hard)
        ohead_gedf = mk_ohead(Overheads.gedf_hard)
        ohead_pfair = mk_ohead(Overheads.pfair_hard)
        ohead_spfair = mk_ohead(Overheads.spfair_hard)
        ohead_gnpedf = None

        def gedf_test(cluster_size, clusters):
            # test, but test clusters lazily
            return  all(imap(lambda c: gedf.is_schedulable(cluster_size, c),
                             clusters))

    tests = [
        # P-EDF
        make_cedf_test(generator, 1, 32, gedf_test, False, True, ohead_pedf),
        # C-EDF
        make_cedf_test(generator, 4,  8, gedf_test, False, True, ohead_cedf),
        # G-EDF
        make_cedf_test(generator, 32, 1, gedf_test, False, True, ohead_gedf),
        # PFAIR
        make_pfair_test(generator, soft, False, 32, ohead_pfair),
        # S-PFAIR
        make_pfair_test(generator, soft, True, 32, ohead_spfair),
        # G-NP-EDF
        make_cedf_test(generator, 32, 1, gedf_test, False, False, ohead_gnpedf)
        ]

    if not soft:
        del tests[-1]

    if not points:
        ret = sample.range_adaptive([umin], [umax],
                                    sample.close_enough_epsilon(epsilon, min_delta=min_delta),
                                    sample.middle_point_scalar,
                                    lambda u: lambda: [u, random.getstate()],
                                    tests, samples=samples, **kargs)
    else:
        ret = sample.range(points,
                           lambda u: lambda: [u, random.getstate()],
                           tests, samples=samples, **kargs)
    ret.sort(key=lambda x: x[0])
    return ret


