import random

from itertools import imap

from overheads import Overheads, const
import irq
import gedf
import binpack
import sample

from util.stats import *

try:
    import cpptest as f
except ImportError:
    pass

from model.tasks import TaskSystem
from exp   import gen_task_system, util_range
from util  import us2ms, sec2us, flatten, storage

def inflate_for_cache_affinity_loss(cache_cost, tasks):
    cache_cost_ms = us2ms(cache_cost)
    for t in tasks:
        t.cost += cache_cost_ms

def inflate_for_overheads(oheads, cluster_size, tasks):
    # scheduling, context switch, cache affinity loss
    gedf.account_for_scheduling_overheads(oheads, tasks)
    # IPI delay: the delay on release due to the latency of an IPI
    if cluster_size > 1:
        irq.global_release_delay(oheads, tasks)
    # account for intterrupts after accounting for other overheads
    # tick and release delay
    irq.global_release_tick_task_centric(oheads, cluster_size, tasks)


def make_cedf_test(generator,
                   cluster_size, num_clusters,
                   sched_test,
                   partition_failure,
                   oheads):
    """
    generator --- task set generator
    """

    # (ucap, cache_cost) -> schedulable?
    def cedf_test(ucap, cache_cost, prng_state):
        # generate
        ts = gen_task_system(generator, ucap, prng_state)
        # partition if necessary
        if num_clusters > 1:
            try:
                clusters = ts.partition(num_clusters, cluster_size)
            except binpack.DidNotFit:
                # If we can't partition, then this
                # task set is not schedulable.
                return partition_failure
        else:
            # fake single "cluster"
            clusters = [ts]

        # setup PMO
        oheads.cache_affinity_loss = const(cache_cost)
        for c in clusters:
            # inflate
            inflate_for_overheads(oheads, cluster_size, c)
            # convert to microsecond resolution
            c.make_integral_us()

        # test, but test clusters lazily
        return sched_test(cluster_size, clusters)

    return cedf_test

def make_cumulative_sched_exp(tests, samples, points, **kargs):
    """
    Cumulative schedulability as a function of cache-affinity-loss overhead.
    tests   --- List of schedulability tests to compare.
    samples --- How many task sets to test per test point?
    point   --- The sampling points (ucap).
    """
    def cumulative_sched_exp(cache_cost):
        # sample
        gen_setup_point = lambda u: lambda: [u, cache_cost, random.getstate()]
        sched_vs_ucap = sample.range(points,
                                     gen_setup_point,
                                     tests, samples=samples, **kargs)
        for row in sched_vs_ucap:
            # add cache cost coordinate
            row.insert(0, cache_cost)
        return sched_vs_ucap

    return cumulative_sched_exp


class SoftResult(object):
    def __init__(self):
        self.bounded = False

    def test(self, cluster_size, clusters, sim_length):
        ok = all([gedf.soft.is_schedulable(cluster_size, c)
                  for c in clusters])
        self.bounded = ok
        if ok:
            if cluster_size == 1:
                self.bound = [0]
                self.stats = []
            else:
                self.bound = [gedf.soft.tardiness_bound(cluster_size, c)
                              for c in clusters]
                # this requires C++ support
                s = [f.edf_observe_tardiness(
                        cluster_size,
                        c.as_cpp_taskset(),
                        sec2us(sim_length))
                     for c in clusters]
                self.stats = [storage(max_tardiness=x.max_tardiness,
                                      total_tardiness=x.total_tardiness,
                                      num_ok_jobs=x.num_ok_jobs,
                                      num_tardy_jobs=x.num_tardy_jobs)
                              for x in s]

# just the test, not the setup
def make_soft_sched_test(sim_length):
    def soft_test(cluster_size, clusters):
        res = SoftResult()
        res.test(cluster_size, clusters, sim_length)
        return [res]
    return soft_test

def combine_soft_results(results):
    n  = len(results)
    ok = [x for x in results if x.bounded]

    bounds = flatten([x.bound for x in ok])
    bavg, bmed, bstd = stats(bounds)

    all_stats = flatten([x.stats for x in ok])
    maxs = [x.max_tardiness for x in all_stats]
    mavg, mmed, mstd = stats(maxs)

    def avg_tardiness(s):
        if s.num_tardy_jobs:
            return s.total_tardiness / float(s.num_tardy_jobs)
        else:
            return 0.0

    avgs = [avg_tardiness(x) for x in all_stats]

    def dl_miss_rate(s):
        total = s.num_tardy_jobs + s.num_ok_jobs
        if total:
            return s.num_tardy_jobs / float(total)
        else:
            return 0

    dls  = [dl_miss_rate(x) for x in all_stats]
    davg, dmed, dstd = stats(dls)

    return [
        len(ok) / float(len(results)), # soft schedulability
        # bound
        safe_max(bounds), # max bound
        bavg,             # avg bound
        bstd,             # std bound
        # observed maximum

        safe_max(maxs),   # max max observed
        mavg,             # avg max observed
        mstd,             # std max observed

        # observed average
        mean(avgs),       # avg avg observed tardiness

        # observed deadline miss ratio
        safe_max(dls),    # max
        davg,             # avg
        dstd,             # std
        ]

def normalize_soft(results, _, __):
    return combine_soft_results(results)

# Number of CPUs (24) hard-coded.
def run(generator, soft,
        samples, ucap_step,
        cache_points,
        simulation_length=60,
        **kargs):
    if soft:
        safe_test = make_soft_sched_test(simulation_length)
        part_fail = [SoftResult()]
        ohead_gedf    = Overheads.from_file('ohead-data/rtss10_G-EDF_avg.csv')
        ohead_cedf_l3 = Overheads.from_file('ohead-data/rtss10_C-EDF-L3_avg.csv')
        ohead_cedf_l2 = Overheads.from_file('ohead-data/rtss10_C-EDF-L2_avg.csv')
        ohead_pedf    = Overheads.from_file('ohead-data/rtss10_P-EDF_avg.csv')

    else:
        def safe_test(cluster_size, clusters):
            # test, but test clusters lazily
            return  all(imap(lambda c: gedf.is_schedulable(cluster_size, c),
                             clusters))
        part_fail = False
        ohead_gedf    = Overheads.from_file('ohead-data/rtss10_G-EDF_wc.csv')
        ohead_cedf_l3 = Overheads.from_file('ohead-data/rtss10_C-EDF-L3_wc.csv')
        ohead_cedf_l2 = Overheads.from_file('ohead-data/rtss10_C-EDF-L2_wc.csv')
        ohead_pedf    = Overheads.from_file('ohead-data/rtss10_P-EDF_wc.csv')

    tests = [
        # G-EDF
        make_cedf_test(generator, 24,  1, safe_test, part_fail, ohead_gedf),
        # C-EDF, L3
        make_cedf_test(generator,  6,  4, safe_test, part_fail, ohead_cedf_l3),
        # C-EDF, L2
        make_cedf_test(generator,  2, 12, safe_test, part_fail, ohead_cedf_l2),
        # P-EDF
        make_cedf_test(generator,  1, 24, safe_test, part_fail, ohead_pedf),
        ]

    if not soft and gedf.using_cpp:
        def no_counter_example(cluster_size, cluster):
            return not f.edf_misses_deadline(cluster_size,
                                             cluster.as_cpp_taskset(),
                                             sec2us(simulation_length))
        def heuristic(cluster_size, clusters):
            # test, but test clusters lazily
            return  all(imap(lambda c: no_counter_example(cluster_size, c),
                             clusters))
        tests += [
            # G-EDF
            make_cedf_test(generator, 24,  1, heuristic,
                           part_fail, ohead_gedf),
            # C-EDF, L3
            make_cedf_test(generator,  6,  4, heuristic,
                           part_fail, ohead_cedf_l3),
            # C-EDF, L2
            make_cedf_test(generator,  2, 12, heuristic,
                           part_fail, ohead_cedf_l2),
            ]

    ucap_points = list(util_range(1.0, 24.0, ucap_step))
    if soft:
        csched_exp = make_cumulative_sched_exp(tests,
                                               samples, ucap_points,
                                               normalize=normalize_soft,
                                               init=list,
                                               **kargs)
    else:
        csched_exp = make_cumulative_sched_exp(tests,
                                               samples, ucap_points,
                                               **kargs)
    result = []
    for cache_cost in cache_points:
        rows = csched_exp(cache_cost)
        result += rows
    return [flatten(r) for r in result]

if __name__ == '__main__':
    import generator.tasksets
    import util
    gen = generator.tasksets.uni_medium
    soft = False

    util.disable_parallelism()

    print 'using cpp:', gedf.using_cpp

    points = [0.0, 500.0, 1000.0, 1500.0, 10000.0]

    data = run(gen, soft, 30, 1, points,
               feedback=sample.TraceProgress())

    util.print_rows(data, col_sep=',', prepend=' ')
