import gfb
import bak
import bar
import bcl
import bcl_iterative as bcli
import rta
import ffdbf

from util.quantor import forall

def should_use_baruah_test(threshold, taskset, no_cpus):
    if not threshold:
        return True
    else:
        slack = no_cpus - taskset.utilization()
        if not slack:
            # can't apply test for zero slack; avoid division by zero
            return False
        n     = len(taskset)
        score = n * (no_cpus * no_cpus) / (slack * slack)
        return score <= threshold


# all (pure Python implementation)
def is_schedulable_py(no_cpus, tasks,
                      baruah_threshold=3000.0,
                      rta_min_step=1000,
                      want_ffdbf=False):
    if tasks.utilization() > no_cpus or not forall(tasks)(lambda t: t.period >= t.cost):
        return False
    else:
        not_arbitrary = tasks.only_constrained_deadlines()
        return ((no_cpus == 1 and tasks.density() <= 1.0) or
                (no_cpus > 1 and
                 bak.is_schedulable(no_cpus, tasks) or
                 (not_arbitrary and
                  ((gfb.is_schedulable(no_cpus, tasks) or
                    bcl.is_schedulable(no_cpus, tasks) or
                    (should_use_baruah_test(baruah_threshold, tasks, no_cpus)
                     and bar.is_schedulable(no_cpus, tasks)) or
                    bcli.is_schedulable(no_cpus, tasks) or
                    rta.is_schedulable(no_cpus, tasks, min_fixpoint_step=rta_min_step) or
                    (want_ffdbf and ffdbf.is_schedulable(no_cpus, tasks))
                    )))))


# Python and C++ schedulablity tests.
# On Topsail, the relative efficiency of these implementations
# is as follows (June 2010):
#
# Tested 156 tasksets.
# Test              Python              C++            Ratio
# GFB       0:00:00.014601   0:00:00.023155             0.63
# BAK       0:00:18.811028   0:00:00.388678            48.40
# BCL       0:00:06.921334   0:00:00.195671            35.37
# BCL-I     0:00:01.762108   0:00:00.069599            25.32
# SKB       0:00:10.818956   0:00:01.914960             5.65
# RTA       0:00:16.463156   0:00:01.766140             9.32
# FF-DBF    1:23:18.865948   0:00:59.787612            83.61
#
# Thus, if the C++ tests are available, then use the C++
# implementation for BAK, BCL, BCL-I, and SKB, but not for GFB,
# where the Python->C++ overhead exceeds the actual performance
# improvement.

# if C++ if available
try:
    import cpptest as cpp

    using_cpp = True

    def is_schedulable_cpp(no_cpus, tasks,
                           baruah_threshold=3000.0,
                           rta_min_step=1000,
                           want_ffdbf=False):
        # Test for overutilization.
        if tasks.utilization() > no_cpus or \
                not forall(tasks)(lambda t: t.period >= t.cost):
            return False
        else:
            # Uniprocessor case
            if no_cpus == 1 and tasks.density() <= 1.0:
                return True
            # Do we have any deadlines past the period?
            not_arbitrary = tasks.only_constrained_deadlines()
            # Density test can handle only constrained deadlines.
            if not_arbitrary and gfb.is_schedulable(no_cpus, tasks):
                return True
            tasks_cpp = tasks.as_cpp_taskset()
            # Baker's test can handle arbitrary periods.
            if cpp.BakerGedf(no_cpus).is_schedulable(tasks_cpp, False):
                return True
            # The rest can only handle constrained deadlines.
            # These we do in C++.
            if not_arbitrary:
                if cpp.BCLGedf(no_cpus).is_schedulable(tasks_cpp, False):
                    return True
                if cpp.BCLIterativeGedf(no_cpus).is_schedulable(tasks_cpp, False):
                    return True
                if should_use_baruah_test(baruah_threshold, tasks, no_cpus) and \
                        cpp.BaruahGedf(no_cpus).is_schedulable(tasks_cpp, False):
                    return True
                if cpp.RTAGedf(no_cpus, rta_min_step).is_schedulable(tasks_cpp, False):
                    return True
                if want_ffdbf and cpp.FFDBFGedf(no_cpus).is_schedulable(tasks_cpp, False):
                    return True
            # Everything failed. Give up.
            return False

except ImportError:
    # Nope, C++ impl. not available. Use Python implementation.
    using_cpp = False
