"""
Generate random task sets for schedulability experiments.

From [CLBDA 06]:

    "We used the uniform, exponential, and bimodal
    distributions to generate task sets as proposed by Baker [3].
    (These distributions allow task-set parameters that are in
    keeping with multiprocessor real-time systems considered in
    our prior work [6].)

        Task periods were uniformly distributed
        over [10; 100] (all units are in ms).

    Task utilizations were distributed differently for each experiment:

        (i)   uniformly, over the range [0:001; 0:1], [0:1; 0:4], or [0:5; 0:9];

        (ii)  exponentially,  with average 0:05 (range [0:001; 0:1]), 0:25 (range
              [0:1; 0:4]), or 0:7 (range [0:5; 0:9]), truncated as needed to
              achieve the desired ranges; or

        (iii) bimodally, distributed uniformly over [0:001; 0:5) with
              probability 8=9, and over [0:5; 0:999] with probability 1/9.

    Task execution costs were
    calculated from periods and utilizations (and may be nonintegral).
    Each task set was created by generating tasks until
    either total utilization was at least four (the number of processors)
    or a predetermined cap on the number of tasks was
    reached, and by then reducing the last task's utilization, if
    necessary, so that total utilization was exactly four. Each experiment
    consisted of 100 different task sets."

"""

import generator.tasks as gen

from random import uniform

# generate(period dist, util distr)
generate = gen.TaskGenerator

def make_period_dist(par):
    if type(par) == tuple:
        return gen.uniform_int(par[0], par[1])
    elif type(par) == list:
        return gen.uniform_choice(par)
    else:
        assert False # can't derive param type

def uniformly(util, period=(10, 100)):
    g = generate(make_period_dist(period),
                 gen.uniform(util[0], util[1]))
    return g.make_task_set

def exponentially(util, uavg,  period=(10, 100)):
    g = generate(make_period_dist(period),
                 gen.exponential(util[0], util[1], uavg,
                                 limiter=gen.truncate))
    return g.make_task_set

def bimodally(util, period=(10, 100)):
    g = generate(
        make_period_dist(period),
        gen.multimodal(
            [(gen.uniform(util[0][0], util[0][1]), util[0][2]),
             (gen.uniform(util[1][0], util[1][1]), util[1][2])]))
    return g.make_task_set

def exponentially_redraw(uavg, period=(10, 100), ulimit=(0.0, 1.0)):
    g = generate(make_period_dist(period),
                 gen.exponential(ulimit[0], ulimit[1], uavg,
                                 limiter=gen.redraw))
    return g.make_task_set


# example: uni_light(cpus, umax, max_tasks)

uni_mixed   = uniformly([0.05, 0.95])

uni_medium_1000  = uniformly((0.1, 0.4), period=(100, 1000))

bimo_light   = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)])
bimo_medium  = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)])
bimo_heavy   = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)])
uni_light    = uniformly([0.001, 0.1])
uni_medium   = uniformly([0.1, 0.4])
uni_heavy    = uniformly([0.5, 0.9])

bimo_light_33  = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)], period=(3, 33))
bimo_medium_33 = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)], period=(3, 33))
bimo_heavy_33  = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)], period=(3, 33))
uni_light_33   = uniformly([0.001, 0.1], period=(3, 33))
uni_medium_33  = uniformly([0.1, 0.4], period=(3, 33))
uni_heavy_33   = uniformly([0.5, 0.9], period=(3, 33))

bimo_light_250   = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)], period=(50, 250))
bimo_medium_250  = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)], period=(50, 250))
bimo_heavy_250   = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)], period=(50, 250))
uni_light_250    = uniformly([0.001, 0.1], period=(50, 250))
uni_medium_250   = uniformly([0.1, 0.4], period=(50, 250))
uni_heavy_250    = uniformly([0.5, 0.9], period=(50, 250))

#### "known good" periods to generate manageable hyperperiods
# These were generated with the constrained that each task has at most 10000
# jobs in a hyperperiod.
#
# periods in [ 10, 100] -> len = 34, lcm = 55440
safe_periods_10_100 = [10, 11, 12, 14, 15, 16, 18, 20, 21, 22, 24, 28, 30, 33,
                       35, 36, 40, 42, 44, 45, 48, 55, 56, 60, 63, 66, 70, 72,
                       77, 80, 84, 88, 90, 99]
# periods in [  3,  33] -> len = 20, lcm = 27720
safe_periods_3_33   = [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 14, 15, 18, 20, 21, 22,
                       24, 28, 30, 33]
# periods in [ 50, 250] -> len = 39, lcm = 332640
safe_periods_50_250 = [54, 55, 56, 60, 63, 66, 70, 72, 77, 80, 84, 88, 90, 96,
                       99, 105, 108, 110, 112, 120, 126, 132, 135, 140, 144,
                       154, 160, 165, 168, 176, 180, 189, 198, 210, 216, 220,
                       224, 231, 240]
# periods in [  1, 2000] -> len = 61, lcm = 7560
safe_periods_1_2000 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 15, 18, 20, 21,
                       24, 27, 28, 30, 35, 36, 40, 42, 45, 54, 56, 60, 63, 70,
                       72, 84, 90, 105, 108, 120, 126, 135, 140, 168, 180, 189,
                       210, 216, 252, 270, 280, 315, 360, 378, 420, 504, 540,
                       630, 756, 840, 945, 1080, 1260, 1512, 1890]

bimo_light_hp   = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)], period=safe_periods_10_100)
bimo_medium_hp  = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)], period=safe_periods_10_100)
bimo_heavy_hp   = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)], period=safe_periods_10_100)
uni_light_hp    = uniformly([0.001, 0.1], period=safe_periods_10_100)
uni_medium_hp   = uniformly([0.1, 0.4], period=safe_periods_10_100)
uni_heavy_hp    = uniformly([0.5, 0.9], period=safe_periods_10_100)

bimo_light_33_hp  = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)], period=safe_periods_3_33)
bimo_medium_33_hp = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)], period=safe_periods_3_33)
bimo_heavy_33_hp  = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)], period=safe_periods_3_33)
uni_light_33_hp   = uniformly([0.001, 0.1], period=safe_periods_3_33)
uni_medium_33_hp  = uniformly([0.1, 0.4], period=safe_periods_3_33)
uni_heavy_33_hp   = uniformly([0.5, 0.9], period=safe_periods_3_33)

bimo_light_250_hp   = bimodally([(0.001, 0.5, 8), (0.5, 0.9, 1)], period=safe_periods_50_250)
bimo_medium_250_hp  = bimodally([(0.001, 0.5, 6), (0.5, 0.9, 3)], period=safe_periods_50_250)
bimo_heavy_250_hp   = bimodally([(0.001, 0.5, 4), (0.5, 0.9, 5)], period=safe_periods_50_250)
uni_light_250_hp    = uniformly([0.001, 0.1], period=safe_periods_50_250)
uni_medium_250_hp   = uniformly([0.1, 0.4], period=safe_periods_50_250)
uni_heavy_250_hp    = uniformly([0.5, 0.9], period=safe_periods_50_250)


# lookup table --- FIXME: this should be structured in some way...
DIST = {'uni_light'   : uni_light,
        'uni_medium'  : uni_medium,
        'uni_heavy'   : uni_heavy,
        'uni_mixed'   : uni_mixed,
        'bimo_light'  : bimo_light,
        'bimo_medium' : bimo_medium,
        'bimo_heavy'  : bimo_heavy,

        'uni_light_33'    : uni_light_33,
        'uni_medium_33'   : uni_medium_33,
        'uni_heavy_33'    : uni_heavy_33,
        'bimo_light_33'   : bimo_light_33,
        'bimo_medium_33'  : bimo_medium_33,
        'bimo_heavy_33'   : bimo_heavy_33,

        'uni_light_250'   : uni_light_250,
        'uni_medium_250'  : uni_medium_250,
        'uni_heavy_250'   : uni_heavy_250,
        'bimo_light_250'  : bimo_light_250,
        'bimo_medium_250' : bimo_medium_250,
        'bimo_heavy_250'  : bimo_heavy_250,

        'uni_light_hp'    : uni_light_hp,
        'uni_medium_hp'   : uni_medium_hp,
        'uni_heavy_hp'    : uni_heavy_hp,
        'bimo_light_hp'   : bimo_light_hp,
        'bimo_medium_hp'  : bimo_medium_hp,
        'bimo_heavy_hp'   : bimo_heavy_hp,

        'uni_light_33_hp'    : uni_light_33_hp,
        'uni_medium_33_hp'   : uni_medium_33_hp,
        'uni_heavy_33_hp'    : uni_heavy_33_hp,
        'bimo_light_33_hp'   : bimo_light_33_hp,
        'bimo_medium_33_hp'  : bimo_medium_33_hp,
        'bimo_heavy_33_hp'   : bimo_heavy_33_hp,

        'uni_light_250_hp'   : uni_light_250_hp,
        'uni_medium_250_hp'  : uni_medium_250_hp,
        'uni_heavy_250_hp'   : uni_heavy_250_hp,
        'bimo_light_250_hp'  : bimo_light_250_hp,
        'bimo_medium_250_hp' : bimo_medium_250_hp,
        'bimo_heavy_250_hp'  : bimo_heavy_250_hp,

        'uni-0.5-0.8'     : uniformly([0.5, 0.8]),
        'uni-0.05-0.8'    : uniformly([0.05, 0.8]),
        'bimo-light-0.8'  : bimodally([(0.001, 0.5, 8), (0.5, 0.8, 1)]),
        'bimo-medium-0.8' : bimodally([(0.001, 0.5, 6), (0.5, 0.8, 3)]),
        'bimo-heavy-0.8'  : bimodally([(0.001, 0.5, 4), (0.5, 0.8, 5)]),

        'uni-0.5-0.7'     : uniformly([0.5, 0.7]),
        'uni-0.05-0.7'    : uniformly([0.05, 0.7]),
        'bimo-light-0.7'  : bimodally([(0.001, 0.5, 8), (0.5, 0.7, 1)]),
        'bimo-medium-0.7' : bimodally([(0.001, 0.5, 6), (0.5, 0.7, 3)]),
        'bimo-heavy-0.7'  : bimodally([(0.001, 0.5, 4), (0.5, 0.7, 5)]),

        'uni-0.5-0.6'     : uniformly([0.5, 0.6]),
        'uni-0.05-0.6'    : uniformly([0.05, 0.6]),
        'bimo-light-0.6'  : bimodally([(0.001, 0.5, 8), (0.5, 0.6, 1)]),
        'bimo-medium-0.6' : bimodally([(0.001, 0.5, 6), (0.5, 0.6, 3)]),
        'bimo-heavy-0.6'  : bimodally([(0.001, 0.5, 4), (0.5, 0.6, 5)]),

        'exp-10-10-100'  : exponentially_redraw(0.10, period=(10, 100)),
        'exp-25-10-100'  : exponentially_redraw(0.25, period=(10, 100)),
        'exp-50-10-100'  : exponentially_redraw(0.50, period=(10, 100)),

        'exp-10-1-2000'  : exponentially_redraw(0.10, period=(1, 2000)),
        'exp-25-1-2000'  : exponentially_redraw(0.25, period=(1, 2000)),
        'exp-50-1-2000'  : exponentially_redraw(0.50, period=(1, 2000)),

        'exp-10-3-33'    : exponentially_redraw(0.10, period=(3, 33)),
        'exp-25-3-33'    : exponentially_redraw(0.25, period=(3, 33)),
        'exp-50-3-33'    : exponentially_redraw(0.50, period=(3, 33)),

        'exp-10-50-250'  : exponentially_redraw(0.10, period=(50, 250)),
        'exp-25-50-250'  : exponentially_redraw(0.25, period=(50, 250)),
        'exp-50-50-250'  : exponentially_redraw(0.50, period=(50, 250)),

        'exp-10-10-100_hp'  : exponentially_redraw(0.10, period=safe_periods_10_100),
        'exp-25-10-100_hp'  : exponentially_redraw(0.25, period=safe_periods_10_100),
        'exp-50-10-100_hp'  : exponentially_redraw(0.50, period=safe_periods_10_100),

        'exp-10-1-2000_hp'  : exponentially_redraw(0.10, period=safe_periods_1_2000),
        'exp-25-1-2000_hp'  : exponentially_redraw(0.25, period=safe_periods_1_2000),
        'exp-50-1-2000_hp'  : exponentially_redraw(0.50, period=safe_periods_1_2000),

        'exp-10-3-33_hp'  : exponentially_redraw(0.10, period=safe_periods_3_33),
        'exp-25-3-33_hp'  : exponentially_redraw(0.25, period=safe_periods_3_33),
        'exp-50-3-33_hp'  : exponentially_redraw(0.50, period=safe_periods_3_33),

        'exp-10-50-250_hp'  : exponentially_redraw(0.10, period=safe_periods_50_250),
        'exp-25-50-250_hp'  : exponentially_redraw(0.25, period=safe_periods_50_250),
        'exp-50-50-250_hp'  : exponentially_redraw(0.50, period=safe_periods_50_250),
        }


# after the fact modification

def implicit_deadlines(ts):
    for t in ts:
        t.deadline = t.period

def uniform_slack_based(min_slack, max_slack):
    def tweak_slack(ts):
        for t in ts:
            slack = t.period - t.cost
            earliest = slack * min_slack
            latest   = slack * max_slack
            t.deadline = t.cost + uniform(earliest, latest)
    return tweak_slack

CONSTRAINED_DEADLINE_DIST = {
    'implicit'      : implicit_deadlines,

    'uni-5-95'    : uniform_slack_based(0.05, 0.95),
    'uni-0-100'  : uniform_slack_based(0.0, 1.0),

    'uni-25-75'   : uniform_slack_based(0.25, 0.75),

    'uni-45-100'  : uniform_slack_based(0.45, 1.0),
    'uni-55-100'  : uniform_slack_based(0.55, 1.0),
    'uni-65-100'  : uniform_slack_based(0.65, 1.0),
    'uni-75-100'  : uniform_slack_based(0.75, 1.0),
    'uni-85-100'  : uniform_slack_based(0.85, 1.0),
}


ARBITRARY_DEADLINE_DIST = {
    'slack_75-125'  : uniform_slack_based(0.75, 1.25),

    'slack_100-200' : uniform_slack_based(1.0, 2.0),
}

DEADLINE_DIST = {}
DEADLINE_DIST.update(CONSTRAINED_DEADLINE_DIST)
DEADLINE_DIST.update(ARBITRARY_DEADLINE_DIST)

def generate_with_deadlines(make_taskset, change_deadlines):
    def _gen_dl(umax, tmax):
        ts = make_taskset(umax, tmax)
        change_deadlines(ts)
        return ts
    return _gen_dl

