"""G-EDF HRT in clusters of virtual CPUs

Based on:

        "Hierarchical Scheduling Framework for Virtual Clustering
         of Multiprocessors"
        Insik Shin, Avrind Easwaran, Insup Lee
        ECRTS'08

Warning: This test fundamentally assumes integral time. Make sure all task parameters
         are integral!
"""

from util import us2ms, sumn, uniq, imerge
from math import ceil, floor
from util.quantor import forall
from irq  import InterruptDBF

def N(tsk, t):
    "Number of full jobs in workload. Equation (3)."
    return int(floor( (t + tsk.period  - tsk.deadline) / float(tsk.period)))

def bf_points_of_change(f, tsk, max_t):
    vals = [f(tsk, t) for t in xrange(max_t + 1)]
    for t in xrange(1, max_t + 1):
        if vals[t - 1] != vals[t]:
            yield (t, vals[t -1], vals[t])

def bf_2points_of_change(f, tsk1, tsk2, max_t):
    vals = [f(tsk1, tsk2, t) for t in xrange(max_t + 1)]
    for t in xrange(1, max_t + 1):
        if vals[t - 1] != vals[t]:
            yield (t, vals[t -1], vals[t])

def N_points_of_change(tsk, max_t):
    yield 0
    t = tsk.deadline
    while t < max_t:
        yield t
        t += tsk.period

def CI(tsk, t):
    "Carry-in job in workload. Equation (3)."
    return min(tsk.cost, max(0, t - N(tsk,t) * tsk.period))

def CI_points_of_change(tsk, max_t):
    c = tsk.cost
    for p in N_points_of_change(tsk, max_t):
        to = min(p + c + 1, max_t)
        while p < to:
            yield p
            p += 1

def W_(tsk, t):
    return N(tsk, t) * tsk.cost + CI(tsk, t)

def W(tsk, t):
    "Worload. Equation (3)."
    n = N(tsk, t)
    tmp = max(0, t - n * tsk.period)
    return n * tsk.cost + min(tsk.cost, tmp)

def W_points_of_change(tsk, max_t):
    c = tsk.cost
    yield 0
    for p in N_points_of_change(tsk, max_t):
        p += 1
        to = min(p + c, max_t)
        while p < to:
            yield p
            p += 1

def I2bar(tsk_i, tsk_k, a_k):
    t = a_k + tsk_k.deadline
    if tsk_i != tsk_k:
        return min(W(tsk_i, t), t - tsk_k.cost)
    else:
        return min(W(tsk_k, t) - tsk_k.cost, a_k)

def I2bar_points_of_change(tsk_i, tsk_k, max_t):
    yield 0
    for t in W_points_of_change(tsk_i, max_t + tsk_k.deadline):
        t2 = t - tsk_k.deadline
        if t2 > 0:
            yield t2

def I2hat(tsk_i, tsk_k, a_k):
    t = a_k + tsk_k.deadline
    if tsk_i != tsk_k:
        return min(N(tsk_i, t) * tsk_i.cost, t - tsk_k.cost)
    else:
        return min(N(tsk_k, t) * tsk_k.cost - tsk_k.cost, a_k)

def I2hat_points_of_change(tsk_i, tsk_k, max_t):
    yield 0
    for t in N_points_of_change(tsk_i, max_t + tsk_k.deadline):
        t2 = t - tsk_k.deadline
        if t2 > 0:
            yield t2

def I2diff(tsk_i, tsk_k, a_k):
    return I2bar(tsk_i, tsk_k, a_k) - I2hat(tsk_i, tsk_k, a_k)

def dem(all_tasks, tsk_k, a_k, m):
    return sum([I2hat(tsk_i, tsk_k, a_k) for tsk_i in all_tasks]) + \
        sumn([I2diff(tsk_i, tsk_k, a_k) for tsk_i in all_tasks], m - 1) + \
        m * tsk_k.cost

class MultiProcessorResource(object):
    "MPR service bound function"
    def __init__(self, maxpar=1, period=1, service=None):
        self.maxpar  = maxpar
        self.period  = period
        if service is None:
            service = period * maxpar
        self.service = service
        self.burst_length = int(ceil(self.service / float(self.maxpar)))
        self.max_zero_service_interval = 2 * (self.period - self.burst_length)

    def __str__(self):
        return "MPR(maxpar=%s, period=%s, service=%s)" % \
            (self.maxpar, self.period, self.service)

    def __call__(self, t):
        toff = t - (self.period - int(ceil(self.service / float(self.maxpar))))
        if toff < 0:
            return 0
        k = int(floor(toff / float(self.period)))
        return k * self.service + \
            max(0,
                (toff - (1 + k) * self.period) * self.maxpar + self.service)

    def utilization(self):
        return self.service / float(self.period)

    def points_of_change(self, max_t):
        yield 0
        t = self.max_zero_service_interval + 1
        while t < max_t:
            b = min(t + self.burst_length + 1, max_t)
            t2 = t
            while t2 < b:
                yield t2
                t2 += 1
            t += self.period

    def ak_of_change(self, tsk_k, max_t):
        yield 0
        for t in self.points_of_change(max_t + tsk_k.deadline):
            t -= tsk_k.deadline
            if t > 0:
                yield t

    def task_schedulable_for_offset(self, all_tasks, tsk_k, a_k):
        return self.dem(all_tasks, tsk_k, a_k) <= self(tsk_k.deadline + a_k)

    def dem(self, all_tasks, tsk_k, a_k):
        return dem(all_tasks, tsk_k, a_k, self.maxpar)

    def task_schedulable(self, all_tasks, tsk_k, max_ak):
        """Try to be clever: just check at poins of change.
        This doesn't really work in practice since for every a_k some curve 
        changes (for most taks sets).
        """
        max_t = max_ak + tsk_k.deadline
        hats = [I2hat_points_of_change(tsk_i, tsk_k, max_t) for tsk_i in all_tasks]
        bars = [I2bar_points_of_change(tsk_i, tsk_k, max_t) for tsk_i in all_tasks]
        sbfs = [self.ak_of_change(tsk_k, max_t)]
        aks_to_test = uniq(imerge(lambda x,y: x < y, *(hats + bars + sbfs)))
        for a_k in aks_to_test:
            if not self.task_schedulable_for_offset(all_tasks, tsk_k, a_k):
                print a_k
                return False
        return True

    def task_schedulable2(self, all_tasks, tsk_k, max_ak):
        "Try not to be clever. Just check every a_k."
        for a_k in xrange(int(ceil(max_ak))):
            if not self.task_schedulable_for_offset(all_tasks, tsk_k, a_k):
                print a_k
                return False
        return True

    def task_schedulable3(self, all_tasks, tsk_k, max_ak):
        """Start at the end; then jump backwards, thereby skipping more.
        This technique is due to Hennadiy.
        """
        a_k = int(floor(max_ak))
        if a_k < 0:
#            print 'no test required: start a_k =>', a_k 
            return True
        while a_k > 0:
            need = self.dem(all_tasks, tsk_k, a_k)
            have = self(tsk_k.deadline + a_k)
            slack  = have - need
            no_per = slack / self.service
 #           print 'a_k=%s, need=%s, have=%s, slack=%s, no_per=%s' % \
 #               (a_k, need, have, slack, no_per)
            if slack < 0:
                return False
            if slack <= self.service:
                a_k -= max(1, int(ceil(slack / self.maxpar)))
            else:
                a_k -= self.period * int(floor(no_per))
  #      print 'a_k => %s' % a_k
        return True

    def ak_bounds(self, all_tasks):
        m  = self.maxpar
        c_sigma = sumn([t.cost for t in all_tasks], m - 1)
        Ut = all_tasks.utilization()
        U  = sum([(t.period - t.cost) * t.utilization() for t in all_tasks])
        B  = self.service * 2 * (1 - (self.service / (float(m) * self.period)))
        slack = self.service / float(self.period) - Ut
        def numerator(tsk_k):
            return c_sigma + m * tsk_k.cost - tsk_k.deadline * slack + U + B
        for tsk_k in all_tasks:
            yield numerator(tsk_k) / slack

    def is_schedulable(self, tasks):
        if tasks.utilization() > self.utilization() or \
                not forall(tasks)(lambda t: t.period >= t.cost):
            return False
        for (tsk_k, max_ak) in zip(tasks, self.ak_bounds(tasks)):
            if not self.task_schedulable3(tasks, tsk_k, max_ak):
                return False
        return True

    @staticmethod
    def from_overheads(overheads, no_cpus, tasks, period=None):
        "MPR from interrupts."
        n          = len(tasks)
        rel_charge = int(ceil(overheads.release(tasks)))
        if period is None:
            period     = tasks.max_period() 
        service    = period
        for t in tasks:
            maxrel = int(ceil(period / float(t.period)))
            service -= maxrel * rel_charge
        no_ticks = ceil(period / float(overheads.quantum_size))
        service -= int(ceil(no_ticks * overheads.tick(tasks)))
        print rel_charge, period, service, (period - service) / float(period)
        return MultiProcessorResource(no_cpus, period, service * no_cpus)

class InterruptedResource(MultiProcessorResource):
    "Like a MPR, but more accurate SBF for interrupts."
    def __init__(self, maxpar, period, interrupts=None):
        if interrupts is None:
            interrupts = []
        self.interrupts = interrupts
        estimate = period
        for irq in interrupts:
            estimate -= irq(period)
        MultiProcessorResource.__init__(self, maxpar, period,
                                        maxpar * max(0, estimate))

    def __call__(self, t):
        "Pessimistic model: all interrupts on all procs."
        service = t
        for irq in self.interrupts:
            service -= irq(t)
        return max(0, service) * self.maxpar

    @staticmethod
    def from_overheads(overheads, no_cpus, tasks, period=None, release_master=False):
        "MPR from interrupts."
        irqs = []
        if not release_master:
            c = int(ceil(overheads.release(tasks)))
            irqs += [InterruptDBF(c, t.period) for t in tasks]
        c = int(ceil(overheads.tick(tasks)))
        irqs += [InterruptDBF(c, overheads.quantum_size) for _ in xrange(no_cpus)]
        if period is None:
            period = tasks.max_period()
        return InterruptedResource(no_cpus, period, irqs)
