#!/usr/bin/env python

class Variable(object):
    def __init__(self, name, prog):
        self.prog = prog
        self.name = name
        self.constraints = []
        self.integer = False
        self.min = None # unbounded
        self.max = None

def format_term(coeff, var):
    if coeff < 0:
        return "- %s %s" % (abs(coeff), var.name)
    else:
        return "+ %s %s" % (coeff, var.name)

def format_terms(terms):
    "format expression of terms"
    return " ".join([format_term(c, v) for (c,v) in terms])

class Constraint(object):
    def __init__(self, prog, bound=0, cmp='<=', name=None):
        self.prog  = prog
        self.name  = name
        self.bound = bound
        self.cmp   = cmp
        self.terms = []

    def add(self, coeff, var):
        v = self.prog.get_var(var)
        v.constraints.append(self)
        self.terms.append((coeff, v))

    def format_lp(self):
        assert self.terms # must have at least one term
        terms = format_terms(self.terms)
        return "%s %s %s" % (terms, self.cmp, self.bound)

class LinearProgram(object):
    def __init__(self, name='Z', goal='maximize'):
        self.vars = {}
        self.v_ord  = []
        self.constraints = {}
        self.c_ord  = []
        self.target = []
        self.name  = name
        self.goal = goal

    def has_integer_vars(self):
        for v in self.v_ord:
            if v.integer:
                return True
        return False

    def add(self, coeff, var):
        self.target.append((coeff, self.V(var)))

    def get_var(self, name=None):
        if name is None:
            name = "X%d" % len(self.vars)
        if not name in self.vars:
            self.vars[name] = Variable(name, self)
            self.v_ord.append(self.vars[name])
        return self.vars[name]

    def get_constraint(self, name=None):
        if name is None:
            name = "C%d" % len(self.constraints)
        if not name in self.constraints:
            self.constraints[name] = Constraint(self, name=name)
            self.c_ord.append(self.constraints[name])
        return self.constraints[name]

    V = get_var
    C = get_constraint


    def format_lp(self):
        prog =  "%s\n" % self.goal
        prog += "  %s : %s\n\n" % (self.name, format_terms(self.target))
        prog += "subject to\n"
        for c in self.c_ord:
            prog += "  %10s : %s \n" % (c.name, c.format_lp())
        prog += "\nbounds\n"
        for v in self.v_ord:
            if v.min is None and v.max is None:
                prog += "  %s free\n" % v.name
            else:
                prog += "  %s <= %s <= %s\n" %\
                    ("-inf" if v.min is None else str(v.min),
                     v.name,
                     "+inf" if v.max is None else str(v.max))
        prog += "\n"
        intvars = [v for v in self.v_ord if v.integer]
        if intvars:
            prog += "integer\n"
            for v in intvars:
                prog += "  %s\n" % v.name
            prog += "\n"
        prog += "end\n"
        return prog

def test2():
    p = LinearProgram()

    p.add(10.9, 'x1')
    p.add(-3.2, 'x2')

    c = p.C()
    c.bound = 100
    c.cmp = '<'
    c.add(3, 'x1')

    c = p.C()
    c.bound = 33
    c.cmp = '>'
    c.add(1.9, 'x2')
    c.add(-0.33, 'x1')

    c = p.C()
    c.add(2, 'x1')
    c.add(2, 'x2')
    c.add(-1, 'x3')

    v = p.V('x1')
    v.integer = True
    v.max = 1000

    p.V('x2').integer = True

    v = p.V('x3')
    v.min = 0
    v.max = 10
    v.integer = True

    return p

