import copy

from math   import floor, ceil
from util   import gcd, lcm, is_multiple, ms2us, us2ms, maybe_int
from util.quantor import forall

from fractions import Fraction

import binpack

class Task(object):
    def __init__(self, cost, period, deadline=None, phase=0, id=None,
                 no_cast=False):
        if deadline is None:
            # implicit deadline by default
            deadline = period
        cast = (lambda x: x) if no_cast else float
        self.id         = id
        self.phase      = cast(phase)
        self.period     = cast(period)
        self.cost       = cast(cost)
        self.deadline   = cast(deadline)
        self.completion = self.deadline

    def all_parameters_integral(self, check_type=False):
        if check_type:
            type_ok  = lambda x: type(x) != float and type(x) != complex
            integral = lambda x: is_multiple(x, 1) and type_ok(x)
        else:
            integral = lambda x: is_multiple(x, 1)
        return integral(self.cost) and integral(self.period) and \
            integral(self.phase) and integral(self.deadline)

    def implicit_deadline(self):
        return self.deadline == self.period

    def constrained_deadline(self):
        return self.deadline <= self.period

    def make_integral_us(self):
        "scale ms params into us int params"
        self.phase      = int(ceil(ms2us(self.phase)))
        self.period     = int(ceil(ms2us(self.period)))
        self.cost       = int(ceil(ms2us(self.cost)))
        self.deadline   = int(ceil(ms2us(self.deadline)))
        self.completion = int(ceil(ms2us(self.completion)))

    def make_integral(self):
        "get integer-ceilings as parameters"
        self.phase      = int(ceil(self.phase))
        self.period     = int(ceil(self.period))
        self.cost       = int(ceil(self.cost))
        self.deadline   = int(ceil(self.deadline))
        self.completion = int(ceil(self.completion))

    def round_to_ms(self):
        "scale us into ms int params"
        self.phase      = int(ceil(us2ms(self.phase)))
        self.period     = int(ceil(us2ms(self.period)))
        self.cost       = int(ceil(us2ms(self.cost)))
        self.deadline   = int(ceil(us2ms(self.deadline)))
        self.completion = int(ceil(us2ms(self.completion)))

    def utilization(self):
        return self.cost / float(self.period)

    def utilization_q(self):
        return Fraction(self.cost, self.period)

    def density(self):
        return self.cost / float(min(self.period, self.deadline))

    def density_q(self):
        return Fraction(self.cost, min(self.period, self.deadline))

    def maxjobs(self, time):
        return int(ceil((time + self.completion) / self.period))

    def __str__(self):
        idstr = ", id=%s" % self.id if self.id is not None else ""
        pstr  = ", phase=%s" % self.phase if self.phase != 0.0 else ""
        dstr  = ", deadline=%s" % self.deadline if self.deadline != self.period else ""

        return "Task(%s, %s%s%s%s)" % (self.cost, self.period, dstr, pstr, idstr)


class TaskSystem(list):
    def __init__(self, tasks=[]):
        self.extend(tasks)
        self.make_integral_us_called = False

    def __str__(self):
        return "\n".join([str(t) for t in self])

    def forall(self, pred):
        return forall(self)(pred)

    def all_parameters_integral(self):
        return self.forall(lambda t: t.all_parameters_integral())

    def only_implicit_deadlines(self):
        return self.forall(lambda t: t.implicit_deadline())

    def only_constrained_deadlines(self):
        return self.forall(lambda t: t.constrained_deadline())

    def higher_prio_tasks(self, tsk):
        """Assumes static priority assignment.
           Returns iterator over tasks with higher priority than tsk in self.
        """
        assert tsk in self
        for t in self:
            if t != tsk:
                yield t
            else:
                break

    def lower_prio_tasks(self, tsk):
        """Assumes static priority assignment.
           Returns iterator over tasks with lower priority than tsk in self.
        """
        assert tsk in self
        skip = True
        for t in self:
            if not skip:
                yield t
            if t == tsk:
                skip = False

    def name_tasks(self):
        for i, t in enumerate(self):
            t.id = i

    def make_integral_us(self):
        # doesn't make much sense to call it twice
        assert not self.make_integral_us_called
        self.make_integral_us_called = True
        for t in self:
            t.make_integral_us()

    def make_integral(self):
        for t in self:
            t.make_integral()

    def round_to_ms(self):
        for t in self:
            t.round_to_ms()

    def sort_by_period(self):
        self.sort(key=lambda t: t.period)

    def index_by_period(self):
        self.sort_by_period()
        self.name_tasks()

    def utilization(self, heaviest=None):
        u = [t.utilization() for t in self]
        if heaviest is None:
            return sum(u)
        else:
            u.sort(reverse=True)
            return sum(u[:heaviest])

    def utilization_q(self, heaviest=None):
        u = [t.utilization_q() for t in self]
        if heaviest is None:
            return sum(u)
        else:
            u.sort(reverse=True)
            return sum(u[:heaviest])

    def density(self):
        return sum([t.density() for t in self])

    def density_q(self):
        return sum([t.density_q() for t in self])

    def hyperperiod(self):
        return lcm(*[maybe_int(t.period) for t in self])

    def max_utilization(self):
        return max([t.utilization() for t in self])

    def max_density(self):
        return max([t.density() for t in self])

    def max_density_q(self):
        return max([t.density_q() for t in self])

    def max_cost(self):
        return max([t.cost for t in self])

    def max_phase(self):
        return max([t.phase for t in self])

    def max_period(self):
        return max([t.period for t in self])

    def min_deadline(self):
        return min([t.deadline for t in self])

    def partition(self, no_clusters, cluster_size, use_density=False):
        weight = Task.density if use_density else Task.utilization
        parts = binpack.worst_fit_decreasing(self, no_clusters,
                                             capacity=cluster_size,
                                             weight=weight,
                                             misfit=binpack.report_failure,
                                             empty_bin=TaskSystem)
        self.partitions = parts
        for p in self.partitions:
            for t in p:
                t.partition = p
        return self.partitions

    def copy(self):
        ts = TaskSystem([copy.copy(t) for t in self])
        return ts

    def as_cpp_taskset(self):
        # requires fast module, and all parameters must be integral
        fts = make_cpp_taskset(self)
        fts.original = self
        return fts

try:
    import cpptest

    def make_cpp_taskset(ts):
        "convert integral TaskSystem ts to C++ TaskSet"
        fts = cpptest.TaskSet(len(ts))
        for i in xrange(len(ts)):
            fts.set_wcet(i, ts[i].cost)
            fts.set_period(i, ts[i].period)
            if ts[i].deadline != ts[i].period:
                # only set if not implicit
                fts.set_deadline(i, ts[i].deadline)
        return fts

except ImportError:
    # no conversion available
    def make_cpp_taskset(_):
        raise ImportError, "C++ implementation not available!"
