import math
import csv
import types

from util import print_rows, print_row, load_column_csv, storage

# all overheads in microseconds

def delistify(x):
    if isinstance(x, types.ListType):
        x = len(x)
    return float(x)

class LinearEqu(object):
    def __init__(self, a, b, coerce=lambda x: x):
        self.a = a
        self.b = b
        self.coerce = coerce

    def __str__(self):
        slope = round(self.b, 3)
        if abs(slope) >= 0.001:
            return '%.3f%+.3f n' % (self.a, slope)
        else:
            return '%.3f' % self.a

    def __call__(self, x):
        return self.a + (self.b * self.coerce(x) if self.b else 0)

    def __add__(self, other):
        return LinearEqu(self.a + other.a, self.b + other.b, self.coerce)

    def __mul__(self, scalar):
        return LinearEqu(self.a * scalar, self.b * scalar, self.coerce)

    def __rmul__(self, scalar):
        return self * scalar

    def is_constant(self):
        return self.b == 0

class PieceWiseLinearEqu(object):
    def __init__(self, points, coerce=lambda x: x,
                 force_monotonicly_increasing=False):
        # points = [(x1, y1), (x2, y2), ...]
        assert len(points) >= 2

        def slope(i):
            dy = points[i+1][1] - points[i][1]
            dx = points[i+1][0] - points[i][0]
            return dy / dx

        def yintercept(i):
            x, y = points[i]
            dy = slope(i) * x
            return y - dy

        self.segments = [(points[i+1][0],
                          LinearEqu(yintercept(i), slope(i),
                                     coerce))
                         for i in xrange(len(points) - 1)]

    def __call__(self, x):
        # find piece (slow and stupid)
        for (limit, f) in self.segments:
            if limit > x:
                break
        # approximate linearly from support point
        return f(x)

    def is_constant(self):
        return all([seg[1].is_constant() for seg in self.segments])

def const(x):
    return LinearEqu(x, 0)

def lin(a, b):
    return LinearEqu(a, b, coerce=delistify)

def scale(alpha, fun):
    return lambda x: fun(x) * alpha

def piece_wise_linear(points, coerce=delistify):
    return PieceWiseLinearEqu(points, coerce)

def make_monotonic(points):
    filtered = points[:1]
    prevprev = None
    _, prev = filtered[0]
    for (x, y) in points[1:]:
        # y values should not decrease
        y = max(prev, y)
        if not prevprev is None and prevprev == y:
            # remove useless intermediate point
            filtered.pop()
        filtered.append((x,y))
        prevprev, prev = prev, y

    # also remove the last one if it is not needed (i.e., constant)
    if len(filtered) == 2 and filtered[-1][1] == filtered[-2][1]:
        filtered.pop()

    return filtered

def is_monotonic(points):
    x1, y1 = points[0]
    for (x2, y2) in points[1:]:
        assert x1 < x2
        if y1 > y2:
            return False
        x1, y1 = x2, y2
    return True

def monotonic_pwlin(points, coerce=delistify):
    ascending = make_monotonic(points)
    if len(ascending) > 1:
        assert is_monotonic(ascending)
        return piece_wise_linear(ascending, coerce)
    elif ascending:
        return const(ascending[0][1])
    else:
        return const(0)


class Overheads(object):
    """Legacy overhead objects"""
    def __init__(self):
        self.zero_overheads()

    FIELD_MAPPING = [
        ('SEND_RESCHED', 'ipi_delay'),
        ('SCHED',        'sched'),
        ('SCHED2',       'sched2'),
        ('RELEASE',      'release'),
        ('CXS',          'ctxsw'),
        ('TICK',         'tick'),
        ('PMO',          'cache_affinity_loss'),
        ('PULL_TIMER',   'pull_timer'),
        ]

    def zero_overheads(self):
        self.quantum_size = 1000 # microseconds

        self.cache_affinity_loss = const(0) # cache-related preemption/migration delay

        self.tick         = const(0)

        self.ctxsw        = const(0)
        self.sched        = const(0)
        self.sched2       = const(0) # rearming timers
        self.release      = const(0)
        self.ipi_delay    = const(0) # new for RTSS'09, not measured in older work
        self.pull_timer   = const(0) # new for semi-part paper

    def zero_non_irq_overheads(self):
        self.cache_affinity_loss = const(0)
        self.ctxsw               = const(0)
        #self.release            = non-zero
        self.sched2              = const(0)
        self.sched               = const(0)
        #self.ipi_delay          = non-zero
        #self.tick               = non-zero

    def __str__(self):
        return " ".join(["%s: %s" % (name, self.__dict__[field])
                         for (name, field) in Overheads.FIELD_MAPPING])

    def scale(self, factor):
        self.cache_afinity_loss = scale(factor, self.cache_affinity_loss)
        self.tick               = scale(factor, self.tick)
        self.ctxsw              = scale(factor, self.ctxsw)
        self.sched              = scale(factor, self.sched)
        self.sched2             = scale(factor, self.sched2)
        self.release            = scale(factor, self.release)
        self.ipi_delay          = scale(factor, self.ipi_delay)

    def get_profile(self, points):
        return [[x] + [self.__dict__[field](x)
                 for (_, field) in Overheads.FIELD_MAPPING]
                for x in points]

    def write_profile(self, fname, points, want_header=True,
                      comment_header=False):
        f = open(fname, 'w')
        if want_header:
            print_row(['NUM_TASKS'] +
                      [name for (name,_) in Overheads.FIELD_MAPPING],
                      col_sep=' ,',
                      f=f,
                      prepend='#' if comment_header else ' ',
                      col_width=20)
        table = self.get_profile(points)
        print_rows(table, f=f, col_sep=',', prepend=' ', col_width=20)
        f.close()

    @staticmethod
    def from_file(fname, non_decreasing=True):
        # no error checking---crash loudly if something is amiss
        f = open(fname)
        d = list(csv.reader(f))
        f.close()
        col_idx = {}
        for i, key in enumerate(d[0]):
            col_idx[key.strip()] = i
        data    = d[1:]

        def col(i):
            for row in data:
                if row:
                    yield float(row[i])

        if not 'NUM_TASKS' in col_idx:
            raise IOError, "no NUM_TASKS?"

        o = Overheads()

        for (name, field) in Overheads.FIELD_MAPPING:
            if name in col_idx:
                points = zip(col(col_idx['NUM_TASKS']), col(col_idx[name]))
                if non_decreasing:
                    o.__dict__[field] = monotonic_pwlin(points)
                else:
                    o.__dict__[field] = piece_wise_linear(points)
        return o

    # ##################################################################
    # Old named profiles.
    #
    # This method has been deprecated --- don't add any new profiles!
    # Use from_file() instead (see exp/rtss10.py for an example).

    def pfair_hard(self):
        self.cache_affinity_loss = const(681.11)
        self.tick    = lin(11.2706107333, 0.301054445333)
        self.sched   = const(30.893333)
        self.ctxsw   = lin(3.10683296667, 0.00993000066667)
        self.sched2  = lin(0.8263334, 0.000234445333333)
        self.release = const(0)

    def spfair_hard(self):
        self.cache_affinity_loss = const(104.13)
        self.tick    = lin(4.83822188333, 0.320902222333)
        self.sched   = const(42.366667)
        self.ctxsw   = lin(3.219, 0.00324666666667)
        self.sched2  = lin(0.81550015, 0.000209999666667)
        self.release = const(0)

    def gedf_hard(self):
        self.cache_affinity_loss = const(375.44)
        self.tick    = lin(3.04333333333, 0.00270666666667)
        self.sched   = lin(53.78849975, 0.248632221667)
        self.ctxsw   = const(29.233333)
        self.sched2  = lin(1.495555, 0.01393889)
        self.release = lin(45.02516645, 0.313565557)

    def cedf_hard(self):
        self.cache_affinity_loss = const(171.62)
        self.tick    = const(3.23)
        self.sched   =  lin(10.4011112, 0.011044444)
        self.ctxsw   = const(6.076667)
        self.sched2  = const(4.356667)
        self.release = const(30.346667)

    def pedf_hard(self):
        self.cache_affinity_loss = const(139.12)
        self.tick    = lin(2.72733346667, 0.00234666766667)
        self.sched   = lin(4.7450002, 0.008505556)
        self.ctxsw   = lin(14.9173331333, 0.0360633343333)
        self.sched2  = lin(3.8195004, 0.00518111066667)
        self.release = lin(4.72749983333, 0.00902333266667)

    def pfair_soft(self):
        self.cache_affinity_loss = const(172.05)
        self.tick    = lin(4.336447, 0.02602898)
        self.sched   = const(3.9941)
        self.ctxsw   = lin(2.5965942, 0.001111436)
        self.sched2  = lin(0.673345, 0.000138163333333)
        self.release = const(0)

    def spfair_soft(self):
        self.cache_affinity_loss = const(89.396)
        self.tick    = lin(2.16945041667, 0.0159464316667)
        self.sched   = const(3.593096)
        self.ctxsw   = lin(2.5391291, 0.000838853)
        self.sched2  = lin(0.65434875, 0.000149641666667)
        self.release = const(0)

    def gedf_soft(self):
        self.cache_affinity_loss = const(73.047)
        self.tick    = lin(2.0797836, 0.002412658)
        self.sched   = lin(11.3124283333, 0.0532636026667)
        self.ctxsw   = const(7.549997)
        self.sched2  = lin(0.5240276, 0.002363528)
        self.release = lin(5.8396803, 0.127082279333)

    def cedf_soft(self):
        self.cache_affinity_loss = const(67.047)
        self.tick    = const(2.86303)
        self.sched   = lin(4.56192583333, 0.00952826066667)
        self.ctxsw   = const(3.173618)
        self.sched2  = const(1.518119)
        self.release = const(16.533077)

    def pedf_soft(self):
        self.cache_affinity_loss = const(72.362)
        self.tick    = lin(2.0590809, 0.001849946)
        self.sched   = lin(2.03880783333, 0.00489201266667)
        self.ctxsw   = lin(4.7277338, 0.004505516)
        self.sched2  = lin(0.6390993, 0.002935457)
        self.release = lin(3.95167691667, 0.00483470833333)

    def gnpedf_soft(self):
        self.gedf_soft();
        self.cache_affinity_loss = const(0)

    # rtss09 starts here.
    # Method with _rm_ in the name imply release master (aka dedicated
    # interrupt handling) mode. These really should be in .csv files...

    def gedf_cache_hard(self):
        self.cache_affinity_loss = const(375.44)

    def gedf_cache_soft(self):
        self.cache_affinity_loss = const(73.047)

    def gsnedf_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 5.060000), (100.0, 5.533333), (150.0, 16.443333), (200.0, 9.550000), (250.0, 10.210000), (300.0, 17.076667), (350.0, 12.116667), (400.0, 12.716667), (450.0, 13.876667)])
        self.release   = monotonic_pwlin([(5.0, 19.220000), (50.0, 45.376667), (100.0, 88.880000), (150.0, 135.913333), (200.0, 150.633333), (250.0, 164.733333), (300.0, 174.490000), (350.0, 186.636667), (400.0, 200.140000), (450.0, 217.460000)])
        self.sched2    = monotonic_pwlin([(0.0, 0.953333), (50.0, 0.953333), (100.0, 1.820000), (150.0, 4.640000), (200.0, 5.916667), (250.0, 6.610000), (300.0, 6.996667), (350.0, 7.273333), (400.0, 7.800000), (450.0, 8.023333)])
        self.sched     = monotonic_pwlin([(5.0, 8.000000), (50.0, 39.346667), (100.0, 92.460000), (150.0, 132.733333), (200.0, 145.306667), (250.0, 156.976667), (300.0, 163.480000), (350.0, 172.140000), (400.0, 179.803333), (450.0, 189.310000)])
        self.ipi_delay = monotonic_pwlin([(50.0, 6.546667), (100.0, 5.000000), (150.0, 5.160000), (200.0, 5.016667), (250.0, 5.080000), (300.0, 7.150000), (350.0, 5.403333), (400.0, 9.433333), (450.0, 7.516667)])
        self.tick      = monotonic_pwlin([(10.0, 8.730000), (50.0, 8.876667), (100.0, 9.226667), (150.0, 9.380000), (200.0, 9.566667), (250.0, 9.613333), (300.0, 9.710000), (350.0, 9.850000), (400.0, 9.896667), (450.0, 10.033333)])

    def gsnedf_rm_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(0.0, 5.053333), (50.0, 5.053333), (100.0, 18.176667), (150.0, 8.273333), (200.0, 13.760000), (250.0, 17.990000), (300.0, 11.106667), (350.0, 16.050000), (400.0, 18.816667), (450.0, 14.010000)])
        self.release   = monotonic_pwlin([(5.0,  18.410000), (50.0, 47.593333), (100.0, 96.990000), (150.0, 143.030000), (200.0, 157.190000), (250.0, 172.933333), (300.0, 185.276667), (350.0, 199.633333), (400.0, 216.396667), (450.0, 238.946667)])
        self.sched2    = monotonic_pwlin([(0.0, 0.960000), (50.0, 0.960000), (100.0, 1.810000), (150.0, 4.656667), (200.0, 7.116667), (250.0, 8.256667), (300.0, 8.470000), (350.0, 8.693333), (400.0, 8.793333), (450.0, 8.846667)])
        self.sched     = monotonic_pwlin([(5.0, 8.653333), (50.0, 46.310000), (100.0, 104.620000), (150.0, 141.116667), (200.0, 153.336667), (250.0, 165.366667), (300.0, 173.026667), (350.0, 181.290000), (400.0, 191.580000), (450.0, 205.030000)])
        self.ipi_delay = monotonic_pwlin([(0.0, 4.403333), (50.0, 4.403333), (100.0, 6.976667), (150.0, 4.293333), (200.0, 4.400000), (250.0, 5.916667), (300.0, 4.660000), (350.0, 14.023333), (400.0, 11.086667), (450.0, 29.290000)])
        self.tick      = monotonic_pwlin([(50.0, 8.856667), (100.0, 9.096667), (150.0, 9.390000), (200.0, 9.610000), (250.0, 9.760000), (300.0, 9.990000), (350.0, 9.873333), (400.0, 9.836667), (450.0, 10.050000)])

    def gfgedf_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 19.186667), (100.0, 5.096667), (150.0, 5.363333), (200.0, 5.956667), (250.0, 6.283333), (300.0, 17.180000), (350.0, 16.470000), (400.0, 16.933333), (450.0, 17.653333)])
        self.release   = monotonic_pwlin([(50.0, 93.643333), (100.0, 172.750000), (150.0, 245.166667), (200.0, 272.833333), (250.0, 319.290000), (300.0, 337.473333), (350.0, 351.213333), (400.0, 361.643333), (450.0, 376.903333)])
        self.sched2    = monotonic_pwlin([(50.0, 1.013333), (100.0, 1.003333), (150.0, 3.526667), (200.0, 5.500000), (250.0, 6.853333), (300.0, 7.670000), (350.0, 8.253333), (400.0, 8.700000), (450.0, 8.940000)])
        self.sched     = monotonic_pwlin([(10.0, 6.926667), (45.0, 6.986667), (50.0, 9.380000), (100.0, 22.640000), (150.0, 59.523333), (200.0, 88.380000), (250.0, 98.723333), (300.0, 108.800000), (350.0, 118.536667), (400.0, 121.163333), (450.0, 119.146667)])
        self.ipi_delay = monotonic_pwlin([(50.0, 12.993333), (100.0, 5.403333), (150.0, 18.396667), (200.0, 38.133333), (250.0, 48.886667), (300.0, 63.120000), (350.0, 72.746667), (400.0, 80.480000), (450.0, 84.466667)])
        self.tick      = monotonic_pwlin([(50.0, 8.950000), (100.0, 9.180000), (150.0, 9.326667), (200.0, 9.480000), (250.0, 9.580000), (300.0, 9.716667), (350.0, 9.843333), (400.0, 9.806667), (450.0, 9.923333)])

    def gfgedf_rm_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 5.070000), (100.0, 5.870000), (150.0, 18.040000), (200.0, 17.973333), (250.0, 17.860000), (300.0, 6.520000), (350.0, 17.273333), (400.0, 16.893333), (450.0, 17.043333)])
        self.release   = monotonic_pwlin([(50.0, 83.260000), (100.0, 158.453333), (150.0, 227.350000), (200.0, 253.710000), (250.0, 298.896667), (300.0, 311.986667), (350.0, 327.796667), (400.0, 340.100000), (450.0, 355.030000)])
        self.sched2    = monotonic_pwlin([(0.0, 0.973333), (50.0, 0.973333), (100.0, 4.740000), (150.0, 7.063333), (200.0, 7.926667), (250.0, 8.536667), (300.0, 8.840000), (350.0, 9.413333), (400.0, 9.660000), (450.0, 9.803333)])
        self.sched     = monotonic_pwlin([(10.0, 7.013333), (45.0, 8.773333), (50.0, 14.976667), (100.0, 41.996667), (150.0, 68.706667), (200.0, 85.810000), (250.0, 92.466667), (300.0, 100.523333), (350.0, 109.010000), (400.0, 111.720000), (450.0, 108.880000)])
        self.ipi_delay = monotonic_pwlin([(0.0, 4.676667), (50.0, 4.676667), (100.0, 19.310000), (150.0, 36.196667), (200.0, 51.040000), (250.0, 57.543333), (300.0, 69.243333), (350.0, 78.793333), (400.0, 84.563333), (450.0, 86.813333)])
        self.tick      = monotonic_pwlin([(50.0, 8.846667), (100.0, 9.113333), (150.0, 9.256667), (200.0, 9.376667), (250.0, 9.476667), (300.0, 9.660000), (350.0, 9.743333), (400.0, 9.846667), (450.0, 10.016667)])

    def gqedf_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.976667), (100.0, 4.863333), (150.0, 4.910000), (200.0, 4.893333), (250.0, 4.926667), (300.0, 4.913333), (350.0, 4.860000), (400.0, 4.960000), (450.0, 4.876667)])
        self.release   = monotonic_pwlin([(50.0, 1.916667), (100.0, 1.966667), (150.0, 2.053333), (200.0, 2.316667), (250.0, 2.220000), (300.0, 2.240000), (350.0, 2.283333), (400.0, 2.336667), (450.0, 2.713333)])
        self.sched2    = monotonic_pwlin([(0.0, 1.000000), (50.0, 1.000000), (100.0, 1.046667), (150.0, 1.093333), (200.0, 1.090000), (250.0, 1.120000), (300.0, 1.133333), (350.0, 1.073333), (400.0, 1.180000), (450.0, 1.080000)])
        self.sched     = monotonic_pwlin([(50.0, 29.340000), (100.0, 33.073333), (150.0, 49.310000), (200.0, 68.703333), (250.0, 80.796667), (300.0, 88.020000), (350.0, 95.270000), (400.0, 101.763333), (450.0, 112.146667)])
        self.ipi_delay = const(0.0)
        self.tick      = monotonic_pwlin([(50.0, 50.280000), (100.0, 70.403333), (150.0, 99.060000), (200.0, 135.003333), (250.0, 169.886667), (300.0, 187.860000), (350.0, 209.936667), (400.0, 231.546667), (450.0, 258.543333)])

    def gqedf_rm_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.900000), (100.0, 4.886667), (150.0, 4.973333), (200.0, 4.963333), (250.0, 4.913333), (300.0, 4.953333), (350.0, 4.876667), (400.0, 4.883333), (450.0, 4.843333)])
        self.release   = monotonic_pwlin([(50.0, 1.936667), (100.0, 2.060000), (150.0, 2.126667), (200.0, 2.230000), (250.0, 2.393333), (300.0, 2.393333), (350.0, 2.413333), (400.0, 2.490000), (450.0, 2.563333)])
        self.sched2    = monotonic_pwlin([(50.0, 0.986667), (100.0, 1.030000), (150.0, 1.110000), (200.0, 1.136667), (250.0, 1.083333), (300.0, 1.140000), (350.0, 1.073333), (400.0, 1.176667), (450.0, 1.150000)])
        self.sched     = monotonic_pwlin([(10.0, 26.760000), (50.0, 28.833333), (100.0, 33.733333), (150.0, 51.483333), (200.0, 68.643333), (250.0, 82.670000), (300.0, 87.850000), (350.0, 94.570000), (400.0, 99.806667), (450.0, 106.246667)])
        self.ipi_delay = const(0.0)
        self.tick      = monotonic_pwlin([(5.0, 37.183333), (50.0, 51.453333), (100.0, 75.266667), (150.0, 107.360000), (200.0, 147.030000), (250.0, 182.093333), (300.0, 201.036667), (350.0, 224.253333), (400.0, 247.013333), (450.0, 272.850000)])

    def sgqedf_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.743333), (100.0, 4.576667), (150.0, 4.573333), (200.0, 4.586667), (250.0, 4.603333), (300.0, 4.660000), (350.0, 4.703333), (400.0, 4.850000), (450.0, 4.886667)])
        self.release   = monotonic_pwlin([(50.0, 2.050000), (100.0, 2.136667), (150.0, 2.003333), (200.0, 2.090000), (250.0, 2.196667), (300.0, 2.600000), (350.0, 2.263333), (400.0, 2.903333), (450.0, 3.060000)])
        self.sched2    = monotonic_pwlin([(50.0, 0.900000), (100.0, 0.916667), (150.0, 0.933333), (200.0, 0.953333), (250.0, 0.953333), (300.0, 0.976667), (350.0, 1.080000), (400.0, 2.670000), (450.0, 3.143333)])
        self.sched     = monotonic_pwlin([(50.0, 5.670000), (100.0, 7.133333), (1502.0, 7.833333), (200.0, 8.493333), (250.0, 9.123333), (300.0, 9.780000), (350.0, 10.933333), (400.0, 13.993333), (450.0, 17.770000)])
        self.ipi_delay = const(0.0)
        self.tick      = monotonic_pwlin([(10.0, 2.800628), (50.0, 6.343333), (100.0, 8.236667), (150.0, 12.363333), (200.0, 18.090000), (250.0, 26.333333), (300.0, 35.053333), (350.0, 51.440000), (400.0, 69.436667), (450.0, 87.080000)])

    def sgqedf_rm_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.723333), (100.0, 4.563333), (150.0, 4.556667), (200.0, 4.606667), (250.0, 4.656667), (300.0, 4.743333), (350.0, 4.783333), (400.0, 4.810000), (450.0, 4.866667)])
        self.release   = monotonic_pwlin([(50.0, 1.900000), (100.0, 1.953333), (150.0, 2.060000), (200.0, 2.273333), (250.0, 2.306667), (300.0, 2.683333), (350.0, 2.613333), (400.0, 2.596667), (450.0, 2.593333)])
        self.sched2    = monotonic_pwlin([(50.0, 0.883333), (100.0, 0.906667), (150.0, 0.943333), (200.0, 0.953333), (250.0, 0.970000), (300.0, 1.006667), (350.0, 1.973333), (400.0, 2.773333), (450.0, 3.246667)])
        self.sched     = monotonic_pwlin([(50.0, 5.446667), (100.0, 7.033333), (150.0, 7.980000), (200.0, 8.653333), (250.0, 9.213333), (300.0, 10.146667), (350.0, 11.920000), (400.0, 15.003333), (450.0, 23.176667)])
        self.ipi_delay = const(0)
        self.tick      = monotonic_pwlin([(50.0, 5.990000), (100.0, 8.690000), (150.0, 13.293333), (200.0, 18.776667), (250.0, 31.766667), (300.0, 43.570000), (350.0, 64.206667), (400.0, 86.473333), (450.0, 107.853333)])

    def ghqedf_hard(self):
        self.gedf_cache_hard()
        self.ctxsw     = monotonic_pwlin([(50.0, 5.036667), (100.0, 5.013333), (150.0, 5.276667), (200.0, 17.106667), (250.0, 6.666667), (300.0, 16.150000), (350.0, 7.500000), (400.0, 17.783333), (450.0, 17.076667)])
        self.release   = monotonic_pwlin([(50.0, 88.910000), (100.0, 149.916667), (150.0, 204.256667), (200.0, 223.543333), (250.0, 269.206667), (300.0, 273.960000), (350.0, 287.126667), (400.0, 288.213333), (450.0, 303.420000)])
        self.sched2    = monotonic_pwlin([(50.0, 0.950000), (100.0, 1.010000), (150.0, 3.250000), (200.0, 5.163333), (250.0, 5.996667), (300.0, 6.590000), (350.0, 7.360000), (400.0, 7.933333), (450.0, 8.283333)])
        self.sched     = monotonic_pwlin([(0.0, 10.400000), (50.0, 10.400000), (100.0, 22.536667), (150.0, 51.993333), (200.0, 80.743333), (250.0, 97.243333), (300.0, 116.350000), (350.0, 135.303333), (400.0, 155.786667), (450.0, 178.926667)])
        self.ipi_delay = monotonic_pwlin([(50.0, 5.270000), (100.0, 5.246667), (150.0, 7.946667), (200.0, 19.013333), (250.0, 29.383333), (300.0, 44.160000), (350.0, 62.690000), (400.0, 85.463333), (450.0, 104.390000)])
        self.tick      = monotonic_pwlin([(5.0, 8.633333), (50.0, 9.106667), (100.0, 9.103333), (150.0, 9.376667), (200.0, 9.443333), (250.0, 9.573333), (300.0, 9.796667), (350.0, 9.803333), (400.0, 9.746667), (450.0, 9.880000)])





    # soft

    def gsnedf_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.257116), (100.0, 4.279206), (150.0, 4.434866), (200.0, 4.580166), (250.0, 4.727793), (300.0, 4.833893), (350.0, 4.902915), (400.0, 4.972359), (450.0, 5.014317)])
        self.release   = monotonic_pwlin([(10.0, 10.360939), (50.0, 13.489388), (100.0, 20.400159), (150.0, 28.426785), (200.0, 36.111630), (250.0, 44.964809), (300.0, 51.642150), (350.0, 60.850067), (400.0, 69.746752), (450.0, 78.899514)])
        self.sched2    = monotonic_pwlin([(20.0, 0.716121), (50.0, 0.721284), (100.0, 0.758751), (150.0, 0.845318), (200.0, 0.954158), (250.0, 1.102249), (300.0, 1.261131), (350.0, 1.381913), (400.0, 1.515223), (450.0, 1.669743)])
        self.sched     = monotonic_pwlin([(10.0, 4.626504), (50.0, 8.094222), (100.0, 15.868094), (150.0, 25.103310), (200.0, 31.101610), (250.0, 36.054611), (300.0, 36.809331), (350.0, 38.494031), (400.0, 38.254152), (450.0, 36.805547)])
        self.ipi_delay = monotonic_pwlin([(50.0, 3.620983), (100.0, 3.368507), (150.0, 3.311931), (200.0, 3.294768), (250.0, 3.301273), (300.0, 3.332466), (350.0, 3.353710), (400.0, 3.400673), (450.0, 3.492159)])
        self.tick      = monotonic_pwlin([(10.0, 1.713356), (50.0, 1.725689), (100.0, 1.776682), (150.0, 1.837533), (200.0, 1.888639), (250.0, 1.921243), (300.0, 1.953355), (350.0, 1.983651), (400.0, 2.007965), (450.0, 2.044645)])

    def gsnedf_rm_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.088879), (100.0, 4.198589), (150.0, 4.460414), (200.0, 4.597347), (250.0, 4.774573), (300.0, 4.852097), (350.0, 4.942499), (400.0, 4.897116), (450.0, 5.028655)])
        self.release   = monotonic_pwlin([(10.0, 9.534481), (50.0, 13.458593), (100.0, 20.910676), (150.0, 29.948755), (200.0, 38.424155), (250.0, 47.840696), (300.0, 55.482255), (350.0, 65.605753), (400.0, 76.124222), (450.0, 87.408731)])
        self.sched2    = monotonic_pwlin([(20.0, 0.711569), (50.0, 0.719373), (100.0, 0.755307), (150.0, 0.834519), (200.0, 0.934839), (250.0, 1.112496), (300.0, 1.287508), (350.0, 1.400638), (400.0, 1.418092), (450.0, 1.671145)])
        self.sched     = monotonic_pwlin([(10.0, 4.957822), (50.0, 9.319991), (100.0, 18.807846), (150.0, 29.322102), (200.0, 35.532560), (250.0, 40.184822), (300.0, 40.443841), (350.0, 41.986560), (400.0, 37.114869), (450.0, 40.057083)])
        self.ipi_delay = monotonic_pwlin([(50.0, 3.091875), (100.0, 2.949932), (150.0, 2.953710), (200.0, 2.965412), (250.0, 2.981535), (300.0, 3.026697), (350.0, 3.041287), (400.0, 3.109150), (450.0, 3.228440)])
        self.tick      = monotonic_pwlin([(15.0, 1.709453), (50.0, 1.725898), (100.0, 1.775776), (150.0, 1.837298), (200.0, 1.896777), (250.0, 1.934195), (300.0, 1.980133), (350.0, 1.992160), (400.0, 2.028137), (450.0, 2.050872)])

    def gfgedf_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = avg = monotonic_pwlin([(50.0, 4.196198), (100.0, 4.076815), (150.0, 4.091428), (200.0, 4.124275), (250.0, 4.191172), (300.0, 4.245588), (350.0, 4.328150), (400.0, 4.349519), (450.0, 4.399700)])
        self.release   = monotonic_pwlin([(5.0, 15.123694), (50.0, 23.536775), (100.0, 38.318864), (150.0, 53.728555), (200.0, 67.164308), (250.0, 82.412810), (300.0, 91.122824), (350.0, 102.550111), (400.0, 109.369906), (450.0, 110.691870)])
        self.sched2    = monotonic_pwlin([(50.0, 0.768484), (100.0, 0.766533), (150.0, 0.792729), (200.0, 0.854960), (250.0, 0.965707), (300.0, 1.083607), (350.0, 1.200222), (400.0, 1.296980), (450.0, 1.434209)])
        self.sched     = monotonic_pwlin([(50.0, 4.552216), (100.0, 5.004378), (150.0, 5.857759), (200.0, 7.176427), (250.0, 8.506393), (300.0, 10.342688), (350.0, 11.927195), (400.0, 13.446336), (450.0, 14.272900)])
        self.ipi_delay = monotonic_pwlin([(50.0, 3.662822), (100.0, 3.483586), (150.0, 3.538007), (200.0, 3.775435), (250.0, 4.090263), (300.0, 4.682636), (350.0, 5.336483), (400.0, 6.135478), (450.0, 6.820003)])
        self.tick      = monotonic_pwlin([(10.0, 1.681990), (50.0, 1.707446), (100.0, 1.757044), (150.0, 1.817713), (200.0, 1.869111), (250.0, 1.905806), (300.0, 1.918752), (350.0, 1.961229), (400.0, 1.959019), (450.0, 1.971021)])

    def gfgedf_rm_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.029668), (100.0, 4.003217), (150.0, 4.071368), (200.0, 4.117165), (250.0, 4.173408), (300.0, 4.246263), (350.0, 4.323759), (400.0, 4.397633), (450.0, 4.451260)])
        self.release   = monotonic_pwlin([(5, 12.597124), (50.0, 20.139925), (100.0, 33.457051), (150.0, 47.915713), (200.0, 60.924967), (250.0, 75.037101), (300.0, 83.141689), (350.0, 94.652170), (400.0, 101.729210), (450.0, 103.121330)])
        self.sched2    = monotonic_pwlin([(25.0, 0.730701), (50.0, 0.730940), (100.0, 0.778376), (150.0, 0.849833), (200.0, 0.937816), (250.0, 1.046366), (300.0, 1.180734), (350.0, 1.291660), (400.0, 1.424485), (450.0, 1.565254)])
        self.sched     = monotonic_pwlin([(50.0, 4.078418), (100.0, 4.816280), (150.0, 5.968767), (200.0, 7.408112), (250.0, 8.509983), (300.0, 10.142849), (350.0, 11.712992), (400.0, 13.156129), (450.0, 13.959959)])
        self.ipi_delay = monotonic_pwlin([(50.0, 3.085628), (100.0, 3.158275), (150.0, 3.417121), (200.0, 3.770703), (250.0, 4.005280), (300.0, 4.527855), (350.0, 5.105332), (400.0, 5.796901), (450.0, 6.422823)])
        self.tick      = monotonic_pwlin([(20.0, 1.680745), (50.0, 1.703186), (100.0, 1.763110), (150.0, 1.821684), (200.0, 1.862838), (250.0, 1.899689), (300.0, 1.923040), (350.0, 1.961775), (400.0, 1.987410), (450.0, 2.009000)])

    def gqedf_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 3.932774), (100.0, 3.830004), (150.0, 3.823897), (200.0, 3.829520), (250.0, 3.844619), (300.0, 3.878260), (350.0, 3.892930), (400.0, 3.976483), (450.0, 3.993735)])
        self.release   = monotonic_pwlin([(20, 1.565351), (50.0, 1.573317), (100.0, 1.605717), (150.0, 1.622647), (200.0, 1.661057), (250.0, 1.693608), (300.0, 1.752715), (350.0, 1.768444), (400.0, 1.833865), (450.0, 1.871730)])
        self.sched2    = monotonic_pwlin([(50.0, 0.748132), (100.0, 0.745381), (150.0, 0.755979), (200.0, 0.767332), (250.0, 0.773118), (300.0, 0.781287), (350.0, 0.797554), (400.0, 0.822520), (450.0, 0.830449)])
        self.sched     = monotonic_pwlin([(50.0, 3.803485), (100.0, 3.679596), (150.0, 3.695616), (200.0, 3.791913), (250.0, 3.894480), (300.0, 3.989817), (350.0, 4.138575), (400.0, 4.508925), (450.0, 4.582703)])
        self.ipi_delay = const(0.0)
        self.tick      = monotonic_pwlin([(50.0, 2.876093), (100.0, 2.998473), (150.0, 3.179663), (200.0, 3.436532), (250.0, 3.812870), (300.0, 4.328181), (350.0, 5.002456), (400.0, 6.007492), (450.0, 7.240790)])

    def gqedf_rm_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 3.827149), (100.0, 3.778489), (150.0, 3.791796), (200.0, 3.818305), (250.0, 3.853977), (300.0, 3.904391), (350.0, 3.931144), (400.0, 3.951878), (450.0, 3.984602)])
        self.release   = monotonic_pwlin([(0.0, 1.561158), (50.0, 1.561158), (100.0, 1.589809), (150.0, 1.633248), (200.0, 1.673685), (250.0, 1.713563), (300.0, 1.780821), (350.0, 1.801693), (400.0, 1.838063), (450.0, 1.895681)])
        self.sched2    = monotonic_pwlin([(0.0, 0.717700), (50.0, 0.717700), (100.0, 0.732346), (150.0, 0.754201), (200.0, 0.764130), (250.0, 0.775280), (300.0, 0.787279), (350.0, 0.796791), (400.0, 0.800558), (450.0, 0.821383)])
        self.sched     = monotonic_pwlin([(50.0, 3.762715), (100.0, 3.747637), (150.0, 3.914522), (200.0, 3.975425), (250.0, 4.012949), (300.0, 4.165312), (350.0, 4.369225), (400.0, 4.515190), (450.0, 4.687490)])
        self.ipi_delay = const(0.0)
        self.tick      = monotonic_pwlin([(50.0, 2.868984), (100.0, 3.005491), (150.0, 3.196195), (200.0, 3.434289), (250.0, 3.896306), (300.0, 4.465542), (350.0, 5.291298), (400.0, 6.442091), (450.0, 7.894566)])

    def ghqedf_soft(self):
        self.gedf_cache_soft()
        self.ctxsw     = monotonic_pwlin([(50.0, 4.182817), (100.0, 4.097358), (150.0, 4.086779), (200.0, 4.097755), (250.0, 4.148575), (300.0, 4.222174), (350.0, 4.324972), (400.0, 4.429505), (450.0, 4.439013)])
        self.release   = monotonic_pwlin([(5.0, 17.776326), (50.0, 27.332953), (100.0, 39.973609), (150.0, 51.788456), (200.0, 60.815847), (250.0, 71.805802), (300.0, 77.028258), (350.0, 84.381160), (400.0, 89.684504), (450.0, 94.170941)])
        self.sched2    = monotonic_pwlin([(50.0, 0.764010), (100.0, 0.762326), (150.0, 0.783979), (200.0, 0.838372), (250.0, 0.931207), (300.0, 1.041730), (350.0, 1.140392), (400.0, 1.255504), (450.0, 1.312786)])
        self.sched     = monotonic_pwlin([(50.0, 5.177300), (100.0, 5.485217), (150.0, 6.131576), (200.0, 7.078387), (250.0, 8.540165), (300.0, 10.799998), (350.0, 13.466103), (400.0, 16.872205), (450.0, 20.679590)])
        self.ipi_delay = monotonic_pwlin([(50.0, 3.685732), (100.0, 3.498932), (150.0, 3.507025), (200.0, 3.587282), (250.0, 3.758739), (300.0, 4.137825), (350.0, 4.734812), (400.0, 5.652826), (450.0, 6.717968)])
        self.tick      = monotonic_pwlin([(15.0, 1.688439), (50.0, 1.712122), (100.0, 1.763488), (150.0, 1.821727), (200.0, 1.870288), (250.0, 1.911194), (300.0, 1.945750), (350.0, 1.995029), (400.0, 2.035951), (450.0, 2.012400)])


    def jsa10_hard(self):
        self.gsnedf_hard()
        self.zero_non_irq_overheads()

    def jsa10_rm_hard(self):
        self.gsnedf_rm_hard()
        self.zero_non_irq_overheads()

    def jsa10_soft(self):
        self.gsnedf_soft()
        self.zero_non_irq_overheads()

    def jsa10_rm_soft(self):
        self.gsnedf_rm_soft()
        self.zero_non_irq_overheads()


class SystemOverheads(object):
    """New-style overhead object"""
    def __init__(self, reference=Overheads(), n=0,
                 migration_cost=None, const_reference=True):
        if const_reference:
            self.tick    = const(reference.tick(n))
            self.sched   = const(reference.sched(n) + reference.sched2(n))
            self.release = const(reference.release(n))
            self.timer_migration = const(reference.pull_timer(n))
            self.ipi_latency = const(reference.ipi_delay(n))
            self.context_switch = const(reference.ctxsw(n))
        else:
            self.tick    = reference.tick
            self.sched   = lambda x: (reference.sched(x) + reference.sched2(x))
            self.release = reference.release
            self.timer_migration = reference.pull_timer
            self.ipi_latency = reference.ipi_delay
            self.context_switch = reference.ctxsw
        if migration_cost:
            self.cpmd = migration_cost
        else:
            self.cpmd = lambda x, y: 0.0


class CPMD(object):

    MEM, L1, L2, L3 = 0, 1, 2, 3
    MAPPING = list(enumerate(["MEM", "L1", "L2", "L3"]))

    def __init__(self, l1=0, l2=0, l3=0, mem=0):
        self.mem_hierarchy  = [const(mem), const(l1), const(l2), const(l3)]
        for (i, name) in CPMD.MAPPING:
            self.__dict__[name] = self.mem_hierarchy[i]

    def cpmd_cost(self, shared_mem_level, working_set_size):
        return self.mem_hierarch[shared_mem_level](working_set_size)

    def max_cost(self, working_set_size):
        return max([f(working_set_size) for f in self.mem_hierarchy])

    @staticmethod
    def from_file(fname, non_decreasing=True):
        id = lambda x: x
        data = load_column_csv(fname, convert=float)
        if not 'WSS' in data.by_name:
            raise IOError, 'no WSS?'

        o = CPMD()

        for idx, name in CPMD.MAPPING:
            if name in data.by_name:
                points = zip(data.by_name['WSS'], data.by_name[name])
                if non_decreasing:
                    o.mem_hierarchy[idx] = monotonic_pwlin(points, coerce=id)
                else:
                    o.mem_hierarchy[idx] = piece_wise_linear(points, coerce=id)
                o.__dict__[name] = o.mem_hierarchy[idx]
        return o

    def get_profile(self, points):
        return [[x] + [cost(x)
                       for cost in self.mem_hierarchy]
                for x in points]

    def write_profile(self, fname, points, want_header=True,
                      comment_header=False):
        f = open(fname, 'w')
        if want_header:
            print_row(['WSS'] +
                      [name for (_,name) in CPMD.MAPPING],
                      col_sep=' ,',
                      f=f,
                      prepend='#' if comment_header else ' ',
                      col_width=20)
        table = self.get_profile(points)
        print_rows(table, f=f, col_sep=',', prepend=' ', col_width=20)
        f.close()


def write_interpolated_files(files):
    for file in files:
        print "Generating profile for %s..." % file
        if 'pmo' in file:
            points = range(0, 10000, 50)
            o = CPMD.from_file(file)
        else:
            points = range(1, 500)
            o = Overheads.from_file(file)
        o.write_profile('overhead-profile-' + os.path.basename(file), points)

def write_interpolated_builtin():
    profiles = [(Overheads.gedf_soft, "gedf-soft"),
                (Overheads.gedf_hard, "gedf-hard"),
                (Overheads.gsnedf_soft, "gsnedf-soft"),
                (Overheads.gsnedf_hard, "gsnedf-hard"),
                ]
    points = range(1, 500)
    for f, name in profiles:
        print "Generating profile for %s..." % name
        o = Overheads()
        f(o)
        o.write_profile('overhead-profile-' + name + '.csv', points)


if __name__ == '__main__':
    import sys
    import os.path

    if len(sys.argv) == 2 and sys.argv[1] == 'builtin':
        write_interpolated_builtin()
    else:
        write_interpolated_files(sys.argv[1:])
