from util      import ms2us, us2ms
from util.quantor   import forall
from quanta    import has_integral_period, quantize_wcet

def account_for_overheads(overheads, tasks):
    """We consider the quantum size to be an
    overhead (since ideally we would have a quantum size of 0).

    The effective quantum size is the amount of time that a job
    can use each quantum to make progress:

    effective quantum size = quantum size - tick - max(migration, preempt)
    - sched - ctxsw

    The inflated job execution cost is then found by the number of
    quanta required to satisfy the uninflated execution cost.

    Account for overheads after everything else has been taken into account.
    
    Note: This only applies to periodic tasks since no release overheads
          are accounted for, i.e., this is for pure time-triggered scheduling.
    """
    T = tasks
    qsize = overheads.quantum_size

    # make sure that all periods are multiple of the quantum size
    assert forall(T)(has_integral_period(us2ms(qsize)))

    equant  = float(qsize)
    equant -= overheads.tick(T)
    equant -= overheads.sched(T)
    equant -= overheads.sched2(T)
    equant -= overheads.ctxsw(T)
    equant -= overheads.cache_affinity_loss(T)
    quantize_wcet(us2ms(qsize), T, us2ms(equant))
    return T

def is_schedulable(no_cpus, tasks):
    return tasks.utilization() <= no_cpus and \
        forall(tasks)(lambda t: t.period >= t.cost)

