#!/usr/bin/env python

import os.path
import sys
import traceback

class Progress(object):
    def point_started(self, pt):
        pass

    def point_finished(self, result):
        pass

    def interval_added(self, intervals):
        pass


class TraceProgress(Progress):
    def point_started(self, pt):
        print 'sampling', pt,

    def point_finished(self, result):
        print '-->', result

    def interval_added(self, intervals):
        print ':: added', intervals[-1]


def get_feedback(kargs, default=Progress()):
    if 'feedback' in kargs:
        tmp = kargs['feedback']
        del kargs['feedback']
        return tmp
    else:
        return default

def measure(process, evaluate):
    def _measure(instance):
        return evaluate(process(instance))
    return _measure

def fraction(x, all, count):
    return float(x) / float(count)

def statistic(stat, default=0.0):
    def normalize(x, all, samples):
        return stat(x) if len(x) > 0 else default
    return normalize

def seq_point(generator, measure, samples=10, normalize=fraction, init=int):
    """Sample one point.

    generator must be a callable and must return a list/tuple of parameters to
    be used in the sampling. It is called for each sample (thus it supports
    randomized sampling).

    measure is a sequence of test functions. For each sample, each element in
    measure will be invoked on what generator() produced. The result of each
    invocation is accumulated.

    Normalize can be used to post-process the samples. By default, it devides
    the accumulated value for each element in measure by the number of samples
    (it normalizes the value).

    """
    value = [init() for _ in measure]
    for _ in xrange(samples):
        instance = generator()
        for i in xrange(len(measure)):
            value[i] += measure[i](*instance)
    return [normalize(x, value, samples) for x in value]

def no_parallelism():
    global point
    point = seq_point

try:
    from itertools    import imap
    from util         import vector_xadd, add_disable_parallelism_callback
    import random
    import os

    try:
        # Python 2.5 with PyProcessing package
        import processing as par
        from   processing import cpuCount as cpu_count
    except ImportError:
        # Ok, let's see if we are runing on Python >= 2.6
        import multiprocessing as par
        from   multiprocessing import cpu_count
        # if this fails we fall back to sequential execution

    _generator = None
    _measure   = None

    def one_sample(_):
        instance = _generator()
        # The processing library re-seeds the prng state, so this is ok
        # for randomized sampling.
        try:
            result = map(lambda i: _measure[i](*instance), xrange(len(_measure)))
        except:
            # print stack trace before it gets mangled by multiprocessing
            print >>sys.stderr, "=" * 80
            print >>sys.stderr, "EXCEPTION:"
            traceback.print_exc(file=sys.stderr)
            print >>sys.stderr, "=" * 80
            raise

        return result

    def par_point(generator, measure, samples=10, normalize=fraction, init=int):
        """Like sample.point(). However, this implementation uses a map-reduce
        approach.
        """
        # make parameters available to one_sample()
        # we must do it this way because Python can't pickle a closure
        global _generator
        global _measure
        _generator = generator
        _measure   = measure

        # we need to spawn workers every single time, otherwise they are not
        # aware of the changes to _generator and _measure
        workers = par.Pool()

        # parallel map
        bunch = max(min(samples / cpu_count(), 10), 1)
        results = workers.imap_unordered(one_sample, xrange(samples),
                                         chunksize=bunch)

        # sequential reduce
        value = reduce(vector_xadd, results, [init() for _ in measure])
        return [normalize(x, value, samples) for x in value]

    # we can use the parallel implementation
    point = par_point
    add_disable_parallelism_callback(no_parallelism)
except ImportError:
    print "# [sample.py] Parallelism unavailable; using sequential sampling."
    # parallel implementation not available, let's use the sequential version
    point = seq_point

def all_points(dims):
    """Yields all points (as lists) along the discrete dimensions
       given in the list dims (dims must be a list as it is evaluated
       multiple times).

       all_points([[0, 1], [2, 3]]) -> [[0, 2], [0, 3], [1, 2], [1, 3]]
    """
    if dims == []:
        yield []
    else:
        for x in dims[0]:
            for r in all_points(dims[1:]):
                yield [x] + r

def field(dimensions, metagen,
          *args, **kargs):
    """Invoke point() for all points along the specified dimensions.
       Returns a list of rows, with the first columns being
       the coordinates of the sampled point.
    """
    feedback = get_feedback(kargs)
    data = []
    for p in all_points(dimensions):
        feedback.point_started(p)
        fp = p + point(metagen(*p), *args, **kargs)
        feedback.point_finished(fp)
        data.append(fp)
    return data

def range(points, *args, **kargs):
    """Invoke point() for each point in points.
       Returns a list of rows, with the first column being
       the sampling points.
    """
    return field([points], *args, **kargs)

def close_enough_epsilon(epsilon, min_delta=None):
    def scalar_delta((p1, val1), (p2, val2)):
        return all(map(lambda x, y: abs(x - y) < epsilon, val1, val2))
    def scalar_min_delta((p1, val1), (p2, val2)):
        return \
            all(map(lambda x, y: abs(x - y) < min_delta, p1, p2)) or \
            all(map(lambda x, y: abs(x - y) < epsilon, val1, val2))
    if min_delta is None:
        return scalar_delta
    else:
        return scalar_min_delta

def middle_point_scalar(p1, p2):
    return [p1[0] + (p2[0] - p1[0]) / 2.0]

def range_adaptive(start, end, close_enough, middle_point, metagen,
                   *args, **kargs):
    points    = [] # return value
    intervals = [] # stack of work to be done

    feedback = get_feedback(kargs)

    def eval(p):
        feedback.point_started(p)
        return point(metagen(*p), *args, **kargs)

    def add(p, val):
        fp = p + val
        points.append(fp)
        feedback.point_finished(fp)

    def push(p1, val1, p2, val2):
        intervals.append((p1, val1, p2, val2))
        feedback.interval_added(intervals)

    def pop():
        return intervals.pop()

    start_val = eval(start)
    add(start, start_val)
    end_val   = eval(end)
    add(end, end_val)
    push(start, start_val, end, end_val)
    while intervals:
        (p1, val1, p2, val2) = pop()
        p3 = middle_point(p1, p2)
        val3 = eval(p3)
        add(p3, val3)
        if not close_enough((p1, val1), (p3, val3)):
            push(p1, val1, p3, val3)
        if not close_enough((p3, val3), (p2, val2)):
            push(p3, val3, p2, val2)
    return points

if __name__ == "__main__":
    import binpack
    import random
    import sys
    import csv

    dump =  csv.writer(sys.stdout).writerows

    if False:
        gen = lambda x: lambda: [[random.randint(0, x) for _ in xrange(10)]]
        dump(range(xrange(10,100,10), gen,
                   [lambda seq: min(seq),
                    lambda seq: sum(seq) / float(len(seq)),
                    lambda seq: max(seq)],
                   samples=100))

        print '\n\n'
        dump(field([xrange(1, 4), xrange(1, 4)],
                   lambda x,y: lambda: (float(x), float(y)),
                   [lambda x, y: x * y, lambda x, y: x / y],
                   samples=1))

    no_parallelism()
    dump(sorted(range_adaptive([0], [10], close_enough_epsilon(0.000001, min_delta=0.05),
                        middle_point_scalar, lambda x: lambda : [x],
                        [lambda x: x * 10], samples=1, feedback=TraceProgress())))
