"""
EDF-fm Semi-partitioned scheduling algorithm (SRT)

Anderson, Bud, Devi, An EDF-based Scheduling Algorithm for Multiprocessor
Soft Real-Time Systems, ECRTS 2005.

(Generalized version: Anderson, Bud, Devi, An EDF-based Restricted-Migration
Scheduling Algorithm for Multiprocessor Soft Real-Time Systems,
Real-Time Systems , Volume 38, Number 2, pp. 85-131, February 2008.)
"""
from __future__ import division
from util.quantor import forall

# EDF-fm assignment phase

from math import ceil, floor

from fractions import Fraction

from binpack import DuctileItem, DuctileBin

from util.storage import storage
from util import ms2us

import sys

DENOM_LIMIT  = 100000
MIN_SHARE    = Fraction(1, DENOM_LIMIT)
TIMER_PERIOD = ms2us(1000)

def jobs_cycle(task, share):
    fraction = share / task.utilization()
    frac = Fraction.from_float(_frac).limit_denominator(DENOM_LIMIT)
    return frac.denominator

def account_for_timer_tick(exec_cost, timer_period, tick_cost, max_cost):
    num_ticks = int(ceil(exec_cost / float(timer_period)))
    exec_cost += num_ticks * tick_cost
    last_ticks = num_ticks

    num_ticks = int(ceil(exec_cost / float(timer_period)))
    while num_ticks > last_ticks and exec_cost <= max_cost:
        exec_cost += (num_ticks - last_ticks) * tick_cost
        last_ticks = num_ticks
        num_ticks  = int(ceil(exec_cost / float(timer_period)))

    return exec_cost

def inflated_utilization(task, cost_inflation):
    return (task.cost + cost_inflation) / float(task.period)

class InflatedTask(DuctileItem):
    def __init__(self, params, rel_irqs=1, tm_irqs=0):
        self.params = params

        # params.jobs_cycle == 0 => static task
        # params.task  --- reference to the task
        # params.cost  --- cost of the task (could be less than params.task.cost)

        # cost for preemption, including CPMD
        # this depends on n, but not on each individual task
        # since we assume max. CPMD for every task
        # params.cpmd  --- cpmd costs

        # this is individual for each task, since periods differ
        self.rel_irqs   = rel_irqs  # number of release interrupts
        self.tm_irqs    = tm_irqs   # number of timer migrations
        # lower estimate for infeasibility check
        if self.params.share is None:
            self.share = 0.0
        else:
            self.share = self.params.share
        self.compute_cost(1)

    def num_releases_during(self, delta):
        return int(ceil(delta / float(self.params.task.period)))

    def num_timer_migrations_during(self, delta):
        if self.params.jobs_cycle:
            return int(ceil(delta /
                            float(self.params.task.period *
                                  self.params.jobs_cycle)))
        else:
            return 0

    def size(self):
        if self.params.jobs_cycle:
            pass
        else:
#            print "%s inflated by %f" % (self, (self.share - self.params.task.utilization()) / self.params.task.utilization() * 100)
            pass
        return self.share

    def copy(self):
        return InflatedTask(self.params,
                            self.rel_irqs,
                            self.tm_irqs)

    def update_cost(self, bin):
        self.compute_cost(len(bin.items))

    def inc_size(self, it):
        self.rel_irqs += it.num_releases_during(self.params.task.period)
        if it.params.jobs_cycle:
            self.tm_irqs += it.num_timer_migrations_during(self.params.task.period)

    def available_share(self, pre_oh):
        return pre_oh / (1.0 + self.cost_inflation() /
                        float(self.params.task.cost))

    def compute_cost(self, n):
        if self.params.share is None:
            fraction = 0.0
        else:
            fraction = self.params.share / self.params.task.utilization()
        # compute preempt charge
        tick_cost = self.params.ohead.tick(n)
        preempt  = tick_cost + self.params.cpmd + \
            2 * (self.params.ohead.context_switch(n) +
                 self.params.ohead.sched(n))

        rel_cost = self.params.ohead.release(n)
        tm_cost  = self.params.ohead.timer_migration(n)
        # compute inflated utilization
        cost = (self.params.task.cost + preempt +
                self.rel_irqs * rel_cost +
                self.tm_irqs * tm_cost)

        cost = account_for_timer_tick(cost, TIMER_PERIOD, tick_cost,
                                      self.params.task.period)
        self.inflated_cost = int(ceil(cost))
        self.share = fraction * self.inflated_cost / \
                     float(self.params.task.period)

    def update_size(self, it, bin):
        self.inc_size(it)
        self.compute_cost(len(bin.items) + 1) # number of tasks in the partition

    def determine_size(self, bin):
        n = len(bin.items) + 1

        # compute interrupt charge
        self.tm_irqs  = 0
        self.rel_irqs = 1 # charge for our own delay
        for it in bin.items:
            self.inc_size(it)

        # estimate cost
        self.compute_cost(n)

    def cost_inflation(self):
        # what we wanted vs. what we cost
        return self.inflated_cost - self.params.task.cost

    def inflated_utilization(self):
        cost_infl = self.cost_inflation()
        # add that into the task's utilization
        return inflated_utilization(self.params.task, cost_infl)

    def __str__(self):
        return "InflatedTask(%s, %s, %s)" % (self.params.task, self.params, self.share)


def inflated_task(task, ohead, cpu, other_cpu=None, share=None):
    cycle = 0 if other_cpu is None else 2 # safe lower bound, we don't know the frequency

    # preemption
    cpmd  = ohead.cpmd(cpu, cpu)
    #print "preempt (%d): %f" % (cpu, cpmd)
    if share is None and other_cpu is None:
        share = task.utilization()
    # create non-changing parameters
    params = storage(task=task, share=share, cpmd=cpmd,
                     jobs_cycle=cycle, ohead=ohead)
    return InflatedTask(params)


def choose_luf(tasks, capacity, cost_inflation_estimate):
    # Assumes tasks is ordered.
    # Find smallest task that doesn't fit.
    last = tasks[0]
    for t in tasks[1:]:
        if inflated_utilization(t, cost_inflation_estimate) > capacity:
            # doesn't fit, keep going
            last = t
        else:
            # this fits; take the previous one
            break
    return last

# assumes ts has already been converted to us
# ohead is new-style overhead object
def assign_ts_with_overheads(ts, no_cpus, ohead, want_bins=False):
    fail = None if want_bins else False

    cpus = [DuctileBin() for _ in range(no_cpus)]
    for cpu in cpus:
        # keep track of migrating tasks
        cpu.migrating = []

    idx = 0
    cpu = cpus[idx]

    # make a copy for pop()
    unassigned = list(ts)
    # decreasing utilizations --- highest utilization first HUF
    unassigned.sort(key = lambda t: t.utilization(), reverse = True)

    while unassigned:
        # try placing the heaviest
        t = unassigned[0]

        # create ductile wrapper
        it = inflated_task(t, ohead, idx)
        if cpu.infeasible_to_fit(it) or not cpu.try_to_add(it):
            # Ok, doesn't fit in one piece.
            # Let's try splitting some task.

            inf = it.cost_inflation()
            t = choose_luf(unassigned, cpu.spare_capacity(), inf)

            # if we reach the last CPU, then overheads made this task set
            # unschedulable
            idx += 1
            if idx == no_cpus:
                return fail

            split_ok = False
            it = inflated_task(t, ohead, idx - 1, idx, None) # assume 0 share for now
            if cpu.try_to_add_no_commit(it):
                # ok, now we can figure out how much we can allocate here
                share = it.available_share(cpu.spare_capacity())
                # makes no sense to fail now
                if not (0 <= share < t.utilization()):
                    print >>sys.stderr, "SOFT ASSERT: 0 <= share < t.utilization()"
                    print >>sys.stderr, "share=%f t.utilization=%f" % (share, t.utilization())
                    print >>sys.stderr, "inf est=%d fact=%d" % (inf, it.cost_inflation())
                if share >= MIN_SHARE:
                    # great, we split it
                    it.params.share = share
                    it.update_cost(cpu)
                    cpu.commit()
                    cpu.migrating.append(t)
                    split_ok = True
                else:
                    # too bad, we can't get enough out of it
                    cpu.rollback()
            # Before we move to the next, let's make sure that
            # we didn't just create an unschedulable CPU.
            if len(cpu.migrating) == 2:
                # must check
                # find inflated tasks
                migs = [it for it in cpu.items if it.params.task in cpu.migrating]
                assert len(migs) == 2
                # check that tardiness is bounded
                migu = migs[0].inflated_utilization() + migs[1].inflated_utilization()
                if (migu) > 1:
                    return fail
            else:
                # can't be more than 2
                assert len(cpu.migrating) < 2
            cpu = cpus[idx]
            did_fit_after_all = False
            if split_ok:
                # must allocate rest of the split task
                rem =  t.utilization() - share
                if rem > 0:
                    it = inflated_task(t, ohead, idx - 1, idx, rem)
                else:
                    # this can happen if we overestimated the overheads during choose_luf()
                    did_fit_after_all = True
            else:
                # Couldn't find an acceptable split => this CPU is full due to
                # overheads. So let's move to the next.
                it = inflated_task(t, ohead, idx)
            if  not (did_fit_after_all or  cpu.try_to_add(it)):
                # Should not fail as the bin is empty.
                # The task must be overloaded.
                assert it.size() > 1
                return fail
            else:
                # Ok, we managed to split it and sqeeze it in.
                # Remove task from unassigned list and continue with heaviest tasks.
                if not did_fit_after_all:
                    # yes, it's actually migrating, keep track of it
                    cpu.migrating.append(t)
                unassigned.remove(t)
        else:
            # was added ok
            # assert it in cpu.items
            unassigned.pop(0) # remove task

    for b in cpus:
        for it in b.items:
            assert (0 <= len(b.migrating) <= 2)
            if (it.params.task in b.migrating):
                it.params.task.tardiness = 0
            elif len(b.migrating) == 0:
                it.params.task.tardiness = 0
            else:
                migs = [i for i in b.items if i.params.task in b.migrating]
                higher_share = migs[0].share
                higher_fraction = migs[0].params.share \
                                  / migs[0].params.task.utilization()
                if (higher_fraction > 1):
                    print "WARNING: higher_fraction > 1, is ", higher_fraction
                higher_cost = migs[0].inflated_cost
                if len(b.migrating) == 2:
                    lower_share = migs[1].share
                    lower_fraction = migs[1].params.share \
                                     / migs[1].params.task.utilization()
                    if (lower_fraction > 1):
                        print "WARNING: lower_fraction > 1, is ", lower_fraction
                    lower_cost = migs[1].inflated_cost
                else:
                    lower_share = 0
                    lower_fraction = 0
                    lower_cost = 0
                it.params.task.tardiness = (higher_cost * (higher_fraction + 1)
                                            + lower_cost * (lower_fraction + 1)
                                           ) / (1 - higher_share - lower_share)
    
    # found an assignment
    if want_bins:
        return cpus
    else:
        return True
