"""
EDF-os Semi-partitioned scheduling algorithm (SRT)

"""
from util.quantor import forall

from math import ceil, floor

from fractions import Fraction

from binpack import DuctileItem, DuctileBin, MaxSpareCapacity, ignore

from util.storage import storage
from util import ms2us
import random
import itertools
import sys

DENOM_LIMIT  = 100000
MIN_SHARE    = Fraction(1, DENOM_LIMIT)
TIMER_PERIOD = ms2us(1000)

def account_for_timer_tick(exec_cost, timer_period, tick_cost, max_cost):
    num_ticks = int(ceil(exec_cost / float(timer_period)))
    exec_cost += num_ticks * tick_cost
    last_ticks = num_ticks

    num_ticks = int(ceil(exec_cost / float(timer_period)))
    while num_ticks > last_ticks and exec_cost <= max_cost:
        exec_cost += (num_ticks - last_ticks) * tick_cost
        last_ticks = num_ticks
        num_ticks  = int(ceil(exec_cost / float(timer_period)))

    return exec_cost

def inflated_utilization(task, cost_inflation):
    return (task.cost + cost_inflation) / float(task.period)

class InflatedTask(DuctileItem):
    def __init__(self, params, rel_irqs=1, remote_releases=0):
        self.params = params

        # params.task  --- reference to the task
        # params.cost  --- cost of the task (could be less than params.task.cost)

        # cost for preemption, including CPMD
        # this depends on n, but not on each individual task
        # since we assume max. CPMD for every task
        # params.cpmd  --- cpmd costs

        # this is individual for each task, since periods differ
        self.rel_irqs   = rel_irqs  # number of release interrupts
        self.remote_releases = remote_releases   # number of remote releases
        # lower estimate for infeasibility check
        self.share      = params.cost / float(params.task.period)
        self.compute_cost(1)

    def num_releases_during(self, delta):
        return int(ceil(delta / float(self.params.task.period)))

    def num_releases_this_cpu_during(self, delta):
        if self.params.migrating:
            if self.share is not None and self.share > 0:
                fraction = self.share / self.params.task.utilization()
            else:
                fraction = self.params.share_bound \
                           / self.params.task.utilization()
            num_tot_releases = ceil((delta + max(0, self.params.task.tardiness))
                                    / self.params.task.period)
            return int(min(num_tot_releases, ceil(fraction * (num_tot_releases - 1)) + 1))
        else:
            return self.num_releases_during(delta)

    # Can't be more than half the number of consecutive releases (plus one for
    # a job that finished early on another CPU and migrates here), or some of
    # those releases would necessarily be consecutive.
    def num_migrate_to_this_cpu_during(self, delta):
        if self.params.migrating:
            if self.share is not None and self.share > 0:
                fraction = self.share / self.params.task.utilization()
            else:
                fraction = self.params.share_bound \
                           / self.params.task.utilization()
            num_tot_releases = ceil((delta + max(0, self.params.task.tardiness))
                                    / self.params.task.period)
            return int(min(ceil((num_tot_releases+1)/2),
                    ceil(fraction * (num_tot_releases - 1)) + 1))
        else:
            return 0

    def available_share(self, pre_oh):
        return pre_oh / (1.0 + self.cost_inflation() /
                        float(self.params.task.cost))

    def size(self):
        if self.params.migrating:
            pass
        else:
#            print "%s inflated by %f" % (self, (self.share - self.params.task.utilization()) / self.params.task.utilization() * 100)
            pass
        return self.share

    def copy(self):
        return InflatedTask(self.params,
                            self.rel_irqs,
                            self.remote_releases)

    def update_cost(self, bin):
        self.compute_cost(len(bin.items))

    def inc_size(self, it):
        self.rel_irqs += it.num_releases_this_cpu_during(self.params.task.period)
        self.remote_releases += it.num_migrate_to_this_cpu_during(self.params.task.period)

    def compute_cost(self, n):
        if self.params.share is None:
            fraction = 0.0
        else:
            fraction = self.params.share / self.params.task.utilization()
        # compute preempt charge
        tick_cost = self.params.ohead.tick(n)
        preempt  = tick_cost + self.params.cpmd + \
            2 * (self.params.ohead.context_switch(n) +
                 self.params.ohead.sched(n))

        rel_cost = self.params.ohead.release(n)

        # Each of these has two costs, depending on whether the predecessor
        # was tardy.  Assume more pessimistic overhead in either case.
        if self.params.migrating:
            my_tm_cost = max(self.params.ohead.timer_migration(n),
                              self.params.ohead.ipi_latency(n))
        else:
            my_tm_cost = 0
        other_tm_cost = max(self.params.ohead.timer_migration(n),
                            self.params.ohead.sched(n))
        # compute inflated utilization
        cost = (self.params.task.cost + preempt +
                self.rel_irqs * rel_cost +
                my_tm_cost +
                self.remote_releases * other_tm_cost)

        cost = account_for_timer_tick(cost, TIMER_PERIOD, tick_cost,
                                      self.params.task.period)
        self.inflated_cost = int(ceil(cost))
        self.share = fraction * self.inflated_cost \
                     / float(self.params.task.period)

    def update_size(self, it, bin):
        self.inc_size(it)
        self.compute_cost(len(bin.items) + 1) # number of tasks in the partition

    def determine_size(self, bin):
        n = len(bin.items) + 1

        # compute interrupt charge
        self.remote_releases  = 0
        self.rel_irqs = 1 # charge for our own delay
        for it in bin.items:
            self.inc_size(it)

        # estimate cost
        self.compute_cost(n)

    def cost_inflation(self):
        # what we wanted vs. what we cost
        return self.inflated_cost - self.params.task.cost

    def inflated_utilization(self):
        cost_infl = self.cost_inflation()
        # add that into the task's utilization
        return inflated_utilization(self.params.task, cost_infl)

    def __str__(self):
        return "InflatedTask(%s, %s, %s)" % (self.params.task, self.params, self.share)

def migrating_task_overutilized(it):
    return (it.cost_inflation() + it.params.task.cost > it.params.task.period)

def migrating_overutilized(cpu):
    # Given new overheads, check if the CPU has an overutilized migrating task
    for it in cpu.migrating:
        if migrating_task_overutilized(it):
            return True
    return False

def try_to_assign_split(task, ohead, bins, cpu_start, cpu_end):
    all_cpus = range(cpu_start, cpu_end + 1)
    already_found_first_cpu = False
    commits_to_make = []
    remaining_share = task.utilization()
    for cpu_idx in all_cpus:
        cpu = bins[cpu_idx]
        other_cpus = [x for x in all_cpus if x != cpu_idx]
        share_bound = min(remaining_share, cpu.spare_capacity())
        failed = False
        if not already_found_first_cpu:
            # Perform tardiness computation.  Start with guess that assumes no
            # other migrating task.
            task.tardiness = task.cost - task.period
            done = False
            while not done:
                it = inflated_task(task, ohead, cpu_idx, other_cpus,
                                   share_bound, 0)
                if not cpu.try_to_add_no_commit(it):
                    done = True
                    failed = True
                elif migrating_overutilized(cpu):
                    done = True
                    cpu.rollback()
                    failed = True
                # May as well update our bound, since we will cause at least
                # this many overheads.
                share_bound = min(remaining_share,
                                  it.available_share(cpu.spare_capacity()))
                if not failed and len(cpu.migrating) > 0:
                    higher = cpu.migrating[0]
                    higher_share = higher.share
                    higher_cost = higher.inflated_cost
                    task.tardiness = (higher.share * (higher.params.task.tardiness +
                                                      higher.params.task.period) +
                                      3 * higher_cost + it.inflated_cost) \
                                     / (1 - higher_share) - task.period
                    cpu.rollback()
                    it = inflated_task(task, ohead, cpu_idx, other_cpus,
                                       share_bound)
                    if not cpu.try_to_add_no_commit(it):
                        done = True
                        failed = True
                    elif migrating_overutilized(cpu):
                        done = True
                        cpu.rollback()
                        failed = True
                    share_bound = min(remaining_share,
                                      it.available_share(cpu.spare_capacity()))
                    # If we increased the cost of the higher-priority task, our
                    # bounds may no longer be valid, so we'll have to iterate
                    # again.  Because of the ceil, this is just fixed-point
                    # iteration.
                    if (higher_share >= higher.share and
                            higher_cost >= higher.inflated_cost):
                        done = True
                        if (not failed) and (migrating_task_overutilized(it)):
                            cpu.rollback()
                            failed = True
                elif not failed:
                    # Initial guess was actually correct
                    done = True
        else:
            # Already have valid tardiness bound, just use it.
            it = inflated_task(task, ohead, cpu_idx, other_cpus, share_bound)
            if not cpu.try_to_add_no_commit(it):
                failed = True
            elif migrating_task_overutilized(it):
                cpu.rollback()
                failed = True
            share_bound = min(remaining_share,
                              it.available_share(cpu.spare_capacity()))
        if (not failed) and share_bound >= MIN_SHARE:
            it.params.share = share_bound
            remaining_share -= share_bound
            it.update_cost(cpu)
            commits_to_make.append((cpu, it))
            # Don't compute tardiness bound again, since won't be first CPU for
            # the task.
            already_found_first_cpu = True
        elif not failed:
            cpu.rollback()
    # Could be less than MIN_SHARE but not zero due to floating point error
    if (remaining_share < MIN_SHARE):
        for (cpu, it) in commits_to_make:
            cpu.commit()
            cpu.migrating.append(it)
        return True
    else:
        for (cpu, it) in commits_to_make:
            cpu.rollback()
        return False

def inflated_task(task, ohead, cpu, other_cpus, share_bound=1, share=None):
    # preemption
    cpmd  = ohead.cpmd(cpu, cpu)
    #print "preempt (%d): %f" % (cpu, cpmd)

    if share is None:
        cost  = task.cost
        if (len(other_cpus) == 0):
            # If there are other CPUs and we don't have a share yet, we need
            # to use share_bound
            share = task.utilization()
    else:
        # extrapolate cost
        cost = int(ceil(share * task.period))
    # create non-changing parameters
    params = storage(task=task, cost=cost, share=share, cpmd=cpmd,
                     share_bound=share_bound, migrating=(len(other_cpus) > 0),
                     ohead=ohead)
    return InflatedTask(params)

# assumes ts has already been converted to us
# ohead is new-style overhead object
def assign_ts_with_overheads(ts, no_cpus, ohead, want_bins=False):
    cpus = [DuctileBin() for _ in range(no_cpus)]
    for i, cpu in enumerate(cpus):
        # keep track of migrating tasks
        cpu.migrating = []
        cpu.id = i

    h = MaxSpareCapacity(initial_bins=cpus)
    inflated_tasks = [inflated_task(t, ohead, 0, []) for t in ts]
    # Decreasing utilization
    inflated_tasks.sort(key = lambda it: it.inflated_utilization(),
                        reverse = True)
    h.binpack(inflated_tasks, report_misfit=ignore) # no error reporting
    sorted_bins = sorted(h.bins, key = lambda b: b.id)
    if h.misfits:
        for b in h.bins:
            if not b.items:
                return None if want_bins else False
        ccpu = 0
        for it in h.misfits:
            c_bin = sorted_bins[ccpu]
            # Small utilization = might fit as fixed after all
            if not c_bin.try_to_add(it):
                ecpu = ccpu + 1
                done = False
                while (not done) and (ecpu < no_cpus):
                    if (try_to_assign_split(it.params.task, ohead, sorted_bins,
                            ccpu, ecpu)):
                        done = True
                        ccpu = ecpu
                    else:
                        ecpu += 1
                if not done:
                    return None if want_bins else False

    # found an assignment
    # Assign tardiness for fixed tasks:
    for b in sorted_bins:
        for it in b.items:
            assert (0 <= len(b.migrating) <= 2)
            if it in b.migrating:
                # Already computed
                pass
            elif len(b.migrating) == 0:
                it.params.task.tardiness = 0
            else:
                higher_share = b.migrating[0].share
                higher_tardiness = b.migrating[0].params.task.tardiness
                higher_period = b.migrating[0].params.task.period
                higher_cost = b.migrating[0].inflated_cost
                if len(b.migrating) == 2:
                    lower_share = b.migrating[1].share
                    lower_tardiness = b.migrating[1].params.task.tardiness
                    lower_period = b.migrating[1].params.task.period
                    lower_cost = b.migrating[1].inflated_cost
                else:
                    lower_share = 0
                    lower_tardiness = 0
                    lower_period = 0
                    lower_cost = 0
                it.params.task.tardiness = (higher_share * (higher_tardiness +
                                                            higher_period) +
                                            3 * higher_cost +
                                            lower_share * (lower_tardiness +
                                                           lower_period) +
                                            3 * lower_cost) / (1 - higher_share
                                                               - lower_share)

    if want_bins:
        return cpus
    else:
        return True

def assign_ts_no_overheads(ts, no_cpus):
    cpus = [DuctileBin() for _ in range(no_cpus)]
    for i, cpu in enumerate(cpus):
        cpu.id = i

        cpu.unused_capacity = Fraction(1)
    bin_items = []
    for t in ts:
        bin_items.append(FixedSizeItem(t, t.utilization()))

    ts.sort(key = lambda t: t.utilization(), reverse = True)
    ccpu = 0
    for i, t in enumerate(ts):
        task_util_frac = Fraction(int(t.cost), int(t.period))
        largest_avail = max(cpus, key = lambda c: c.unused_capacity)
        if (i >= len(ts) - 3):
            #Forced migrating
            segment = 0
            t.fracs = []
            share = Fraction(0)
            remaining_utilization = task_util_frac
            while (remaining_utilization > 0):
                if (segment >= 2):
                    desired_share = remaining_utilization
                else:
                    desired_frac_den = random.randrange(2, 10)
                    desired_frac_num = random.randrange(1, desired_frac_den)
                    desired_share = min(remaining_utilization,
                                        Fraction(desired_frac_num,
                                                 desired_frac_den)
                                        * task_util_frac)
                new_share = min(desired_share, cpus[ccpu].unused_capacity)
                new_fraction = new_share / task_util_frac
                t.fracs.append((new_fraction.numerator,
                                new_fraction.denominator, ccpu))
                remaining_utilization -= new_share
                cpus[ccpu].unused_capacity -= new_share
                if (remaining_utilization > 0) or (cpus[ccpu].unused_capacity
                                                   == 0):
                    ccpu += 1
                    if ccpu >= no_cpus:
                        return None
                segment += 1
        elif largest_avail.unused_capacity >= task_util_frac:
            t.fracs = [(1, 1, largest_avail.id)]
            largest_avail.unused_capacity -= task_util_frac
        else:
            #Needs to be migrating
            t.fracs = []
            share = Fraction(0)
            remaining_utilization = task_util_frac
            while remaining_utilization > 0:
                new_share = min(remaining_utilization,
                                cpus[ccpu].unused_capacity)
                new_fraction = new_share / task_util_frac
                t.fracs.append((new_fraction.numerator,
                                new_fraction.denominator, ccpu))
                remaining_utilization -= new_share
                cpus[ccpu].unused_capacity -= new_share
                if (cpus[ccpu].unused_capacity == 0):
                    ccpu += 1
                    if ccpu >= no_cpus:
                        return None
    return None

def write_assigned_tasks(ts, filename):
    f = open(filename, 'w')
    #wcet period ncpus
    #num denom cpu ...
    for t in ts:
        wline = "%s %s %s\n" % (int(t.cost), int(t.period), len(t.fracs))
        f.write(wline)
        for frac in t.fracs:
            wline = "%s %s %s\n" % frac
            f.write(wline)
    f.close()
