# EDF-WM assignment phase

from math import ceil, floor

from binpack import DuctileItem, DuctileBin, MaxSpareCapacity, ignore

from model.tasks import TaskSystem

from util.storage import storage
from util.seq     import neighbours
from util    import lcm

def account_for_timer_tick(exec_cost, timer_period, tick_cost, max_cost):
    num_ticks = int(ceil(exec_cost / float(timer_period)))
    exec_cost += num_ticks * tick_cost
    last_ticks = num_ticks

    num_ticks = int(ceil(exec_cost / float(timer_period)))
    while num_ticks > last_ticks and exec_cost <= max_cost:
        exec_cost += (num_ticks - last_ticks) * tick_cost
        last_ticks = num_ticks
        num_ticks  = int(ceil(exec_cost / float(timer_period)))

    return exec_cost, num_ticks

class InflatedTask(DuctileItem):
    def __init__(self, params, rel_irqs=1, tm_irqs=0, num_tasks=1, density=None):
        self.params = params

        # params.task  --- reference to the task
        # params.cost  --- cost of the task (could be less than params.task.cost)
        # params.deadline --- (possibly) shortened deadline
        # params.cpmd --- cpmd charge, pre-computed (migratory or preemption)
        # params.slice --- True if this is a slice that causes timer migrations

        # this is individual for each task, since periods differ
        self.rel_irqs   = rel_irqs  # number of release interrupts
        self.tm_irqs    = tm_irqs   # number of timer migrations
        self.num_tasks  = num_tasks # number of other tasks
        if density is None:
            self.compute_density()
        else:
            self.density = density

    def num_releases_during(self, delta):
        return int(ceil(delta / float(self.params.task.period)))

    def num_timer_migrations_during(self, delta):
        if self.params.slice:
            return int(ceil(delta / float(self.params.task.period)))
        else:
            return 0

    def size(self):
        return self.density

    def copy(self):
        return InflatedTask(self.params,
                            self.rel_irqs,
                            self.tm_irqs,
                            self.num_tasks,
                            self.density)

    def inc_size(self, it):
        self.rel_irqs  += it.num_releases_during(self.params.deadline)
        self.tm_irqs   += it.num_timer_migrations_during(self.params.deadline)
        self.num_tasks += 1

    def compute_density(self):
        # compute preempt charge
        tick_cost = self.params.ohead.tick(self.num_tasks)
        preempt  = tick_cost + self.params.cpmd + \
            2 * (self.params.ohead.context_switch(self.num_tasks) +
                 self.params.ohead.sched(self.num_tasks))

        rel_cost = self.params.ohead.release(self.num_tasks)
        tm_cost  = self.params.ohead.timer_migration(self.num_tasks)
        # compute inflated utilization
        cost = (self.params.cost + preempt +
                self.rel_irqs * rel_cost +
                self.tm_irqs * tm_cost)

        if self.params.slice:
            # add in IPI latency
            cost += self.params.ohead.ipi_latency(self.num_tasks)

        cost, self.num_ticks = account_for_timer_tick(
            cost, self.params.ohead.tick_period, tick_cost, self.params.task.period)

        self.inflated_cost = int(ceil(cost))
        self.density = self.inflated_cost / float(self.params.deadline)


    def update_size(self, it, bin):
        self.inc_size(it)
        self.compute_density()

    def determine_size(self, bin):
        n = len(bin.items) + 1

        # compute interrupt charge
        self.num_tasks = 1 # us
        self.tm_irqs   = 0
        self.rel_irqs  = 1 # charge for our own delay
        for it in bin.items:
            self.inc_size(it)

        # estimate cost
        self.compute_density()

    def __str__(self):
        return "InflatedTask(%s, %d, %d, %d, %.2f, %.2f, slice=%s)" % (
            self.params.task,
            self.inflated_cost,
            self.params.cost,
            self.params.deadline,
            self.params.cost / float(self.params.deadline),
            self.size(), self.params.slice)


def inflated_task(task, ohead, cpu, is_slice=False, prev_cpu=None,
                  cost=None, deadline=None):
    # preemption
    cpmd  = ohead.cpmd(cpu, cpu)
    # 2nd charge --- preemption or migration
    if prev_cpu is not None:
        cpmd += ohead.cpmd(prev_cpu, cpu)

    if cost is None:
        cost  = task.cost
    if deadline is None:
        deadline = task.deadline
    # create non-changing parameters
    params = storage(task=task, cost=cost, deadline=deadline, cpmd=cpmd,
                     slice=prev_cpu is not None, ohead=ohead)
    return InflatedTask(params)


def assign_ts_with_overheads(ts, no_cpus, ohead, want_bins=False):
#    print 'EDF-WM', no_cpus, ts.utilization()
    cpus = [DuctileBin() for _ in range(no_cpus)]
    for i, cpu in enumerate(cpus):
        cpu.id = i
    for (x, y) in neighbours(cpus):
        y.prev_cpu = x
    cpus[0].prev_cpu = None

    h = MaxSpareCapacity(initial_bins=cpus)
    # XXX assumes preemptions have the same cost everywhere---true for Ludwig.
    inflated_tasks = [inflated_task(t, ohead, 0) for t in ts]
    # decreasing density
    inflated_tasks.sort(key = lambda it: it.size(), reverse = True)
    h.binpack(inflated_tasks, report_misfit=ignore) # no error reporting
    if h.misfits:
        # some didn't make it => let's split them
        # ...but only if there aren't any empty procs
        # (empty procs == can't fit by itself => infeasible)
        for b in h.bins:
            if not b.items:
                return None if want_bins else False
            else:
                b.debug_len = len(b.items)
        for it in h.misfits:
            if not try_to_split(it.params.task, h.bins, ohead):
                # => not schedulable
                #print "Done => No."
                return None if want_bins else False

    # great, everything fit

#    for b in h.bins:
#        print "CPU%d" % b.id
#        for it in b.items:
#            print it
#        sc = b.spare_capacity()
#        uc = sum([it.params.cost / float(it.params.task.period) for it in b.items])
#        print 'sc', sc
#        print 'u', uc, sc + uc
#        print '=' * 10
    for t in ts:
        t.tardiness = 0
#    print "Done => Yes."
    return h.bins if want_bins else True


### Here comes the nasty part...

def try_to_split(task, cpus, ohead):
    num_slices = 2
    # start a transaction on each proc
    for cpu in cpus:
        cpu.prepare()
    while num_slices <= len(cpus):
        rel_deadline = int(floor(task.deadline / float(num_slices)))
#        print 'try_to_split', task, '->', num_slices, '=>', rel_deadline

        prev_id = None
        best = []
        remaining_cpus = cpus
        for _ in range(num_slices):
            new_slices = [alloc_max_slice(cpu, task, rel_deadline, ohead,
                                          prev_id) for cpu in remaining_cpus]
            best_slice = max(new_slices, key=lambda it: it.params.cost)
            for bad_slice in [s for s in new_slices if s != best_slice]:
                bad_slice.cpu.rollback()
                bad_slice.cpu.prepare()
            best.append(best_slice)
            remaining_cpus = [cpu for cpu in remaining_cpus
                              if cpu != best_slice.cpu]
            prev_id = best_slice.cpu.id
        total = sum([it.params.cost for it in best])
        if total >= task.cost:
            # we got ourselves a nice split
            for slice in best:
                # accept the ones that we keep
                slice.cpu.commit()
            # update last slice
            complete = sum([it.params.cost for it in best[:-1]])
            # make sure slice size matches
            assert it.params.task.cost - complete <= best[-1].params.cost
            best[-1].params.cost -= it.params.task.cost - complete
            best[-1].compute_density()
            # we are done here
            break
        else:
            # nope, try again
            for slice in best:
                slice.cpu.rollback()
                slice.cpu.prepare()
            num_slices += 1

#    if num_slices <= len(cpus):
#        print "Split %s into %d pieces" % (it, num_slices)

    return num_slices <= len(cpus)


def dbf(it, t):
    """Demand bound function for slice it  at time t."""
    if t <= 0:
        return 0
    return  max(0,
                (floor((t - it.params.deadline) / float(it.params.task.period)) + 1) *
                it.params.cost)


def dbf_points_of_change(it, max_t):
    """Return iterator over t where dbf(tsk, t) changes."""
    yield 0
    t = it.params.deadline
    while t <= max_t:
        yield t
        t += it.params.task.period

def alloc_max_slice(cpu, task, rel_deadline, ohead, prev_cpu):
    tsk_it = inflated_task(task, ohead, cpu.id, True, prev_cpu, cost=0,
                           deadline=rel_deadline)
    tsk_it.cpu = cpu

    cpu.temporary_assign(tsk_it)
    extra_cost = tsk_it.inflated_cost # the overheads added to the budget (does not include ticks)

    # skip new split task
    already_allocated = cpu.items[:-1]

    usum = 0.0
    for it in already_allocated:
        usum += (it.inflated_cost / float(it.params.task.period))
    initial_cost = int(floor((1 - usum) * task.period))

    # it also can't be more than the rel_deadlie (otherwise, density > 1)
    cost = min(initial_cost, rel_deadline)

    # hyperpriod
    hp = lcm(*[it.params.task.period for it in cpu.items])
    # Since cost can't possibly become larger than what we assume now,
    # we can also compute an upper bound based on L_a
    la = new_interval_end(already_allocated, cost, rel_deadline, it.params.task)

#    print 'hp', hp, 'la', la
    max_t = min(hp, la)

    all_points = [(dbf_points_of_change(it, max_t), it) for it in cpu.items]
    for point, it in all_points:
        if cost <= extra_cost:
#            print 'early no slack exit', cost, extra_cost
            break
        try:
            L = point.next()
            while L < rel_deadline:
                L = point.next()
            while L < max_t:
#                print 'testing', L, 'max_t', max_t
                demand = sum([dbf(x, L) for x in already_allocated])
                new_cost = compute_max_cost(demand, L, it.params.task, rel_deadline)
#                if not new_cost <= rel_deadline:
#                    print [str(x) for x in cpu.items]
#                    print 'XXXXXX', new_cost, rel_deadline, cost
#                    print 'dbf', demand, 'L', L, 'per', it.params.task.period, 'dl', rel_deadline
#                    print it.params.task, str(it)
#                    assert False
                if new_cost < cost:
#                    assert new_cost <= rel_deadline
                    # new upper bound
                    cost = max(new_cost, 0)
                    interval_end = min(max_t,
                                       new_interval_end(already_allocated, cost,
                                                        rel_deadline,
                                                        it.params.task))
#                    print 'improved max_t <-', interval_end, 'from', max_t
                    max_t = min(max_t, interval_end)
                if cost <= extra_cost or L + it.params.task.period >= max_t:
                    break # determined max. cost
                else:
                    # go to next testpoint
                    L = point.next()
        except StopIteration:
            pass # move on the next point

    assert cost <= rel_deadline


    # Take off what we need for overheads. This is a bit tricky since the
    # number of tick IRQs depend on the execution budget, and the execution
    # budget depends on the number of ticks.
    high = cost - extra_cost
    tsk_it.params.cost = high
    count = 0
    max_ticks = None
    # print 'tick-period=%d tick-cost=%d' % (ohead.tick_period, int(ceil(ohead.tick(0))))
    while True:
        count +=1
        tsk_it.compute_density()
        # print tsk_it
        # print "tsk_it.params.cost=%d => density=%f" % (tsk_it.params.cost, tsk_it.size())
        if tsk_it.size() <= 1:
            # If this creates a negative cost, then try_to_split will filter &
            # and rollback this CPU anyway.
            break
        # need to reduce cost
        # print "num_ticks=%d tick_cost=%f" % (tsk_it.num_ticks, ohead.tick(0) * tsk_it.num_ticks)
        # print "tick periods = %d" % tick_periods
        if max_ticks is None:
            max_ticks = tsk_it.num_ticks
        max_ticks -= 1
        tick_periods = max_ticks * ohead.tick_period
        # try one less tick
        tsk_it.params.cost = tick_periods - extra_cost - 1
        # print "tsk_it.params.cost = %d" % tsk_it.params.cost
    return tsk_it

def compute_max_cost(demand, L, tsk, rel_deadline):
    # XXX floating point / integer correct here?
    return int((L - demand) / (floor((L - rel_deadline) / float(tsk.period)) + 1))

# Equation (9) # stolen from andrea's impl.
def new_interval_end(items, cost, rel_deadline, tsk):
    max_d = max(t.params.deadline for t in items)
    # with an implitcit-periodic model, this will cancel.
    wdiff_p = sum(((t.params.task.period - t.params.deadline) * t.params.task.utilization()) \
                   for t in items)
    u_split = cost / float(tsk.period)
    # if we fill the cpu exactly, La denominator will become 0
    usum = sum([it.params.task.utilization() for it in items])
    if (u_split == (1 - usum)):
        return max_d
    else:
        return max(((wdiff_p + (tsk.period - rel_deadline) * u_split ) / \
                        (1 - usum - u_split)), max_d)
