"""
NPS-F Semi-partitioned scheduling algorithm

Bletsas and Andersson, Preemption-light multiprocessor scheduling of sporadic
tasks with high utilization bound, 30th RTSS, 2009
"""
from math import floor, ceil

from binpack import DuctileBin, DuctileItem, BasicBin, Bin, FirstFit, DidNotFit, report_failure, GlobalConstraintBin

from model.tasks import TaskSystem

from util.quantor import forall
from util.seq     import neighbours
from util.storage import storage

# ############# Basic concepts from the paper #############

def npsf_slot_length(ts, delta):
    """How long is one slot for a given task system?"""
    return int(floor(ts.min_deadline() / float(delta)))

def npsf_inflate(usum, delta):
    """How much budget is required such that tasks with usum utilization
    will all meet their deadllines for a given delta?"""
    return (usum * (delta + 1)) / (usum + delta)

def npsf_deflate(budget, delta):
    """What's the utilization that we can guarantee with a budget-length
    reservation?"""
    return (delta * budget) / ((delta + 1) - budget)

# ############# Non-clustered Phase 1 #################

def make_size_phase1(ohead, n, max_cpmd):
    """During phase 1, we don't know yet if a job will be migrating.  Thus, we
    charge overheads as if it were migrating. When the actual physical
    assignment occurs, then overheads will be recomputed and this pessimism
    removed."""
    cxsw  = ohead.context_switch(n)
    sched = ohead.sched(n)
    ipi   = ohead.ipi_latency(n)
    # standard inflation of each job during phase 1
    preempt = 2 * (cxsw +  sched + max_cpmd) + ipi
    def size_phase1(task):
        return (task.cost + preempt) / float(task.period)
    return size_phase1

def estimate_switch_cost(ohead, n, max_cpmd):
    cxsw  = ohead.context_switch(n)
    sched = ohead.sched(n)
    ipi   = ohead.ipi_latency(n)
    # Upper bound on switch cost: a notional processor is split across two
    # physical processors, so *each* allocation incurs scheduling and context
    # switch charges and the CPMD charge. Further, the continuing reserve is
    # charged with one IPI and an additional context switch and scheduling decision.
    return 3 * (cxsw + sched) + ipi + 2 * max_cpmd

def max_notional_processor_capacity(n, ohead, max_cpmd, delta, slot_length):
    """The maximum size notional processor that does not overload a slot if
    it has to migrate."""
    sched = ohead.sched(n)
    cxsw  = ohead.context_switch(n)
    ipi   = ohead.ipi_latency(n)
    budget = floor(slot_length - 3 * (sched + cxsw) - ipi - 2 * max_cpmd)
    cap = npsf_deflate(budget / float(slot_length), delta)
#    print "XXX => %.2f (%.2f)" % (cap, cap * slot_length)
    return cap


class NotionalProcessor(GlobalConstraintBin):
    """A notional processor is bin with maximum capacity that decreases as
    items are added since we need to leave room for interrupts. Optionally,
    a global constraint is enforced (this is needed for clustering)."""

    def __init__(self, irq_util_loss, *args, **kargs):
        GlobalConstraintBin.__init__(self, *args, **kargs)
        self.irq_util_loss = irq_util_loss

    def capacity(self):
        # Account for the interrupts caused by the tasks in this notional
        # processors.  This is tricky: we don't count the irq cost in each
        # server as allocated utilization (this would be very pessimistic),
        # instead we reduce the capcity of each bin. If we do not do this, then
        # we might end up creating notional processors that are so tightly
        # packed that they will not fit on a processor by themselves once
        # interrupts are accounted for with the IRQLoss object during the
        # mapping to physical processors.
        return self.max_capacity - self.irq_util_loss(len(self.items))


def binpack_phase1(tasks, ohead, max_cpmd, delta, slot_length):
    """First step of the (non-clustered) assignment: binpack task system into
    as many unit-capacity bins as needed using first-fit-decreasing."""
    size_phase1 = make_size_phase1(ohead, len(tasks), max_cpmd)
    tasks.sort(key=size_phase1, reverse=True)
    cap = max_notional_processor_capacity(len(tasks), ohead, max_cpmd, delta,
                                          slot_length)
    irq_util_loss = make_irq_utilization_loss(ohead, slot_length)
    def make_bin():
        return NotionalProcessor(irq_util_loss, initial_items=TaskSystem(),
                                 size=size_phase1, capacity=cap)
    h = FirstFit(make_bin=make_bin)
    h.binpack(tasks, report_misfit=report_failure)
    return h.bins


# ############# Interrupt Accounting & Mapping #############

def irq_cost(ohead, num_tasks, slot_length, tick_period):
    """How much time is lost to IRQs (at most) during one slot?"""
    tick    = ohead.tick(num_tasks)
    release = ohead.release(num_tasks)
    num_ticks = int(ceil(slot_length / float(tick_period)))
#    print "tx", num_ticks, '->', tick * num_ticks
#    print 'tt', num_tasks, '->', num_tasks * release
    cost = num_tasks * release + tick * num_ticks
    return cost

def make_irq_utilization_loss(ohead, slot_length):
    # capture constant parameters
    def irq_util_loss(num_tasks):
        """What is the fraction of the slot that we lose to IRQs?"""
        return (irq_cost(ohead, num_tasks, slot_length, ohead.tick_period)
                / float(slot_length))
    return irq_util_loss

NO_TASKS = frozenset()


class IRQLoss(DuctileItem):
    """Ductile item in each "slot". Expands to soak up time lost to IRQs. """
    def __init__(self, params, num_tasks=0):
        self.params = params
        # params.ohead --- how to compute overheads
        # params.slot_length --- how long is a slot?
        self.num_tasks = num_tasks
        self.compute_cost()
        # compatability with Reservation items in the
        # same bin. This reservation serves to "soak up"
        # lost capacity and thus does not host any
        # tasks.
        self.params.tasks = NO_TASKS

    def size(self):
        return self.cost

    def copy(self):
        return IRQLoss(self.params, self.num_tasks)

    def compute_cost(self):
        self.cost = irq_cost(
            self.params.ohead, self.num_tasks,
            self.params.slot_length, self.params.ohead.tick_period)
        self.cost = int(ceil(self.cost))

    def update_size(self, it, bin):
        # count how many tasks execute on this server
        self.num_tasks += len(it.params.tasks)
        self.compute_cost()

    def determine_size(self, bin):
        for res in bin.items:
            self.num_tasks += len(res.params.tasks)
        self.compute_cost()

    def __str__(self):
        return "IRQLoss(%s, %s, %s)" % (self.params.slot_length, self.cost, self.num_tasks)


class Reserve(DuctileItem):
    """Reserve Item. Knows about how to account for its own overheads.
    Wraps notional processors including inflate() and overhead accounting.
    """
    def __init__(self, params, num_tasks = 0, budget = None):
        self.params = params
        self.num_tasks = len(params.tasks) if num_tasks == 0 else num_tasks
        if budget is None:
            self.compute_budget()
        else:
            self.budget = budget

    def size(self):
        return self.budget

    def copy(self):
        return Reserve(self.params, self.num_tasks, self.budget)

    def compute_budget(self):
        # this is how much the client wants
        client_time = self.params.budget
        # add in overheads and delays incurred at the beginning of the
        # reserve switch
        sched = self.params.oheads.sched(self.num_tasks)
        ctxsw = self.params.oheads.context_switch(self.num_tasks)
        if self.params.continuing:
            # continuing reserves must wait for the previous CPU to signal
            # that it is save to switch, this causes extra IPI latency
            ipi = self.params.oheads.ipi_latency(self.num_tasks)
            switch_time = 2 * (sched + ctxsw) + ipi + self.params.cpmd
        else:
            switch_time = sched + ctxsw + self.params.cpmd

        self.budget = int(ceil(client_time + switch_time))

    def update_size(self, it, bin):
        self.num_tasks += len(it.params.tasks)
        self.compute_budget()

    def determine_size(self, bin):
        for res in bin.items:
            self.num_tasks += len(res.params.tasks)
        self.compute_budget()

    def __str__(self):
        return "Reserve(%s, %s, %s)" % (self.budget, self.params.budget, self.num_tasks)


def make_reserve(
    tasks,         # The tasks assigned to this notional server.
    oheads,        # Runtime overheads.
    cpus,          # tuple of the form (from, to)
    continuing,    # True if this a continuing reserver (One with
    # IPI delay)
    slot_length,   # How long is each scheduling slot?
    delta,         # The delta used for coming up with the slot size.
    budget=None):  # used to override tasks.utilization()
    cpmd = oheads.cpmd(cpus[0], cpus[1])
    # figure out how much to allocate on this server
    if budget is None:
        budget = int(ceil(npsf_inflate(tasks.utilization(), delta) * slot_length))

    params = storage(oheads      = oheads,
                     budget      = budget,
                     continuing  = continuing,
                     slot_length = slot_length,
                     delta       = delta,
                     tasks       = tasks,
                     cpmd        = cpmd)
    return Reserve(params)


def make_bins_for_cpus(procs, ohead, slot_length):
    cpus = []
    for cpu_id in procs:
        cpu = DuctileBin(capacity=slot_length)
        # keep track of CPU ids to figure out migration costs (L1, L2, L3)
        cpu.id = cpu_id
        # add a fake reserve that soaks up IRQ time
        cpu.add(IRQLoss(storage(slot_length=slot_length, ohead=ohead)))
        cpus.append(cpu)
    return cpus

# ############# FLAT MAPPING ###############

def assign_notional_processors(tasksets, cpus, slot_length,
                               delta, oheads, want_bins):
    fail = None if want_bins else False
    idx = 0
    cpu = cpus[0]

    # Try to place every taskset (= notional processor).
    for ts in tasksets:
        res = make_reserve(ts, oheads, (cpu.id, cpu.id),
                           False, slot_length, delta)
        required = res.params.budget
        if cpu.infeasible_to_fit(res) or not cpu.try_to_add(res):
            # Does not fit => let's try to split it.
            idx += 1

            if idx == len(cpus):
                #print "can't place", res
                return fail

            nxt_cpu = cpus[idx]

            # try to place an empty reservation
            res = make_reserve(ts, oheads, (cpu.id, nxt_cpu.id),
                               False, slot_length, delta, budget=1)
            if cpu.try_to_add(res):
                # almost empty fits => figure out how much we can allocate
                alloc = 1 + cpu.spare_capacity()
                # allocated a slice, go for it
                #print "alloc:", alloc, "req", required, cpu.id, '->', nxt_cpu.id
                #print "pre:", oheads.cpmd(cpu.id, cpu.id), "mig:", oheads.cpmd(cpu.id, nxt_cpu.id)
                #assert alloc < required # otherwise, why would it fail?
                res.params.budget = alloc
                res.compute_budget()
                assert cpu.spare_capacity() == 0  # otherwise, where did come from?
            else:
                # empty wouldn't even fit, allocate to next
                alloc = 0
            # allocate remaining

            res = make_reserve(ts, oheads,
                               (cpu.id, nxt_cpu.id) if alloc
                               else (nxt_cpu.id, nxt_cpu.id),
                               alloc != 0,
                               slot_length, delta,
                               budget = required - alloc)
            cpu = nxt_cpu
            if not cpu.try_to_add(res):
                #assert res.size() > 0 # it should be fitting unless overutilized
                #print "is overtutilized", res, res.params.tasks
                return fail
#        else:
#            # was added ok
#            assert res in cpu.items

    # found an assignment
    # We must make sure that both halfs of a split server do not
    # exceed the slot length (including overheads); they  would overlap
    # in that case.

#    for cpu in cpus:
#        for res in cpu.items[1:]:
#            print "overhead (%d, %d) -> %.2f" % (res.params.budget, res.size(), res.size() /  float(res.params.budget))
#            assert res.size() <= slot_length

    # assumption: split servers should always be the first and the last in
    # their resp. bins.
    # for (cpu1, cpu2) in neighbours(cpus):

    #     if len(cpu1.items) > 1 and len(cpu2.items) > 1:
    #         last  = cpu1.items[-1]
    #         # skip over the initial IRQ reserve
    #         first = cpu2.items[1]
    #         if first.params.tasks == last.params.tasks:
    #             assert first.params.continuing
    #             if first.size() + last.size() > slot_length:
    #                 print "XXXX BAD: slot length violated across both reservations", last, first, slot_length, last.size() + first.size(), last.params.budget + first.params.budget
    #                 return fail
    #         else:
    #             assert not first.params.continuing
    #         # this must be a split reserve

    return cpus if want_bins else True



def assign_ts_with_overheads(tasks, num_procs, delta,
                             oheads,    # SystemOverheads()
                             max_cpmd,  # what's the maximum migration cost?
                             want_bins=False):
    """Implementation of the NPS-F flat mapping assignment"""
    fail = None if want_bins else False
    slot_length = npsf_slot_length(tasks, delta)
    ## Phase 1: bin-pack into as many notional processors as needed
    try:
        bins = binpack_phase1(tasks, oheads, max_cpmd, delta, slot_length)
    except DidNotFit as dnf:
        # This can happen if the max. capacity of notational servers became too
        # small. In this case we can't do much; if we allow greater servers than
        # we may not be able to assign them to physical processors later on.
        return fail

    tasksets = [b.items for b in bins]

    ## Phase 2: map to physical processors
    # First make some processor bins.
    cpus = make_bins_for_cpus(range(num_procs), oheads, slot_length)

    return assign_notional_processors(tasksets, cpus, slot_length,
                                      delta, oheads, want_bins)

# ################### Clustered NPS-F support ########################


class Cluster(BasicBin):
    """A cluster is a bin consisting of multiple processors."""

    def __init__(self, num_procs, size, cap, delta, switch_util, irq_util):
        def make_bin():
            return NotionalProcessor(irq_util,
                                     global_validate=self.not_overutilized,
                                     initial_items=TaskSystem(), size=size,
                                     capacity=cap)

        self.ff = FirstFit(make_bin=make_bin)
        self.num_procs   = num_procs
        self.switch_util = switch_util
        self.irq_util    = irq_util
        self.delta       = delta

    def total_inflated_utilization(self):
        usum = len(self.ff.bins) * self.switch_util
        n = 0
        for np in self.ff.bins:
            # Make sure we use the correct size() function, not just
            # np.items.utilization() (which isn't aware of overheads).
            task_util = np.allocated_capacity()
            # Note: this is where the fact that NotionalProcessors "shrink" as
            # tasks are added is used. If we were to just count IRQ loss as
            # allocted utilization, then usum below would be quite pessimistic.
            usum += npsf_inflate(task_util, self.delta)
            n += len(np.items)
        usum += self.num_procs * self.irq_util(n)
        return usum

    def not_overutilized(self):
        usum = self.total_inflated_utilization()
#        print "utilization check:", usum, self.num_procs
        return self.num_procs >= usum

    # nothing to prepare

    def temporary_assign(self, reserve):
        self.candidate = reserve

    def validate(self):
        # Here we do the actual work. We don't actually support rollback
        # since we don't update the state of the cluster if the item doesn't fit.
        self.ff.binpack([self.candidate])
        if self.ff.misfits:
            self.ff.misfits.pop()
            self.last_ok = False
            # chop empty bin, if any
            if not self.ff.bins[-1].items:
                del self.ff.bins[-1]
            return False
        else:
            self.last_ok = True
            return True

    # nothing to rollback
    def rollback(self):
        assert not self.last_ok

    def commit(self):
        assert self.last_ok


def clustered_assign_ts_with_overheads(
    tasks, num_procs, cluster_size, delta,
    oheads,    # SystemOverheads()
    max_cpmd,  # what's the maximum migration cost?
    want_bins=False):

    fail = None if want_bins else False
    slot_length = npsf_slot_length(tasks, delta)

    n    = len(tasks)
    size = make_size_phase1(oheads, n, max_cpmd)
    cap  = max_notional_processor_capacity(n, oheads, max_cpmd, delta,
                                           slot_length)
    switch_cost = estimate_switch_cost(oheads, n, max_cpmd)

#    print 'delta', delta, 'slot_length', slot_length

    irq_util_loss = make_irq_utilization_loss(oheads, slot_length)

    num_clusters = num_procs / cluster_size
    clusters = [Cluster(cluster_size, size, cap, delta,
                        switch_cost / float(slot_length),
                        irq_util_loss)
                for _ in xrange(num_clusters)]

#    print 'clust', n, cap, num_procs, cluster_size, num_clusters, switch_cost / float(slot_length)

    ff = FirstFit(initial_bins = clusters)

    ## Phase 1: bin-pack into clusters
    # Clustering adds another phase.
    # A task is added to a cluster if there exists a notional processor
    # within that cluster that can take the task without crossing
    # the total utilization bound.
    try:
        tasks.sort(key=size, reverse=True)
        ff.binpack(tasks, report_failure)
    except DidNotFit as dnf:
#        print "XXX could not find cluster for", dnf.item
        return fail

#    print 'assigned to clusters'

    ## Phase 2: map to physical processors
    assigned_bins = []
    for i, cluster in enumerate(clusters):
        cpus = make_bins_for_cpus(range(i * cluster_size, (i + 1) * cluster_size),
                                  oheads, slot_length)
        tasksets = [np.items for np in cluster.ff.bins if np.items]
        bins = assign_notional_processors(tasksets, cpus, slot_length,
                                          delta, oheads, want_bins)
        if bins == fail:
#            print 'fail', i, cluster.total_inflated_utilization(), len(cpus)
            return fail
        else:
#            print 'ok'
            assigned_bins.append(bins)

    return assigned_bins if want_bins else True



def test1():
    from generator.tasksets import uni_light_250_hp, DIST
    from exp.eurosys11 import make_wss_cost_function, linear_shared_cache
    from overheads import Overheads, SystemOverheads, CPMD
    from util import ms2us, load_random_state

    # make it predictable
    load_random_state('tests/prng.state')

    ohead_cnpsf = Overheads.from_file('ohead-data/rtas11_C-NPS-F_avg.csv')
    load_cpmd = CPMD.from_file('ohead-data/pmo-ospert10-load-avg.csv')
    load_cost = make_wss_cost_function(load_cpmd, linear_shared_cache)

    gen   = DIST['exp-10-10-100_hp']
    ucap  = 16
    cpus  = 24
    wss   =  0
    delta =  4

    ts = gen(ucap, None)
    so = SystemOverheads(reference=ohead_cnpsf, n=len(ts),
                         migration_cost=load_cost(wss))
    so.tick_period = ms2us(1)

    ts.make_integral_us()

    print "load_cpmd.L3:", load_cpmd.L3(wss)

    bins = clustered_assign_ts_with_overheads(ts, cpus, 6, delta, so, load_cpmd.L3(wss))
#    bins = assign_ts_with_overheads(ts, cpus, delta, so, load_cpmd.MEM(wss))

    print "ucap=%s n=%s bins=%s" % (ucap, len(ts), bins)

if __name__ == '__main__':
    test1()
