#!/usr/bin/env python

import unittest
import overheads as o

class MonoticInterpolation(unittest.TestCase):
    def setUp(self):
        self.data1 = [(1, 2.0), (2, 3.0), (10, -1.0), (11, 19.0)]
        self.data2 = [(1, 2.0), (2, -3.0), (10, -1.0), (11, 2.0)]

        self.data3 = zip(range(25),
                         [0.00, 5.60, 4.82, 4.58, 4.54, 4.79, 4.29,
                          4.16, 4.09, 4.21, 3.92, 4.36, 3.72, 4.67,
                          3.87, 3.85, 3.45, 5.18, 3.83, 3.39, 3.30,
                          3.21, 2.64])

    def test_make_monotonic(self):
        filtered = o.make_monotonic(self.data1)
        self.assertEqual(4, len(filtered))
        self.assertEqual((1,2), filtered[0])
        self.assertEqual((10, 3), filtered[2])

    def test_make_single(self):
        filtered = o.make_monotonic(self.data2)
        self.assertEqual(1, len(filtered))
        self.assertEqual((1,2), filtered[0])

    def test_pwlin_monotonic(self):
        f = o.monotonic_pwlin(self.data1)
        self.assertAlmostEqual(2.5, f(1.5))
        self.assertAlmostEqual(0, f(-1))
        self.assertAlmostEqual(19 + 16 / 1.0, f(12))
        self.assertFalse(f.is_constant())

    def test_pwlin_single(self):
        f = o.monotonic_pwlin(self.data2)
        self.assertAlmostEqual(2.0, f(1.5))
        self.assertAlmostEqual(2.0, f(-1))
        self.assertAlmostEqual(2.0, f(12))
        self.assertTrue(f.is_constant())

    def test_data3(self):
        filtered = o.make_monotonic(self.data3)
        self.assertEqual([(0, 0.00), (1, 5.60)], filtered[0:2])
        self.assertEqual([(0, 0.00), (1, 5.60), (22, 5.60)], filtered[0:3])
        self.assertEqual(3, len(filtered))

class CPMD(unittest.TestCase):
    def setUp(self):
        self.load_wc  = o.CPMD.from_file('ohead-data/pmo-ospert10-load-wc.csv')
        self.load_avg = o.CPMD.from_file('ohead-data/pmo-ospert10-load-avg.csv')
        self.idle_wc  = o.CPMD.from_file('ohead-data/pmo-ospert10-idle-wc.csv')
        self.idle_avg = o.CPMD.from_file('ohead-data/pmo-ospert10-idle-avg.csv')

    def init_ok(self):
        x = o.CPMD()
        self.assertTrue(x.mem_hierarchy[o.CPMD.MEM].is_constant())
        self.assertTrue(x.mem_hierarchy[o.CPMD.L1].is_constant())
        self.assertTrue(x.mem_hierarchy[o.CPMD.L2].is_constant())
        self.assertTrue(x.mem_hierarchy[o.CPMD.L3].is_constant())

    def test_loaded_ok(self):
        self.assertFalse(self.load_wc.mem_hierarchy[o.CPMD.MEM].is_constant())
        self.assertFalse(self.load_wc.mem_hierarchy[o.CPMD.L1].is_constant())
        self.assertFalse(self.load_wc.mem_hierarchy[o.CPMD.L2].is_constant())
        self.assertFalse(self.load_wc.mem_hierarchy[o.CPMD.L3].is_constant())

        self.assertFalse(self.load_avg.mem_hierarchy[o.CPMD.MEM].is_constant())
        self.assertFalse(self.load_avg.mem_hierarchy[o.CPMD.L1].is_constant())
        self.assertFalse(self.load_avg.mem_hierarchy[o.CPMD.L2].is_constant())
        self.assertFalse(self.load_avg.mem_hierarchy[o.CPMD.L3].is_constant())

        self.assertFalse(self.idle_avg.mem_hierarchy[o.CPMD.MEM].is_constant())
        self.assertFalse(self.idle_avg.mem_hierarchy[o.CPMD.L1].is_constant())
        self.assertFalse(self.idle_avg.mem_hierarchy[o.CPMD.L2].is_constant())
        self.assertFalse(self.idle_avg.mem_hierarchy[o.CPMD.L3].is_constant())

        self.assertFalse(self.idle_wc.mem_hierarchy[o.CPMD.MEM].is_constant())
        self.assertFalse(self.idle_wc.mem_hierarchy[o.CPMD.L1].is_constant())
        self.assertFalse(self.idle_wc.mem_hierarchy[o.CPMD.L2].is_constant())
        self.assertFalse(self.idle_wc.mem_hierarchy[o.CPMD.L3].is_constant())


    def test_select_value(self):
        self.assertEqual(19.05, self.load_avg.L1(256))
        self.assertEqual(30.03, self.load_avg.L1(1024))
        self.assertEqual(120.47, self.load_avg.L3(128))
        self.assertEqual(19.05, self.load_avg.L1(512))

if __name__ == '__main__':
    unittest.main()
