# assorted sequence helpers

from itertools    import ifilter, izip, count, chain, repeat
from heapq        import heapify, heappop, heappush
from collections  import defaultdict

def listify(f):
    "wrap return value of f into list"
    def listified(*args, **kargs):
        return [f(*args, **kargs)]
    return listified

def flatten(lst):
    "make a list of lists into a single list"
    res = []
    for x in lst:
        if type(x) == list:
            res += x
        else:
            res.append(x)
    return res

def neighbours(lst):
    # Could be implemented more succintly with izip and drop, but then it
    # wouldn't work for iterators, but only for lists. This way, we
    # call next() only once.
    it = iter(lst)
    prev = it.next()
    while True:
        cur = it.next()
        yield (prev, cur)
        prev = cur

def itimes(it, f=lambda _:1):
    return chain(*[repeat(x, f(x)) for x in iter(it)])

def iskip(it, n=0):
    for i,x in izip(count(), iter(it)):
        if i >= n:
            yield x

def iremove(it, bad=[]):
    bad_count = defaultdict(int)
    for b in bad:
        bad_count[b] += 1
    def is_good(x):
        bad_count[x] -= 1
        return bad_count[x] < 0
    return ifilter(is_good, it)


class PrioObj(object):
    def __init__(self, val, le):
        self.val = val
        self.le  = le

    def __str__(self):
        return str(self.val)

    def __le__(self, other):
        return self.le(self.val, other.val)


def imerge(le, *iters):
    nxtheap = []
    _le = lambda a, b: le(a[0], b[0])
    for i in iters:
        try:
            it = iter(i)
            nxtheap.append(PrioObj((it.next(), it), _le))
        except StopIteration:
            pass
    heapify(nxtheap)
    while nxtheap:
        wrapper = heappop(nxtheap)
        x, it = wrapper.val
        yield x
        try:
            wrapper.val = (it.next(), it)
            heappush(nxtheap, wrapper)
        except StopIteration:
            pass

def uniq(seq):
    it = iter(seq)
    last = it.next()
    yield last
    for x in it:
        if x != last:
            last = x
            yield x


class LazyList(object):
    "Takes an iterator and turns it into a list, but only lazily so."

    def __init__(self, iterator):
        self.backing_it = iter(iterator)
        self.reduced    = []

    def __iter__(self):
        return LazyListIter(self)

    def grow(self):
        self.reduced.append(self.backing_it.next())

    def __len__(self):
        try:
            while True:
                self.grow()
        except StopIteration:
            return len(self.reduced)

    def longer_than(self, n):
        for i,x in enumerate(self):
            if i == n: # i is zero based
                return True
        return False

    def len_min(self, x, y):
        if self.longer_than(y - x):
            return y
        else:
            return len(self) + x


class LazyListIter(object):
    def __init__(self, ll):
        self.llist = ll
        self.idx   = 0

    def __iter__(self):
        return self

    def next(self):
        while len(self.llist.reduced) <= self.idx:
            self.llist.grow()
        x = self.llist.reduced[self.idx]
        self.idx += 1
        return x

